theory Heap
imports MapExt
uses ("traceaux.ML")
  ("obtain_tactical.ML") 
  ("heaps.ML")
  ("instdefs.ML")
begin

section {* Background Theory For Heaps *}

subsection {* Addresses *}

datatype addr = Addr nat
  
definition
  "NULL \<equiv> Addr 0"

definition
  "addr_ivl p n \<equiv> map Addr(case p of Addr a \<Rightarrow> [ a ..< a+n])"

lemma addr_ivl_base_contained[simp]:
  "n \<noteq> 0 \<Longrightarrow> p \<in> set (addr_ivl p n)"
  apply (cases p)
  apply (simp add: addr_ivl_def image_def split: addr.split)
  done


fun addr_add :: "addr \<Rightarrow> nat \<Rightarrow> addr" (infix "\<oplus>" 65)
where "addr_add (Addr p) n = Addr (p + n)"

fun addr_diff :: "addr \<Rightarrow> addr \<Rightarrow> nat" (infix "\<ominus>" 65)
where "addr_diff (Addr p) (Addr q) = p - q"

declare 
  addr_add.simps[simp del]
  addr_diff.simps[simp del]

lemma addr_add_0[simp]:
  "p \<oplus> 0 = p"
  apply (cases p)
  apply (simp add: addr_add.simps)
  done

subsection {* Heap Content *}

datatype heap_val =
  VBool bool | VInt int | VNat nat | VPtr addr 

datatype heap = Heap "addr \<rightharpoonup> heap_val"

fun hcnt :: "heap \<Rightarrow> (addr \<rightharpoonup> heap_val)"
where "hcnt (Heap c) = c"

fun hdom :: "heap \<Rightarrow> addr set"
where "hdom (Heap c) = dom c"

lemma dom_hcnt_hdom_conv[simp]:
  "dom (hcnt h) = hdom h"
  by (cases h) simp

lemma Heap_hcnt[simp]:
  "Heap (hcnt h) = h"
  by (cases h) simp

definition
  heap_disjoint :: "heap \<Rightarrow> heap \<Rightarrow> bool" (infix "\<bottom>" 70)
where
  "h \<bottom> h' \<equiv> hdom h \<inter> hdom h' = {}"

definition
 heap_union :: "heap \<Rightarrow> heap \<Rightarrow> heap" (infixr "\<uplus>" 70)
where
 "h \<uplus> h' \<equiv> Heap (hcnt h ++ hcnt h')"

definition
  "empty_heap \<equiv> Heap Map.empty"

lemma hdom_union[simp]:
  "hdom (h \<uplus> h') = hdom h \<union> hdom h'"
  by (cases h, cases h') (auto simp add: Un_commute heap_union_def)

lemma hdom_empty[simp]:
  "hdom empty_heap = {}"
  by (simp add: empty_heap_def)

lemma hcnt_empty[simp]:
  "hcnt empty_heap = empty"
  by (simp add: empty_heap_def)

lemma heap_op_simps[simp]:
  "(h1 \<uplus> h2) \<bottom> h3 = (h1 \<bottom> h3 \<and> h2 \<bottom> h3)"
  "h3  \<bottom> (h1 \<uplus> h2) = (h1 \<bottom> h3 \<and> h2 \<bottom> h3)"
  "h \<bottom> h' = h' \<bottom> h"
  "(h1  \<uplus> h2) \<uplus> h3 = h1 \<uplus> (h2 \<uplus> h3)"
  "h1 \<bottom> h2 \<or> h2 \<bottom> h1 \<Longrightarrow> (h1  \<uplus> h2) = h2 \<uplus> h1"
  "hcnt (h1 \<uplus> h2) = hcnt h1 ++ hcnt h2"
  apply (unfold heap_disjoint_def heap_union_def)
  apply (cases h1, cases h2, cases h3, fastforce)+
  apply (cases h1, cases h2, cases h3)
  apply (fastforce simp add: map_add_comm map_add_comm[symmetric])
  apply (fastforce simp add: map_add_comm map_add_comm[symmetric])
  done

lemma heap_union_comm:
  "h \<bottom> h' \<Longrightarrow> h \<uplus> h' = h' \<uplus> h"
  by (simp add: heap_disjoint_def heap_union_def map_add_comm[symmetric])

lemma heap_disjoint_map_restr[intro]:
  "A \<inter> B = {} \<Longrightarrow> Heap (h |` A) \<bottom> Heap (h |` B)"
  by (auto simp add: heap_disjoint_def)

lemma heap_union_map_restr[simp]:
  "Heap (h |` A) \<uplus> Heap (h |` B) = Heap (h |` (A \<union> B))"
  by (simp add: heap_union_def map_restrict_add)

 lemma heap_cnt_restrict_dom[simp]:
  "Heap (hcnt h |` hdom h) = h"
  by (cases h) simp

section {* Extensible Shallow Embedding for Types *}

record 'a cty =
  ty_size :: "nat"
  ty_rep :: "'a \<Rightarrow> heap_val list"
  ty_val  :: "heap_val list \<Rightarrow> 'a"

text {*
  Could also have chosen a low-level rendering, see
  Gast, H: Semi-automatic proofs about object graphs in separation logic.
  in Merz, S. and Luettgen, Gerald: Automated Verification of Critical Systems 
  (AVoCS '12)
*}

definition
  "ty_ok (ct :: ('a cty)) \<equiv> (\<forall>v. 
     length (ty_rep ct v) = ty_size ct \<and>      
     ty_val ct (ty_rep ct v) = v)"

lemma tyrep_size_by_ty_ok:
  "ty_ok ct \<Longrightarrow> length (ty_rep ct v) = ty_size ct"
  by (simp add: ty_ok_def)

lemma ty_val_ty_rep_inv:
  "ty_ok ct \<Longrightarrow> ty_val ct (ty_rep ct v) = v"
  by (simp only: ty_ok_def)

definition
  "cty_int :: int cty \<equiv> \<lparr> ty_size = 1, ty_rep = \<lambda>i. [VInt i ], ty_val = (\<lambda>v. case hd v of VInt i \<Rightarrow> i | _ \<Rightarrow> undefined) \<rparr>"

lemma cty_int_simps[simp]:
  "ty_ok cty_int"
  "ty_size cty_int = 1"
  by (simp_all add: ty_ok_def cty_int_def)
  
definition
  "cty_ptr :: addr cty \<equiv> \<lparr> ty_size = 1, ty_rep = \<lambda>a. [VPtr a ], ty_val = (\<lambda>v. case hd v of VPtr a \<Rightarrow> a | _ \<Rightarrow> undefined) \<rparr>"

lemma cty_ptr_simps[simp]:
  "ty_ok cty_ptr"
  "ty_size cty_ptr = 1"
  by (simp_all add: ty_ok_def cty_ptr_def)

definition
  "cty_bool :: bool cty \<equiv> 
   \<lparr> ty_size = 1, ty_rep = \<lambda>b. [VBool b], ty_val = (\<lambda>v. case hd v of VBool b \<Rightarrow> b | _ \<Rightarrow> undefined) \<rparr>"

lemma cty_bool_simps[simp]:
  "ty_ok cty_bool"
  "ty_size cty_bool = 1"
  by (simp_all add: ty_ok_def cty_bool_def)

definition
  "cty_void :: unit cty \<equiv> \<lparr> ty_size = 0, ty_rep = \<lambda>_. [], ty_val = \<lambda>_. () \<rparr>"

lemma cty_void_simps[simp]:
  "ty_ok cty_void"
  "ty_size cty_void = 0"
  by (simp_all add: ty_ok_def cty_void_def)
 
section {* Abstraction over Heap Accesses *}

definition
  heap_get :: "'a cty \<Rightarrow> addr \<Rightarrow> heap \<Rightarrow> 'a"
where
  "heap_get ct p h \<equiv> ty_val ct (map (the \<circ> hcnt h) (addr_ivl p (ty_size ct)))"

definition
  heap_put :: "'a cty \<Rightarrow> addr \<Rightarrow> 'a \<Rightarrow> heap \<Rightarrow> heap"
where
  "heap_put ct p v h \<equiv>
     Heap (hcnt h ++ (\<lambda>q. if q \<in> set (addr_ivl p (ty_size ct))
                          then Some (nth (ty_rep ct v) (q \<ominus> p))
                          else None))"

definition
  "heap_acc_cond (ct :: 'a cty) p \<equiv> 
    \<lambda>h. set (addr_ivl p (ty_size ct)) \<subseteq> hdom h"

lemma upto_shift_down:
  "[(i :: nat) ..< j] = map (\<lambda>k. i + k) [0..< j - i]"
  apply (induct j)
  apply (auto simp add: Suc_diff_le)
  done

lemma map_addr_ivl_indentity:
  "\<lbrakk> n = length xs \<rbrakk> \<Longrightarrow> map (\<lambda>q. xs ! (q \<ominus> p)) (addr_ivl p n) = xs"
  apply (clarsimp simp add: addr_diff.simps addr_ivl_def split: addr.split)
  apply (subst upto_shift_down)
  apply (simp add: map_map comp_def addr_diff.simps map_nth)
  done
  
lemma heap_get_put_inv:
  "\<lbrakk> ty_ok ct \<rbrakk> \<Longrightarrow> heap_get ct p (heap_put ct p v h) = v"
  apply (simp add: heap_get_def heap_put_def comp_def)
  apply (subst map_cong[OF refl])
  apply (auto simp add: map_addr_ivl_indentity ty_ok_def map_add_dom_app_simps)
  done

section {* Separation Logic *}

type_synonym
  hassn = "heap \<Rightarrow> bool"

definition
  "emp \<equiv> \<lambda>h. hdom h = {}"

definition
  "true \<equiv> \<lambda>h. True"

definition
  points_to :: "addr \<Rightarrow> 'a cty \<Rightarrow> 'a \<Rightarrow> hassn" ("_ \<mapsto> _. _" [60,1000,60] 60)
where
  "p \<mapsto> ct. v \<equiv>
    \<lambda>h. hdom h = set (addr_ivl p (ty_size ct)) \<and> heap_get ct p h = v \<and> p \<noteq> NULL"

definition
  star :: "hassn \<Rightarrow> hassn \<Rightarrow> hassn" (infixr "\<star>" 58)
where
  "P \<star> Q \<equiv> \<lambda>h0. \<exists>h h'. h \<bottom> h' \<and> h0 = h \<uplus> h' \<and> P h \<and> Q h'"

subsection {* Lifting Boolean Operations to Heaps *}

definition
  heap_ex :: "('a \<Rightarrow> hassn) \<Rightarrow> hassn" (binder "\<E>" 10)
where
  "\<E> x. P x \<equiv> \<lambda>h. \<exists>x. P x h"

definition
  heap_conj :: "hassn \<Rightarrow> bool \<Rightarrow> hassn" (infixl "\<bullet>" 33)
where
  "P \<bullet> c \<equiv> \<lambda>h. P h \<and> c"

lemma heap_exI[intro]:
  "\<lbrakk> P x h \<rbrakk> \<Longrightarrow> (\<E> x. P x) h"
  by (auto simp add: heap_ex_def)

lemma heap_exE[elim!]:
  "\<lbrakk> (\<E> x. P x) h;
    \<And>x. P x h \<Longrightarrow> R
   \<rbrakk> \<Longrightarrow> R"
  by (auto simp add: heap_ex_def)

lemma heap_conjI[intro]:
  "\<lbrakk> P h; Q \<rbrakk> \<Longrightarrow> (P \<bullet> Q) h"
  by (auto simp add: heap_conj_def)

lemma heap_conjE[elim!]:
  "\<lbrakk> (P \<bullet> Q) h;
     \<lbrakk> P h; Q \<rbrakk> \<Longrightarrow> R
   \<rbrakk> \<Longrightarrow> R"
  by (auto simp add: heap_conj_def)

lemmas seplogic_defs =
  heap_disjoint_def heap_union_def   
  points_to_def star_def emp_def
  heap_ex_def heap_conj_def 

lemma extract_aux_symbols[simp]:
  "\<And>P Q. (\<E> x. P x) \<star> Q = (\<E> x. P x \<star> Q)"
  "\<And>P Q. P \<star> (\<E> x. Q x) = (\<E> x. P \<star> Q x)"
  "\<And>P Q c. (P \<bullet> c) \<star> Q = ((P \<star> Q) \<bullet> c)"
  "\<And>P Q c. P \<star> (Q \<bullet> c) = ((P \<star> Q) \<bullet> c)"
  "\<And>P Q R. (P \<star> Q) \<star> R = (P \<star> (Q \<star> R))"
  "\<And>P Q. ((\<E> x. P x) \<bullet> c) = (\<E> x. P x \<bullet> c)"
  "\<And>P Q c. (P \<bullet> c \<bullet> d) = (P \<bullet> (c \<and> d))"
  "\<And>P. (P \<bullet> True) = P"
  apply (auto intro!: ext simp add: seplogic_defs)
  apply (rename_tac ha hb hc)
  apply (simp only: ex_simps[symmetric])
  apply (rule_tac x=hb in exI)
  apply (rule_tac x="hc \<uplus> ha" in exI)
  apply (rule_tac x=hc in exI)
  apply (rule_tac x=ha in exI)
  apply (simp add: Int_Un_distrib Int_Un_distrib2 heap_union_def)
  
  apply (rename_tac ha hb hc)
  apply (simp only: ex_simps[symmetric])
  apply (rule_tac x="hb \<uplus> ha" in exI)
  apply (rule_tac x=hc in exI)
  apply (rule_tac x=ha in exI)
  apply (rule_tac x=hb in exI)
  apply (simp add: Int_Un_distrib Int_Un_distrib2 heap_union_def)
  apply (simp add: map_add_comm  map_add_comm[symmetric])
  done

lemma sep_trivial[simp]:
  "(P \<bullet> True) = P"
  "(P \<bullet> False) = (\<lambda>x. False)"
  "(P \<star> emp) = P"
  "(emp \<star> Q) = Q"
  "(\<lambda>h. False) \<star> Q = (\<lambda>h. False)"
  "P \<star> (\<lambda>h. False) = (\<lambda>h. False)"
  apply (auto intro!: ext simp add: seplogic_defs)
  apply (case_tac h, case_tac h', simp)
  apply (case_tac h0, simp)
  apply (rule_tac x="Heap fun" in exI)
  apply (rule_tac x="empty_heap" in exI)
  apply simp 
  apply (case_tac h, case_tac h')
  apply simp
  apply (rule_tac x="empty_heap" in exI)
  apply (rule_tac x="h0" in exI)
  apply simp
  done
  
lemma drop_True_conj:
  "(True \<and> P) = P"
  "(P \<and> True) = P"
  by auto

lemma points_to_NULL[simp]:
  "points_to NULL ct v = (\<lambda>h. False)"
  by (simp add: points_to_def)

section {* Basic Properties *}

lemma star_assoc:
  "\<And>P Q R. (P \<star> Q) \<star> R = (P \<star> (Q \<star> R))"
  by simp (* proven earlier *)

lemma star_comm:
  "P \<star> Q = Q \<star> P"
  apply (auto intro!: ext simp add: seplogic_defs)
  apply (fastforce simp add: map_add_comm[symmetric])
  apply (fastforce simp add: map_add_comm[symmetric])
  done

lemma star_swap_front:
  "P \<star> Q \<star> R = Q \<star> P \<star> R"
  apply (subst star_assoc[symmetric])
  apply (subst star_comm)
  apply simp
  done

lemmas star_ac =
  star_comm star_assoc star_swap_front

lemma star_extractLL:
  "(P \<star> Q) \<star> R = P \<star> (Q \<star> R)"
  by simp

lemma star_extractLR:
  "(Q \<star> P) \<star> R = P \<star> (Q \<star> R)"
  by (simp add: star_swap_front)

lemma star_match_components:
  "\<lbrakk> (P \<star> Q) h;
     \<And>h. Q h \<Longrightarrow> Q' h
   \<rbrakk> \<Longrightarrow> (P \<star> Q') h"
  apply (clarsimp simp add: seplogic_defs)
  apply (rule_tac x=h in exI)
  apply (rule_tac x=h' in exI)
  apply simp
  done

text {* A subset of extract aux *}
lemma norm_heap_simps:
  "\<And>P Q. (\<E> x. P x) \<star> Q = (\<E> x. P x \<star> Q)"
  "\<And>P Q. P \<star> (\<E> x. Q x) = (\<E> x. P \<star> Q x)"
  "\<And>P Q. ((\<E> x. P x) \<bullet> c) = (\<E> x. P x \<bullet> c)"
  "\<And>P Q c. (P \<bullet> c) \<star> Q = ((P \<star> Q) \<bullet> c)"
  "\<And>P Q c. P \<star> (Q \<bullet> c) = ((P \<star> Q) \<bullet> c)"
  "\<And>P Q c. (P \<bullet> c \<bullet> d) = (P \<bullet> (c \<and> d))"
  by simp_all

lemma normalize_heap_goal[simp]:
  "\<And>P. (\<exists>x. (\<E> y. P x y) h) = (\<E> x y. P x y) h"
  "\<And>P. ((\<E> x. P x) h \<and> Q) = (\<E> x. P x \<bullet> Q) h"
  "\<And>P. ((P \<bullet> Q) h \<and> R) = (P \<bullet> (Q \<and> R)) h"
  by (auto simp add: heap_ex_def heap_conj_def)

lemma heap_unify_with_emp:
  "P h \<Longrightarrow> (P \<star> emp) h"
  by simp

section {* Derived Inequalities of Pointers *}

definition
  "union_list as = (foldr (op \<union>) as {})"

lemma union_list_simps[simp]:
  "union_list [] = {}"
  "union_list (a # as) = a \<union> union_list as"
  by (simp_all add: union_list_def)

lemma union_list_append[simp]:
  "union_list (as @ bs) = union_list as \<union> union_list bs"
  by (induct as) auto

lemma union_list_map:
  "union_list (map f xs) = (\<Union>x \<in> set xs. f x)"
  by (induct xs) auto

lemma union_list_singletons[simp]:
  "union_list (map (\<lambda>x. {x}) xs) = set xs"
  "union_list (map ((\<lambda>x. {x}) \<circ> f) xs) = f ` set xs"
  by (induct xs, simp_all)

fun
  disjoint_list :: "addr set list \<Rightarrow> heap \<Rightarrow> bool"
where
  "disjoint_list [] h = True"
| "disjoint_list (a # as) h = (a \<subseteq> hdom h \<and> NULL \<notin> a \<and> a \<inter> union_list as = {} \<and> disjoint_list as h)"

lemma disjoint_list_heap_union[intro]:
  "disjoint_list as h \<Longrightarrow> disjoint_list as (h \<uplus> h')"
  "disjoint_list as h' \<Longrightarrow> disjoint_list as (h \<uplus> h')"
  by (induct as) auto

lemma disjoint_list_unionD:
  "disjoint_list as h \<Longrightarrow> union_list as \<subseteq> hdom h"
  by (induct as) auto

lemma disjoint_list_star:
  "\<lbrakk> (P \<star> Q) h;
     \<And>h'. P h' \<Longrightarrow> disjoint_list as h';
     \<And>h'. Q h' \<Longrightarrow> disjoint_list bs h'
   \<rbrakk> \<Longrightarrow> disjoint_list (as @ bs) h"
  apply (induct as)
  apply (fastforce simp add: star_def)
  apply (clarsimp simp add: star_def heap_disjoint_def)
  apply (rule conjI)
  apply fast
  apply (simp add: Int_Un_distrib)
  apply (drule meta_spec, drule(1) meta_mp)+
  apply (clarsimp dest!: disjoint_list_unionD simp add: hdom_def)
  apply blast
  done

lemma disjoint_list_empty:
  "P \<Longrightarrow> disjoint_list [] h"
  by simp

lemma disjoint_list_append:
  "disjoint_list (A @ B) h
   = (disjoint_list A h \<and> disjoint_list B h \<and>
      union_list A \<inter> union_list B = {})"
  apply (induct A)
  apply auto
  done

lemma disjoint_list_mapD:
  "disjoint_list (map (\<lambda>p. {p}) xs) h \<Longrightarrow> distinct xs"
  "disjoint_list (map ((\<lambda>p. {p}) \<circ> f) ys) h \<Longrightarrow> distinct (map f ys)"
  apply (induct xs)
  apply auto
  apply (induct ys)
  apply auto
  done
  
lemma disjoint_list_destructD[dest!]:
  "disjoint_list (Ps @ Qs) h
  \<Longrightarrow> disjoint_list Ps h \<and> disjoint_list Qs h \<and> union_list Ps \<inter> union_list Qs = {}"
  "disjoint_list [] h \<Longrightarrow> True"
  "disjoint_list (P # Ps) h
  \<Longrightarrow> NULL \<notin> P \<and> P \<inter> union_list Ps = {} \<and> disjoint_list Ps h"
  "NULL \<noteq> p \<Longrightarrow> p \<noteq> NULL"
  "disjoint_list (map (\<lambda>p. {p}) xs) h
   \<Longrightarrow> distinct xs"
  "disjoint_list (map ((\<lambda>p. {p}) \<circ> g) ys) h 
   \<Longrightarrow> distinct (map g ys)"
  apply (simp_all add: disjoint_list_append)
  apply (fastforce dest: disjoint_list_mapD)+
  done

declare
  disjoint_list.simps[simp del]

section {* Side-conditions from heap components *}

lemma side_cond_star:
  "\<lbrakk> (P \<star> Q) h; \<And>h. P h \<Longrightarrow> c; \<And>h. Q h \<Longrightarrow> d \<rbrakk> \<Longrightarrow> c \<and> d"
  by (auto simp add: seplogic_defs)

lemma side_cond_TrueI:
  "P h \<Longrightarrow> True"
  by auto

lemma disjoint_sym:
  "A \<inter> B = {} \<Longrightarrow> B \<inter> A = {}"
  by auto

section {* Interaction Heap and Accesses *}

lemma heap_get_resolve_simp:
  "(p \<mapsto> ct. v \<star> R) h \<Longrightarrow> heap_get ct p h = v"
  apply (auto intro!: ext simp add: seplogic_defs heap_get_def)  
  apply (simp add: comp_def)
  apply (subst map_cong[OF refl])
  apply (subst map_add_dom_app_simps(3)) 
  apply (simp, blast)
  apply simp_all
  done

lemma heap_put_hdom:
  "set (addr_ivl p (ty_size ct)) \<subseteq> hdom h 
   \<Longrightarrow> hdom (heap_put ct p v h) = hdom h"
  apply (cases h)
  apply (fastforce simp add: heap_put_def)
  done

lemma heap_upd_forward_base:
  "\<lbrakk> (p \<mapsto> ct. a \<star> H) h;
     ty_ok ct;
     \<And>h. (p \<mapsto> ct. v \<star> H) h \<Longrightarrow> Q h
   \<rbrakk> \<Longrightarrow>  Q (heap_put ct p v h)"
  apply (clarsimp simp add: points_to_def star_def heap_get_put_inv)
  apply (drule meta_spec, drule meta_mp)
  prefer 2
  apply assumption
  apply (rule_tac x="heap_put ct p v h" in exI)
  apply (rule_tac x="h'" in exI)
  apply (simp add: heap_disjoint_def heap_get_put_inv heap_put_hdom Int_commute)
  apply (simp add: heap_put_def heap_union_def map_add_comm map_add_comm[symmetric])
  done
 
lemma heap_acc_condI:
  "(p \<mapsto> ct. a \<star> H) h \<Longrightarrow> heap_acc_cond ct p h"
  by (clarsimp simp add: heap_acc_cond_def star_def points_to_def)

section {* Symbolic Execution *}

definition
  "HeapOp \<equiv> Let"

syntax
  "_HeapOp" :: "id \<Rightarrow> logic \<Rightarrow> logic \<Rightarrow> logic" ("heapop _ \<leftarrow> _; //_" [ 0,0,10] 10)

translations
  "heapop x \<leftarrow> s; e" == "CONST HeapOp s (%x. e)"

lemma heap_op_unfold:
  "HeapOp = (\<lambda>s e. e s)"
  by (intro ext) (simp only: HeapOp_def Let_def)

lemma heap_put_forward_heap_op:
  "\<lbrakk> (p \<mapsto> ct. a \<star> H) h; 
     ty_ok ct; 
     \<And>h. (p \<mapsto> ct. v \<star> H) h \<Longrightarrow> Q h
   \<rbrakk> \<Longrightarrow> HeapOp (heap_put ct p v h) Q"
  apply (rule heap_upd_forward_base, assumption+)
  apply (simp add: HeapOp_def)
  done

lemma heap_get_forward_heap_op:
  "\<lbrakk> (p \<mapsto> ct. v \<star> H) h; 
     ty_ok ct; 
     Q v
   \<rbrakk> \<Longrightarrow> HeapOp (heap_get ct p h) Q"
  by (simp add: heap_get_resolve_simp HeapOp_def)

section {* Tactic Support *}

ML_file "traceaux.ML"
ML_file "obtain_tactical.ML" 
ML_file "instdefs.ML"
setup InstDefs.setup

ML_file "heaps.ML"
setup Heaps.setup

lemma points_to_addrs:
  "\<lbrakk> ty_size ct \<noteq> 0; (p \<mapsto>ct. v) h \<rbrakk> \<Longrightarrow> disjoint_list [ {p} ] h"
  by (clarsimp simp add: points_to_def disjoint_list.simps)
 
declare points_to_addrs[of cty_int, simplified, heap_addrs]
declare points_to_addrs[of cty_bool, simplified, heap_addrs]

lemma tst_forw_disjoint_cache_heap_structure:
  "\<And>x y z. \<lbrakk> (p \<mapsto>cty_int. a \<star> q \<mapsto>cty_bool. b) h \<rbrakk> \<Longrightarrow> p \<noteq> q \<and> q \<noteq> p"
  "\<And>x y z. \<lbrakk> (q \<mapsto>cty_bool. b \<star> p \<mapsto>cty_int. a ) h' \<rbrakk> \<Longrightarrow> p \<noteq> q"
  "\<And>x y z. \<lbrakk> (p \<mapsto>cty_int. a \<star> q \<mapsto>cty_bool. b) h' \<rbrakk> \<Longrightarrow> p \<noteq> q"
  apply simp
  apply simp
  apply simp
  done

lemma tst_forw_disjoint_contradiction:
  "\<And>x y z. \<lbrakk> (p \<mapsto>cty_int. a \<star> p \<mapsto>cty_bool. b) h \<rbrakk> \<Longrightarrow> False"
  apply fwd_disjoint
  apply simp
  done

lemma test_direct_matching:
  "A (h :: heap)
  \<Longrightarrow> A h"
  "(A \<star> B) (h :: heap)
  \<Longrightarrow> (A \<star> B) h"
  "(A \<star> B) (h :: heap)
  \<Longrightarrow> (B \<star> A) h"
  "(A \<star> B \<star> C) (h :: heap)
  \<Longrightarrow> (B \<star> A \<star> C) h"
  "(A \<star> B \<star> C) (h :: heap)
  \<Longrightarrow> (C \<star> A \<star> B) h"
  "(A \<star> B \<star> C) (h :: heap)
  \<Longrightarrow> (C \<star> B \<star> A) h"
  "(A \<star> B \<star> C \<star> D) (h :: heap)
  \<Longrightarrow> (B \<star> C \<star> A \<star> D) h"
  "(A \<star> (B \<star> C) \<star> D) (h :: heap)
  \<Longrightarrow> (B \<star> C \<star> A \<star> D) h"
  "((A \<star> (B \<star> C)) \<star> D) (h :: heap)
  \<Longrightarrow> (B \<star> C \<star> A \<star> D) h"
  "((A \<star> (B \<star> C)) \<star> D) (h :: heap)
  \<Longrightarrow> (\<E> R. (B \<star> D \<star> R)) h"
  by heap_cancel+

lemma test_explicit_exI:
  "(\<E> a f. p \<mapsto>cty_int. a \<bullet> f = (F :: int \<Rightarrow> int)) heap
  \<Longrightarrow> (\<E> a f. p \<mapsto>cty_int. a \<bullet> f = F) heap"
  apply (erule heap_exE heap_conjE)+
  apply (rule heap_exI)+
  using [[trace_heaps]]
  apply heap
  done

lemma test_implicit_nested_exI:
  "(\<E> a f. p \<mapsto>cty_int. a \<bullet> f = (F :: int \<Rightarrow> int)) heap
  \<Longrightarrow> \<exists>a b. (\<E> f. p \<mapsto>cty_int. a \<bullet> f = F) heap \<and> a = b"
  apply simp
  apply heap
  done

section {* Test Unfoldings with Lists *}

definition
  "fo_data :: nat \<equiv> 0"

definition
  "fo_next :: nat \<equiv> 1"

definition
  "node p vn \<equiv> \<lambda>h. let (v,n) = vn 
        in (p \<oplus> fo_data \<mapsto>cty_int. v \<star> p \<oplus> fo_next \<mapsto>cty_ptr. n) h"

lemma node_NULL[simp]:
  "node NULL v = (\<lambda>h. False)"
  by (simp add: node_def fo_data_def)

fun list :: "addr \<Rightarrow> int list \<Rightarrow> addr \<Rightarrow> hassn"
where
  "list p [] q = (emp \<bullet> (p = q))"
| "list p (x # xs') q = (\<E> n. node p (x, n) \<star> list n xs' q)"

lemma list_append[simp]:
  "list p (xs @ ys) q = (\<E> r. list p xs r \<star> list r ys q)"
  by (induct xs arbitrary: p) (auto intro!: ext)

lemma list_NULL_NULL[simp]:
  "list NULL xs NULL = (emp \<bullet> (xs = []))"
  by (cases xs) (auto simp add: heap_ex_def)

lemma node_unfold_ex:
  "node p c = (\<E> v' n'. p \<oplus> fo_data \<mapsto> cty_int. v' \<star> p \<oplus> fo_next \<mapsto> cty_ptr. n' \<bullet> v' = fst c \<and> n' = snd c)"
  apply (rule ext)
  apply (fastforce intro: heap_exI heap_conjI simp add: node_def)
  done

lemma node_unfold_direct:
  "node p (d,n) = p \<oplus> fo_data \<mapsto>cty_int.  d \<star> p \<oplus> fo_next \<mapsto>cty_ptr. n"
  by (simp add: node_def)

lemma list_step:
  "p \<noteq> q \<Longrightarrow> list p xs q = (\<E> x n xs'. node p (x, n) \<star> list n xs' q \<bullet> xs = x # xs')"
  apply (rule ext)
  apply (cases xs)
  apply force
  apply (rule iffI)
  apply (force simp add: seplogic_defs)
  apply (erule heap_exE heap_conjE)+
  apply (auto simp add: seplogic_defs)
  done

declare list_step[sep_unfold]
declare node_unfold_direct[sep_unfold]
declare list_append[sep_unfold]

lemma test_right_unfoldings:
  "\<lbrakk> (node q a \<star> p \<oplus> fo_data \<mapsto>cty_int. x \<star> p \<oplus> fo_next \<mapsto>cty_ptr. n \<star> list n xs' q) h;
     p \<noteq> q; xs = x # xs'; p \<noteq> NULL
   \<rbrakk> \<Longrightarrow> (list p xs q \<star> node q a) h"
  "(p \<oplus> fo_next \<mapsto>cty_ptr. n \<star> node q z \<star> p \<oplus> fo_data \<mapsto>cty_int. d) h
   \<Longrightarrow> (node q z \<star> node p (d,n)) h"
  by heap+
  
lemma test_left_unfoldings:
  "(node p (d,n)) h
   \<Longrightarrow> (p \<oplus> fo_next \<mapsto>cty_ptr.  n \<star> p \<oplus> fo_data \<mapsto>cty_int.  d) h"
  "\<lbrakk> (list p xs q \<star> node q (y,y')) h
   \<rbrakk> \<Longrightarrow> (\<E> n R. (q \<oplus> fo_data \<mapsto>cty_int.  n) \<star> R) h"
  "\<lbrakk> (list p xs q \<star> node q x) h; p \<noteq> q
   \<rbrakk> \<Longrightarrow> (\<E> x a b R. (node q a \<star> (p \<oplus> fo_data \<mapsto>cty_int.  x) \<star> R)) h"
  by heap+
 
lemma test_ex_both_unfolding:
  "(\<E> d n. node p (d,n)) h \<Longrightarrow> (\<E> x y. node p (x,y)) h"
  "(\<E> d n. node p (d,n) \<bullet> d > 0) h \<Longrightarrow> (\<E> x y. node p (x,y)) h"
  "\<lbrakk> (list p xs q \<star> node q x) h; p \<noteq> q
   \<rbrakk> \<Longrightarrow> (\<E> x a b R. (node q a \<star> (p \<oplus> fo_data \<mapsto>cty_int. x) \<star> R)) h"
  by heap+

lemma test_ex_nested_unfolds_unfolding:
  "\<lbrakk> (list p xs q \<star> node q x) h; p \<noteq> q
   \<rbrakk> \<Longrightarrow> \<exists>a b. (\<E> x R. (node q a \<star> (p \<oplus> fo_data \<mapsto>cty_int. x) \<star> R)) h \<and> a = b"
  apply (rule exI conjI)+
  apply heap
  apply simp
  done

lemma test_heap_conj_unfold:
  "\<lbrakk> (node p (d,n) \<star> list n xs q) h; p \<noteq> q
   \<rbrakk> \<Longrightarrow> (list p (d # xs) q \<bullet> (p \<noteq> q)) h"
  by heap

lemma test_emp_frame:
  "(p \<mapsto>ct. A \<star> q \<mapsto>ct. B) h \<Longrightarrow> (\<E> x y R. q \<mapsto>ct. y \<star> p \<mapsto>ct. x \<star> R) h"
  by heap

lemma test_frame_compute:
  "(p \<mapsto>ct. A \<star> q \<mapsto>ct. B \<star> r \<mapsto>ct. C)  h \<Longrightarrow> \<exists>y R. (q \<mapsto>ct. y \<star> R) h"
  apply (rule exI)+
  apply heap
  done

end