theory SimplC
imports Vcg Heap
keywords
  "verify_statement" :: thy_goal and
  "define_struct" :: thy_decl
uses
  ("mlyacclib/MLY_base-sig.ML")
  ("mlyacclib/MLY_join.ML")
  ("mlyacclib/MLY_lrtable.ML")
  ("mlyacclib/MLY_stream.ML")
  ("mlyacclib/MLY_parser2.ML")
  ("Pos.ML")
  ("Ast.ML")
  ("decls.ML")
  ("ghost.ML")
  ("simplc.grm.sig")
  ("simplc.grm.ML")
  ("simplc.lex.ML")
  ("parser.ML")
  ("PrimTab.ML")
  ("traceaux.ML") 
  ("TAst.ML")
  ("structdef.ML")
  ("prepvc.ML")
  ("sym_exec.ML")
  ("simplc_goals.ML")
begin

section {* Basic Background Theory *}

datatype throw_case = ThrowBreak | ThrowReturn

datatype guards = ArrayBounds | HeapAlloc

section {* Theory of Arrays *}

datatype 'a array = Array "int \<times> (int \<Rightarrow> 'a)"

definition
  "amap f a \<equiv> case a of Array a' \<Rightarrow> Array (f a')"

definition
  "aacc f a \<equiv> case a of Array a' \<Rightarrow> f  a'"

definition
  "ainit n \<equiv> Array (n, (\<lambda>i. undefined))"

definition
  "aset i x a \<equiv> amap (\<lambda>(n,c). (n, c(i := x))) a"

definition
  "aget a i \<equiv> aacc (\<lambda>(n,c). c i) a"

definition
  "acheck i a \<equiv> aacc (\<lambda>(n,c). 0 \<le> i \<and> i < n) a"

definition
  "alength a \<equiv> aacc (\<lambda>(n,c). n) a"

lemmas array_defs = 
  amap_def aacc_def 
  ainit_def aget_def aset_def acheck_def alength_def

lemma array_simps[simp]:
  "aget (aset i x a) i  = x"
  "i \<noteq> j \<Longrightarrow> aget (aset j x a) i  = aget a i"
  "acheck i a = (0 \<le> i \<and> i < alength a)"
  "alength (ainit n) = n"
  "alength (aset i x a) = alength a"
  by (simp_all add: array_defs split: array.split)

definition
  "elems a i j \<equiv> map (aget a \<circ> int) [nat i ..< nat j]"

lemma elems_empty[simp]:
  "elems a i i = []"
  by (simp add: elems_def)

lemma elems_single[simp]:
  "\<lbrakk> 0 \<le> i \<rbrakk> \<Longrightarrow> elems a i (i + 1) = [aget a i]"
  by (simp add: elems_def nat_add_distrib)
  
lemma elems_single'[simp]:
  "\<lbrakk> 0 < i \<rbrakk> \<Longrightarrow> elems a (i - 1) i = [aget a (i - 1)]"
  apply (subgoal_tac "\<exists>k. i = k + 1")
  apply clarsimp
  apply arith
  done
 
lemma elems_sep[simp]:
  "k \<notin> { i..<j } \<Longrightarrow> elems (aset k y a) i j = elems a i j"
  by (auto simp add: elems_def)

lemma elems_split:
  "\<lbrakk> 0 \<le> i; i \<le> k; k \<le> j \<rbrakk> \<Longrightarrow> elems a i j = elems a i k @ elems a k j"
  apply (simp add: elems_def)
  apply (subgoal_tac "\<exists>d. d \<ge> 0 \<and> j = k + d")
  apply clarsimp
  txt {* find_theorems "nat (?a + ?b)" *}
  apply (simp add: nat_add_distrib)
  apply (subst upt_add_eq_append)
  apply simp
  apply (simp only: map_append[symmetric])
  txt {* Typisch: Bei Existenzquantoren muss man Hilfestellung geben *}
  apply (rule_tac x="j - k" in exI)
  apply auto
  done

lemma elems_split_first:
  "\<lbrakk> 0 \<le> i; i < j \<rbrakk> \<Longrightarrow> elems a i j = aget a i # elems a (i+1) j"
  by (rule trans, rule_tac k="i + 1" in elems_split, auto)

lemma elems_split_last:
  "\<lbrakk> 0 \<le> i; i < j \<rbrakk> \<Longrightarrow> elems a i j = elems a i (j - 1) @ [ aget a (j - 1) ]"
  by (rule trans, rule_tac k="j - 1" in elems_split, auto)
  
lemma elems_split_update:
  "\<lbrakk> 0 \<le>  i; i \<le> k; k < j 
   \<rbrakk> \<Longrightarrow> elems (aset k x a) i j = elems a i k @ [ x ] @  elems a (k + 1) j"
  apply (subst elems_split)
  apply assumption+
  apply simp
  apply simp
  apply (subst elems_split)
  apply simp
  apply (rule less_add_one[THEN less_imp_le])
  apply simp
  apply simp
  done

lemma elems_length[simp]:
  "\<lbrakk> 0 \<le> i; i \<le> j \<rbrakk> \<Longrightarrow> length (elems a i j) = nat (j - i)"
  by (simp add: elems_def)

lemma elems_prefix:
  "\<lbrakk> 0 \<le> i; i \<le> k; k \<le> j \<rbrakk> \<Longrightarrow> elems a i k = take (nat (k - i)) (elems a i j)"
  apply (subst elems_split) back
  txt {* Nach einigen simp/assumption habe ich das elems-length lemma gefunden *}
  apply auto
  done

section {* Records and the Heap *}

definition
  "field_get f p \<equiv> f p"

definition
  "field_upd f p x \<equiv> f (p := x)"

lemma field_get_simps[simp]:
  "p \<noteq> q \<Longrightarrow> field_get (field_upd f p v) q = field_get f q"
  "field_get (field_upd f p v) p = v"
  by (simp add: field_get_def field_upd_def)+
 
text {*
Simpl defines its own encoding of field accesses without
auxiliary constants. We therefore remove its syntax setup,
but re-introduce it for output only.
*}
no_syntax
"_faccess"  :: "'ref \<Rightarrow> ('ref \<Rightarrow> 'v) \<Rightarrow> 'v"
   ("_\<rightarrow>_" [65,1000] 100)

syntax
  "_field_get"  :: "addr \<Rightarrow> (addr \<Rightarrow> 'v) \<Rightarrow> 'v"
   ("_\<rightarrow>_" [65,1000] 100)

translations
  "p \<rightarrow> f" == "CONST field_get f p"

syntax (output)
 "_faccess"  :: "'ref \<Rightarrow> ('ref \<Rightarrow> 'v) \<Rightarrow> 'v"
   ("_\<rightarrow>_" [65,1000] 100)


ML_file "prepvc.ML"
setup PrepVC.setup

ML_file "sym_exec.ML"
setup SymExec.setup

section {* The embedding of SimplC *}

ML_file "mlyacclib/MLY_base-sig.ML";
ML_file "mlyacclib/MLY_join.ML";
ML_file "mlyacclib/MLY_lrtable.ML";
ML_file "mlyacclib/MLY_stream.ML";
ML_file "mlyacclib/MLY_parser2.ML";

ML_file "Pos.ML";
ML_file "Ast.ML";
ML_file "decls.ML";
ML_file "ghost.ML";

ML_file "simplc.grm.sig";
ML_file "simplc.grm.ML";
ML_file "simplc.lex.ML";
ML_file "parser.ML";

section {* Embedding into Simpl *}

ML_file "PrimTab.ML"
ML_val {* 
let open Ast 
    val c = PrimTab.choose (EVar "x")
in
    [ 
      c OP_PLUS [TInt, TInt],
      c OP_PLUS [TPtr TInt, TInt],
      c OP_EQUAL [TInt, TInt],
      c OP_EQUAL [TPtr TInt, TPtr TInt],
      c OP_EQUAL [TPtr TInt, TPtr TVoid]
    ]
end
*}
ML_file "traceaux.ML"
ML_file "TAst.ML"

ML_file "simplc_goals.ML"
setup SimplC_Goals.setup

ML_file "structdef.ML"

method_setup split_vc = {* 
  Scan.succeed (fn ctx => SIMPLE_METHOD' (REPEAT_ALL_NEW (resolve_tac [@{thm conjI}, @{thm impI}])))
*}

declare [[heap_model=seplogic]]

define_struct
"struct tst {
  int x;
  int y;
}"

verify_statement sep_deref_read
vars: "int *p; int r;"
pre: "(p\<mapsto>cty_int. X) heap"
post: "(p\<mapsto>cty_int. X \<bullet> r = X) heap"
{*
  r = *p;
*}
  apply vcg
  apply prep_vc
  apply run
  apply simp
  done

verify_statement sep_deref_write
vars: "int *p; int r;"
pre: "(p\<mapsto>cty_int. X \<bullet> r = R) heap"
post: "(p\<mapsto>cty_int. R \<bullet> r = R) heap"
{*
  *p = r;
*}
  apply vcg
  apply prep_vc
  apply run
  done

verify_statement sep_multi_write
vars: "int *p; int *q; int *r;"
pre: "(p\<mapsto>cty_int. X \<star> q\<mapsto>cty_int. Y \<star> r\<mapsto>cty_int. Z) heap"
post: "(p\<mapsto>cty_int. 1 \<star> q\<mapsto>cty_int. 2 \<star> r\<mapsto>cty_int. 3) heap"
{*
  *p = 1;
  *q = 2;
  *r = 3;
*}
  apply vcg
  apply prep_vc
  apply step
  apply step
  apply step
  apply heap
  done


declare tst_unfold[sep_unfold]
verify_statement sep_read_struct
vars: "struct tst*p; int r;"
pre: "(tst p v) heap"
post: "(p\<oplus>fo_tst_x\<mapsto>cty_int. tst_x v \<star> p\<oplus>fo_tst_y\<mapsto>cty_int. tst_y v \<bullet> r = tst_x v) heap"
{*
  r = p->x;
*}
  apply vcg
  apply prep_vc
  apply step
  apply simp
  done

verify_statement sep_write_struct
vars: "struct tst*p; int r;"
pre: "(tst p v \<bullet> r = R) heap"
post: "(p\<oplus>fo_tst_x\<mapsto>cty_int. R \<star> p\<oplus>fo_tst_y\<mapsto>cty_int. tst_y v) heap"
{*
  p->x = r;
*}
  apply vcg
  apply prep_vc
  apply step
  apply simp
  done
  
declare [[heap_model=burstall]]

verify_statement null_literal
vars: "int *p;"
pre: "True"
post: "p = NULL"
{* p = null; *}
  apply vcg
  done
verify_statement short_circuit_and
vars: "int a[]; int i; int n;"
pre: "n = alength a"
post: "0 \<le> i \<and> i < n \<and> aget a i > 0 \<longrightarrow> i = 0"
{*
  if (0 <= i && i <  n && a[i] > 0)
    i = 0;
*}
  apply vcg
  apply auto
  done

verify_statement short_circuit_or
vars: "int a[]; int i; int n;"
pre: "n = alength a"
post: "0 > i \<or> i \<ge> n \<and> aget a i > 0 \<longrightarrow> i = 0"
{*
  if (i < 0 || i >= n || a[i] > 0)
    i = 0;
*}
  apply vcg
  apply auto
  done

verify_statement ghost_test
vars: "int i;"
ghost: 
  lst :: "int list"
pre: "i = I"
post: "lst = [I]"
{*
//@ lst = [ i ]
*}
  apply vcg
  apply simp
  done

verify_statement simple_assign
vars: {*
  int i;
  int j;
*}
pre: "i = I"
post: "j = I \<and> i = I"
{* j = i; *}
  apply vcg
  apply simp
  done

verify_statement inner_effect
vars: {*
  int a[];
  int i;
*}
pre: "0 \<le> i \<and> i + 1 < alength a \<and> aget a (i + 1) = K"
post: "i = K"
{*
i = a[i = i + 1];
*}
  apply vcg
  apply simp
  done

verify_statement array_get
vars: {*
  int a[];
  int i;
  int j;
*}
pre: "aget a i = K \<and> 0 \<le> i \<and> i < alength a"
post: "j = K"
{*
j = a[i];
*}
  apply vcg
  apply (simp(no_asm_use))
  apply simp
  done

verify_statement array_set
vars: {*
  int a[];
  int i;
*}
pre: "0 \<le> i \<and> i < alength a"
post: "aget a i = 42"
{*
a[i] = 42;
*}
  apply vcg
  apply simp
  done

verify_statement array_set2
vars: {*
  int a[];
  int i;
*}
pre: "0 \<le> i \<and> i < alength a \<and> aget a i = K"
post: "aget a i = K + 3"
{*
a[i] = a[i] + 1;
a[i] = a[i] + 2;
*}
  apply vcg
  apply simp
  done

verify_statement swap
vars: {*
 int i;
 int j;
 int t;
*}
pre: "i = I \<and> j = J"
post: "i = J \<and> j = I"
{*
 t = i;
 i = j;
 j = t;
*}
  apply vcg
  apply simp
  done

verify_statement incr2
vars: {*
 int i;
 int j;
*}
pre: "i = I"
post: "j = I + 1"
{*
 j = i + 1;
*}
  apply vcg
  apply simp
  done

verify_statement cases2
vars: {*
  int i;
  int r;
*}
pre: "i = I"
post: "r = (if I > 0 then 1 else 2)"
{*
if (i > 0)
   r = 1;
else
   r = 2;
*}
  apply vcg
  apply simp
  done

verify_statement cases3
vars: {*
  int i;
  int r;
*}
pre: "i = I"
post: "r = (if I - 1 > 0 then 1 else 2)"
{*
if ((i = i - 1) > 0)
   r = 1;
else
   r = 2;
*}
  apply vcg
  apply simp
  done

verify_statement count_c_style
vars: {*
  int i;
  int j;
*}
pre: "i = I \<and> i \<ge> 0"
post: "j = I"
{* j = 0;
  /*@ i \<ge> 0 \<and> j = I - i */
  while ( (i = i - 1) >= 0) j = j + 1;
*}
  apply vcg
  apply auto
  done

verify_statement with_break
vars: {*
 int i;
  int j;
*}
pre: "True"
post: "i = 1"
{*
 j = 0;
 i = 0;
/*@ i = 0 \<and> j < 10 */
while (j < 10) {
  i = i + 1;
  break;
  j = j + 1;
}
*}
  apply vcg
  apply auto
  done

declare [[heap_model=burstall]]

define_struct "struct tst2 { int x; int y; }"

verify_statement field_read
vars: "struct tst2 *p; int r;"
pre: "p\<rightarrow>tst2_x = X \<and> tst2_alloc p"
post: "r = X"
  {*
  r = p->x;
  *}
  apply vcg
  apply simp
  done

verify_statement field_set
vars: "struct tst2 *p; struct tst2 *q; int k;"
pre: "p\<rightarrow>tst2_x = X \<and> q\<rightarrow>tst2_x = Y \<and> p \<noteq> q \<and> k = K \<and> tst2_alloc p \<and> tst2_alloc q"
post: "p\<rightarrow>tst2_x = K \<and> q\<rightarrow>tst2_x = Y"
{*
p->x = k;
*}
  apply vcg
  apply simp
  done

end

