structure TypedAst =
struct

(** Basic definitions of typed ASTs *)

datatype ty = datatype Ast.ty
datatype var_kind = datatype Ast.var_kind
type struct_info = Ast.struct_info

(**
  The variable context
*)
datatype vty =
  CVar of ty | GhostVar of typ

type vctx = (string * (var_kind * vty)) list

datatype var_context = VarContext of { 
   fun_name : string option,
   cur      : vctx
}

fun vctx_mk fun_name cur =
  VarContext { fun_name=fun_name,cur=cur }

fun vctx_add_cur cur' (VarContext {fun_name,cur}) =
  VarContext { fun_name=fun_name, cur = cur' @ cur }

fun vctx_get_cur (VarContext {cur,...}) v =
  case AList.lookup (op=) cur v of
      NONE => error ("Variable "^v^" not found in "^commas (map fst cur))
    | SOME r => r

fun vctx_fun_name (VarContext {fun_name,...}) = fun_name

val heap_var = "heap"  
val return_name = "ret_"  
val thrown_name = "thrown_"

fun is_global_var vctx name =
  not (name = thrown_name) andalso
  (name = heap_var orelse
   member (op=) [ Global ] (fst (vctx_get_cur vctx name)))

fun var_name vctx v =
  case vctx_fun_name vctx of
      NONE => Hoare.varname v
    | SOME fun_name =>
      if is_global_var vctx v
      then Hoare.varname v
      else Hoare.varname (suffix ("_'"^fun_name) v)
      
(** The Syntax Trees Themselves *)


val mk_locals = map (apsnd (fn t => (Local, CVar t)))

exception NotDone of string

type primop = string (* now a fully-named operator *)

datatype exp =
    EVar of string
  | ECchar of char
  | ECint of int
  | ECuint of int
  | ENull
  | EDeref of ty * exp
  | EAddr of exp
  | EArrow of exp * ty * string
  | EIndex of exp * exp
  | EAssign of ty * exp * exp
  | EPre of (primop * ty) * exp
  | EPost of (primop * ty) * exp
  | EPrim1 of (primop * ty) * exp
  | EPrim2 of (primop * ty * ty) * exp * exp
  | EFree of ty * exp
  | EFreeArray of ty * exp * exp
  | ENew of ty
  | ENewArray of ty * exp
  | EAnd of exp * exp
  | EOr of exp * exp
  | ECall of string * exp list

datatype ghost =
  GAssign of string * string

datatype sm =
    SSkip
  | SSeq of sm * sm
  | SExec of exp
  | SBlock of (string * ty) list * sm
  | SIf of exp * sm * sm
  | SWhile of string option * exp * sm
  | SBreak
  | SContinue
  | SReturn of exp
  | SGhost of ghost

fun string_of_ty (t : ty) = PolyML.makestring t
fun string_of_exp (e : exp) = PolyML.makestring e
fun string_of_sm (s : sm) = PolyML.makestring s

(** Keeping struct and function information *)

structure StructData = Generic_Data
(type T = struct_info list
 val empty = []
 val extend = I
 fun merge (a,b) = distinct (op=) (a @ b))

fun find_struct thy sname =
  find_first (fn {name,...} => name = sname)
          (StructData.get (Context.Theory thy))

fun get_struct_info thy s =
  case find_struct thy s of
      NONE => error ("Unknown struct "^s)
    | SOME info => info

fun put_struct_info s thy =
    case find_struct thy (#name s) of
        NONE => Context.theory_map (StructData.map (fn d => s :: d)) thy
      | SOME _ => error ("Re-definition of struct " ^ #name s)

fun check_struct_info thy (s : struct_info) =
  let val field_names = map fst o #fields
      val fs = field_names s
  in
      case duplicates (op=) fs of
          [] => ()
      (* force distinct field names between structs 
          (case get_first (fn s' => 
                                   case duplicates (op=) (fs @ field_names s') of
                                       [] => NONE
                                     | dups => SOME(#name s', dups))
                     (StructData.get (Context.Theory thy)) of
                     NONE => ()
                   | SOME (s',dups) =>
                     (error ("fields already used in "^s'^": "^commas dups)))
       *)
        | dups => error ("duplicate field names in " ^ #name s ^ ": "^commas dups)
  end

fun get_struct_field_type thy s f =
  case AList.lookup (op =) (#fields (get_struct_info thy s)) f of
      NONE => error ("Field "^f^" unknown in struct "^s)
    | SOME ty => ty

type fun_info = {
   name : string,
   params : (string * ty) list,
   rty    : ty
}

structure Fun_Info = Generic_Data
   (type T= fun_info Symtab.table; 
    val empty = Symtab.empty;
    val merge = Symtab.merge (K false);
    val extend = I);
    
fun get_fun_info ctx name =
  case Symtab.lookup (Fun_Info.get (Context.Proof ctx)) name of
      NONE => error ("Function "^name^" not defined")
    | SOME d => d

(** Type checking *)


fun tc_exp ctx (vs : var_context) e =
  let  fun loop (Ast.EVar x) =
         (let val (_,t) = vctx_get_cur vs x
          in
              case t of
                  CVar t => (EVar x, t)
                | _ => error("cannot use ghost variable here: "^x)
          end)
        | loop (Ast.ECchar x) = (ECchar x, TChar)
        | loop (Ast.ECint x) = (ECint x, TInt)
        | loop (Ast.ECuint x) = (ECuint x, TUInt)
        | loop Ast.ENull= (ENull, TPtr TVoid)
        | loop (Ast.EDeref(e)) =
         let val (e',t) = loop e
         in
             case t of 
                 TPtr t' => (EDeref(t',e'), t')
               | t => error ("not a pointer "^string_of_ty t)
         end
        | loop (Ast.EAddr e) =
         let val (e',t) = loop e
         in 
             (EAddr e', TPtr t)
         end
        | loop (Ast.EArrow(e,f)) = 
         let val (e',t) = loop e
         in case t of
             TPtr (TStruct s) =>
             (EArrow(e',(TStruct s),f), get_struct_field_type ctx s f)
              | _ => error ("not a pointer to struct "^string_of_ty t)
         end
        | loop (Ast.EIndex(a,i)) =
         let val (a',ta) = loop a
             val (i',ti) = loop i
         in
             case (ta,ti) of
                 (TArray(t',_), TInt) => (EIndex(a',i'), t')
               | t => error ("type error in array access: "^PolyML.makestring t)
         end
        | loop (e0 as (Ast.EAssign(a,b))) =
         let val (a',ta) = loop a
             val (b',tb) = loop b
         in
             if ta <> tb andalso b <> Ast.ENull
             then error ("Assignment with different types "^
                            PolyML.makestring ta^" <> "^PolyML.makestring tb^
                            " in\n"^PolyML.makestring e0)
             else ();
           (EAssign(ta,a',b'), ta)
         end
      (*
         | loop (EPre(prim,e)) res = loop e res
         | loop (EPost(_,e)) res = loop e res
         *)
         | loop (e0 as (Ast.EPrim1(p,e))) = 
         let val (e',t) = loop e
             val (pname,[t1],res) = PrimTab.choose e0 p [t]
         in
             (EPrim1((pname,t1), e'), res)
         end
         | loop (e0 as (Ast.EPrim2(p,a,b))) = 
         let val (a',ta) = loop a
             val (b',tb) = loop b
             val (pname,[t1,t2],res) = PrimTab.choose e0 p [ta,tb]
         in
             (EPrim2((pname,t1,t2),a',b'), res)
         end

         | loop (Ast.EAnd(a,b)) = 
         let val (a',ta) = loop a
             val (b',tb) = loop b
         in
             if ta <> TInt orelse tb <> TInt
             then error ("Boolean AND must be applied to two int")
             else ();
           (EAnd(a',b'), TInt)
         end
         | loop (Ast.EOr(a,b)) =
         let val (a',ta) = loop a
             val (b',tb) = loop b
         in
             if ta <> TInt orelse tb <> TInt
             then error ("Boolean OR must be applied to two int")
             else ();
           (EOr(a',b'),TInt)
         end

      (*
        | loop (EFree(_,e)) res = loop e res
        | loop (EFreeArray(_,a,b)) res = loop a (loop b res)
        | loop (ENew _) res = res
        | loop (ENewArray(_,e)) res = loop e res
        | loop (ECall(f,args)) res = fold loop args (f :: res)
                *)
         | loop e = (raise (NotDone (PolyML.makestring e)))
  in
      loop e
  end

fun tc_sm ctx (vs : var_context) sm =
  let fun loop vs (Ast.SExec e) = SExec (fst (tc_exp ctx vs e))
        | loop vs (Ast.SSeq (a,b)) = SSeq (loop vs a, loop vs b )
        | loop vs (Ast.SBlock (decls,a)) = 
            SBlock(decls, loop (vctx_add_cur (mk_locals decls) vs) a)
        | loop vs (Ast.SIf(e,a,b)) = 
        let val (e',te) = tc_exp ctx vs e
            val _ =
            if te <> TInt
            then error ("if condition must be an int")
            else ();
        in
            SIf (e', loop vs a, loop vs b)
        end
        | loop vs (Ast.SWhile(I,e,a)) =
        let val (e',te) = tc_exp ctx vs e
            val _ =
            if te <> TInt
            then error ("while condition must be an int")
            else ();
        in
            SWhile (I, e',loop vs a)
        end

        | loop vs Ast.SSkip = SSkip
        | loop vs Ast.SBreak = SBreak
        | loop vs Ast.SContinue = SContinue
        | loop vs (Ast.SReturn e) = 
        let val (e',te) = tc_exp ctx vs e
        in
            (** TODO check against return type *)
            SReturn e'
        end
        | loop vs (Ast.SGhost (Ast.GAssign(l,r))) = (SGhost (GAssign(l,r)))
  in
      loop vs sm
  end
end