structure Embed =
struct

type cfun_data = {
   name   : string,
   pre    : term,
   post   : term,
   aux    : term list
}

structure CFun_Data = Generic_Data
   (type T=cfun_data Symtab.table; 
    val empty = Symtab.empty;
    val merge = Symtab.merge (K true);
    val extend = I);
    
fun get_fun_data name ctx =
  case Symtab.lookup (CFun_Data.get (Context.Proof ctx)) name of
      NONE => error ("Function "^name^" not defined")
    | SOME d => d


fun write_list msg ps =
  writeln (Pretty.string_of (Pretty.big_list msg ps))

(* Constructors for common terms *)
val sFalse = Syntax.const @{const_syntax False}
val sTrue = Syntax.const @{const_syntax True}
val sUndefined = Syntax.const @{const_syntax undefined}
fun sCollect P = Syntax.const @{const_syntax Collect} $ P
fun sK t = Abs ("_", dummyT, incr_boundvars 1 t);

val sSkip = Syntax.const @{const_syntax Skip}
fun sBasic b = Syntax.const @{const_syntax Basic} $ b
fun sSeq a b = Syntax.const @{const_syntax Seq} $ a $ b
fun sLet a b = Syntax.const @{const_syntax Let} $ a $ b
val sSkip = Syntax.const @{const_syntax Skip}
fun sGuard f g a = Syntax.const @{const_syntax Guard} 
                     $ HOLogic.mk_string f 
                     $ (sCollect g) $ a
fun sBind a b = Syntax.const @{const_syntax bind} $ a $ b
fun sCond a b c = Syntax.const @{const_syntax Cond} $ a $ b $ c
fun sWhile I e b = Syntax.const @{const_syntax whileAnno} $ e $ I $ sUndefined $ b
fun sCatch a b = Syntax.const @{const_syntax Catch} $ a $ b
val sThrow = Syntax.const @{const_syntax Throw}
fun sBlock init body ret c = Syntax.const @{const_syntax block} $ init $ body $ ret $ c
fun sCall init p ret c = Syntax.const @{const_syntax call} $ init $ p $ ret $ c

(* Type Representations *)

open CSyntax

fun mk_local (x,ty) = (x, (Local,ty))
fun mk_global (x,ty) = (x, (Global,ty))

fun embed_type ctx TInt = @{const cty_int}
  | embed_type ctx TBool = @{const cty_bool}
  | embed_type ctx (TPtr t) = @{const cty_ptr}
  | embed_type ctx TVoid = @{const cty_void}
  | embed_type ctx (TStruct nm) =
    (Consts.read_const (Proof_Context.consts_of ctx) ("cty_struct_"^nm, Position.none)
     handle ERROR _ => error ("embed_type: undefined struct type "^nm))
  | embed_type ctx ty = error ("Unexpected type: "^string_of_ty ty)

fun embed_type_HOL ctx ty =
  case embed_type ctx ty of
      Const (_, Type(_, [ rep, @{typ unit} ])) => rep
    | t => error ("Could not determine HOL embeding for "^
                     PolyML.makestring t)

(* Primitive operators *)

fun casePtrBinVoid pat (ctx,[ (TPtr elem_ty),_ ]) =
  let val sz = if elem_ty = TVoid
               then HOLogic.mk_number @{typ Word} 1
               else (Const(@{const_name ty_size},dummyT) $ embed_type ctx elem_ty)
  in
      map_aterms
      (fn (Free("sz",_)) => sz
         | t => t) pat
  end
  | casePtrBinVoid _ _ = error "casePtrBinVoid expected pointer arg"

val raw_tab = [
  ("prim_plus_int",  K @{term "(a :: Word) + b"} ),
  ("prim_minus_int", K @{term "(a :: Word) - b"} ),
  ("prim_mul_int",   K @{term "(a :: Word) * b"} ),
  ("prim_div_int",   K @{term "(a :: Word) div b"} ),
  ("prim_mod_int",   K @{term "(a :: Word) mod b"} ),
  ("prim_and_int",   K @{term "(a :: Word) AND b"} ),
  ("prim_or_int",    K @{term "(a :: Word) OR b"} ),
  ("prim_tilde_int", K @{term "NOT (a :: Word)"} ),
  ("prim_incr_int",  K @{term "(a :: Word) + 1"} ),
  ("prim_decr_int",  K @{term "(a :: Word) - 1"} ),
  ("prim_eq_int",    K @{term "(a :: Word) = b"} ),
  ("prim_neq_int",   K @{term "(a :: Word) ~= b"} ),
  ("prim_less_int",  K @{term "word_sless (a :: Word) b"} ),
  ("prim_leq_int",   K @{term "word_sle (a :: Word) b"} ),
  ("prim_greater_int", K @{term "word_sless b (a :: Word)"} ),
  ("prim_geq_int",     K @{term "word_sle b (a :: Word)"} ),
  ("prim_plus_uint",   K @{term "(a :: Word) + b"} ),
  ("prim_minus_uint",  K @{term "(a :: Word) - b"} ),
  ("prim_mul_uint",    K @{term "(a :: Word) * b"} ),
  ("prim_div_uint",    K @{term "(a :: Word) div b"} ),
  ("prim_mod_uint",    K @{term "(a :: Word) mod b"} ),
  ("prim_and_uint",    K @{term "(a :: Word) AND b"} ),
  ("prim_or_uint",     K @{term "(a :: Word) OR b"} ),
  ("prim_tilde_uint",  K @{term "NOT (a :: Word)"} ),
  ("prim_incr_uint",   K @{term "(a :: Word) + 1"} ),
  ("prim_decr_uint",   K @{term "(a :: Word) - 1"} ),
  ("prim_eq_uint",     K @{term "(a :: Word) = b"} ),
  ("prim_neq_uint",    K @{term "(a :: Word) ~= b"} ),
  ("prim_less_uint",   K @{term "(a :: Word) < b"} ),
  ("prim_leq_uint",    K @{term "(a :: Word) <= b"} ),
  ("prim_greater_uint", K @{term "(a :: Word) > b"} ),
  ("prim_geq_uint",     K @{term "(a :: Word) >= b"} ),
 
  ("prim_plus_char",    K @{term "(a :: Word) + b"}  ),
  ("prim_minus_char",   K @{term "(a :: Word) - b"}  ),
  ("prim_mul_char",     K @{term "(a :: Word) * b"}  ),
  ("prim_div_char",     K @{term "(a :: Word) div b"}  ),
  ("prim_mod_char",     K @{term "(a :: Word) mod b"}  ),
  ("prim_incr_char",    K @{term "(a :: Word) + 1"} ),
  ("prim_decr_char",    K @{term "(a :: Word) - 1"} ),
  ("prim_tilde_char",   K @{term "NOT (a :: Word)"} ),
  ("prim_eq_char",      K @{term "(a :: Word) = b"} ),
  ("prim_neq_char",     K @{term "(a :: Word) ~= b"} ),
  ("prim_less_char",    K @{term "(a :: Word) < b"} ),
  ("prim_leq_char",     K @{term "(a :: Word) <= b"} ),
  ("prim_greater_char", K @{term "(a :: Word) > b"} ),
  ("prim_geq_char",     K @{term "(a :: Word) >= b"} ),

  ("prim_eq_bool",      K @{term "(a :: bool) = b"} ),
  ("prim_neq_bool",     K @{term "(a :: bool) ~= b"} ),
  ("prim_not",          K @{term "Not (a :: bool)"} ),

  ("prim_plus_ptr", casePtrBinVoid @{term "(a :: Word) + sz * b"}),

(*
  ("prim_plus_ptr_unsigned", K @{term False} ),
  ("prim_minus_ptr",         K @{term False} ),
  ("prim_incr_ptr",         K @{term False} ),
  ("prim_decr_ptr",         K @{term False} ),
*)
  ("prim_eq_ptr",           K @{term "(a :: Word) = b"} ),
  ("prim_neq_ptr",          K @{term "(a :: Word) ~= b"} ),
  ("prim_less_ptr",         K @{term "(a :: Word) < b"} ),
  ("prim_leq_ptr",          K @{term "(a :: Word) <= b"} ),
  ("prim_greater_ptr",      K @{term "(a :: Word) > b"} ),
  ("prim_geq_ptr",          K @{term "(a :: Word) >= b"} ),
                
 (*
 ("prim_tilde_char",
  { args = [ "TChar" ], res = "TChar",
    embed = fn [a] =>  to_Char (NOT (to_char a)) }),
 ("conv_int_char",
  { args = [ "TInt" ], res = "TChar",
    embed = fn [a] =>  to_Char (scast (to_int a)) }),
 ("conv_char_int",
  { args = [ "TChar" ], res = "TInt",
    embed = fn [a] =>  to_Int (scast (to_char a)) }),
 ("conv_uint_char",
  { args = [ "TUInt" ], res = "TChar",
    embed = fn [a] =>  to_Char (scast (to_uint a)) }),
 ("conv_char_uint",
  { args = [ "TChar" ], res = "TUInt",
    embed = fn [a] =>  to_UInt (scast (to_char a)) }),
*)

 ("cast_int_ptr",  K @{term "(a :: Word)"} ),
 ("cast_ptr_int",  K @{term "(a :: Word)"} ),
 ("cast_int_uint", K @{term "(a :: Word)"} ),
 ("cast_uint_int", K @{term "(a :: Word)"} )
]

val primop_tab =
  let fun prep (nm, e) =
      (* TODO: pull out preparation of term to speed up *)
      (nm,
       fn (ctx,arg_tys) => fn args =>
         let val e = e (ctx,arg_tys)
            val frees =
                (* make sure that parameters are independent of the order in the term,
                   which may be surprising due to abbreviations *)
                sort (string_ord o pairself fst) 
                         (Term.add_frees e [])
          val closed = fold_rev absfree frees e
         in Term.list_comb (closed,args) end)
  in
      Symtab.make (map prep raw_tab)
  end

fun embed_primop ctx primop arg_tys args =
  case Symtab.lookup primop_tab primop of
      NONE => error ("embedding of primitive operator "^primop^" not done")
    | SOME f => Envir.beta_norm (f (ctx,arg_tys) args)

val heap_var = "heap"  
val return_name = "ret_"  
val thrown_name = "thrown_"

fun is_global_var vctx name =
  not (name = thrown_name) andalso
  (name = heap_var orelse
   member (op=) [ Global, GhostGlobal ] (fst (vctx_get_cur vctx name)))

fun var_name vctx v =
  case vctx_fun_name vctx of
      NONE => Hoare.varname v
    | SOME fun_name =>
      if is_global_var vctx v
      then Hoare.varname v
      else Hoare.varname (suffix ("_'"^fun_name) v)

fun is_state_comp ctx vctx name =
  case try (var_name vctx) name of
    NONE => false
    | SOME nm =>
      is_some (StateSpace.get_comp (Context.Proof ctx) nm)

val global_register =
   (Syntax.const @{const_syntax StateSpace.globals},
    Syntax.const @{const_syntax StateSpace.globals_update})
    
val local_register =
   (Syntax.const @{const_syntax StateSpace.locals},
    Syntax.const @{const_syntax StateSpace.locals_update})

fun var_register vctx v =
  if is_global_var vctx v
  then global_register
  else local_register

fun read_var ctx vctx v =
  let val (sel,_) = var_register vctx v
  in
      Abs("s",dummyT,
          StateSpace.gen_lookup_tr ctx (sel $ Bound 0) (var_name vctx v))
  end

fun set_var ctx vctx v t =
  let val (sel,upd) = var_register vctx v
  in
      Abs("s",dummyT,
          upd $ 
             (sK (StateSpace.gen_update_tr false ctx (var_name vctx v) 
                    (Term.incr_boundvars 1 t) (sel $ (Bound 0)))) $ Bound 0)
  end
  
fun replace_globals ctx src =
  let val (sel,upd) = global_register
  in Abs("s",dummyT,
         upd $ (sK (sel $ Term.incr_boundvars 1 src)) $ Bound 0)
  end  
  
fun heap_put ctx vctx t p v =
   Abs("s",dummyT,
     set_var ctx vctx heap_var
       (Syntax.const @{const_syntax heap_put} 
           $ embed_type ctx t 
           $ incr_boundvars 1 p 
           $ incr_boundvars 1 v 
           $ (read_var ctx vctx heap_var $ Bound 0))
          $ Bound 0)
         

fun heap_get ctx vctx t p =
  Abs("s",dummyT,
   Syntax.const @{const_syntax heap_get} $ embed_type ctx t 
         $ incr_boundvars 1 p
         $ (read_var ctx vctx heap_var $ Bound 0))
  
fun heap_acc_cond ctx vctx t p =
  Abs("s",dummyT,
      Syntax.const @{const_syntax heap_acc_cond} $ embed_type ctx t 
         $ incr_boundvars 1 p
         $ (read_var ctx vctx heap_var $ Bound 0))

fun dest_TPtr (TPtr t) = t
  | dest_TPtr t = error ("dest_TPtr fails "^PolyML.makestring t)

fun dest_TStruct (TStruct s) = s
  | dest_TStruct t = error ("dest_TStruct fails "^PolyML.makestring t)

fun field_addr ctx p s f =
  let val intern = Consts.intern (Proof_Context.consts_of ctx)
  in
      (Const(@{const_name plus},dummyT) 
          $ p $ (Const(intern ("fo_"^dest_TStruct s^"_"^f),dummyT)))
  end
  
fun dest_name_safe t =
  SOME (dest_name t)
  handle ERROR _ => NONE

fun s_throw ctx vctx which =
  sSeq (sBasic (set_var ctx vctx thrown_name which)) sThrow

fun s_catch ctx vctx which a b =
  sCatch a
    (sCond 
        (sCollect 
            (Abs("s",dummyT,
                 HOLogic.eq_const @{typ throw_case} 
                    $ (read_var ctx vctx thrown_name $ Bound 0)
                    $ which)))
        b
        sThrow)

fun lookup_var ctx vctx vs x =
  AList.lookup (op=) vs x
    

(*
  Some variable names are mapped to internal names to
  avoid clashes.
*)
fun fix_var_name "ret" = return_name
  | fix_var_name "heap" = heap_var
  | fix_var_name v = v
  
val pre_state = Free("pre_state_",dummyT)
      
(**
  Assertions are HOL terms of type bool which are allowed to
  inspect the state bound at some outer level.  
*)
fun embed_HOL ctx vctx vs state_depth t =
  let fun resolve_const t =
        let val nm = case t of
                         Const(c,_) => c
                       | Free(f,_) => f
        in
            ProofContext.read_const_proper ctx false nm
            handle ERROR _ => t
        end
      fun var d t is_pre_state =
        if dest_name t = ("\\"^"<sigma>")
        then Bound d
        else
        let val var_name = fix_var_name (dest_name t)
        in
            case AList.lookup (op=) vs var_name of
                NONE => (* check for ghost variables *)
                if is_state_comp ctx vctx var_name
                then read_var ctx vctx var_name $ Bound d
                else resolve_const t
              | SOME _ => 
                read_var ctx vctx var_name
                $ (if is_pre_state
                   then error "References to pre_state not finished" 
                         (* pre_state *)
                   else  Bound d)
        end
      
      fun loop d (t as Const _) = var d t false   
        | loop d (t as Free _) = var d t false
        | loop d (Const(@{syntax_const "@prestate"},_) $ (t as Const _)) = 
          var d t true
        | loop d (Const(@{syntax_const "@prestate"},_) $ (t as Free _)) =
          var d t true
        | loop d (Const(@{syntax_const "_faccess"},_) $ t $ f) =
           (case dest_name_safe t of
                NONE => (*TODO*) error("Sorry: struct access done only for pointer variables")
              | SOME v =>
                (case lookup_var ctx vctx vs v of
                    NONE => error ("unknown variable in field access: "^v)
                   | SOME(_, TPtr (ty as (TStruct s))) =>
                     let val f = dest_name f
                         val ft = get_struct_field_type ctx s f
                         val addr = field_addr ctx (read_var ctx vctx v $ Bound d) ty f
                     in 
                         heap_get ctx vctx ft addr $ Bound d
                     end
                   | SOME _ => error ("variable "^v^" is not pointer to struct")))
        | loop d (a $ b) = loop d a $ loop d b
        | loop d (Abs(x,T,t')) = Abs(x,T, loop (d+1) t')
        | loop d t = t
  in
      loop state_depth t
  end

fun embed_HOL_assn ctx vctx vs fix_pre_state P =
  let val P' = embed_HOL ctx vctx vs 0 P
      val P'' = sCollect (Abs("s", dummyT, P'))
  in
      if false andalso fix_pre_state
      then Const(@{const_name Lattices.inf},dummyT) 
          $ P''
          $ HOLogic.mk_set dummyT [pre_state]
      else P''
  end

(**
  The embedding of an expression a statement with a "hole", where
  the result of the expression is accessible as (Bound 0) and which
  must be filled by a statement. That statement is passed along as
  a continuation "k", both to the overall function and the loop.

  The VCG is slow if bind nodes are introduced too directly.
  We therefore use an intelligent bind function that understands
  the Simpl language and inspects the usages of the bound parameter.
  If they occur only in places where the state has not yet changed,
  the function inlines the references.  
*)

fun abs_k k = Abs("_",dummyT, k)

(*
 inc is the number of outer binders introduced besides this bind.
*)
fun bind ctx inc v k = 
  let exception StateChange
      open TraceAux
      (* b is the binding index, s the optional state index *)
      fun tr msg b t =
        tracing ("bind fail/"^msg^"/"^string_of_int b^
                    ":"^Syntax.string_of_term ctx t)
      
      fun check_occ msg b t =
        if member (op=) (loose_bnos t) b
        then (tr ("occ/"^msg) b t; raise StateChange)
        else ()
      
      fun repl b s (Bound i) =
         if i > b
         then Bound (i - 1)
         else if b = i 
         then (if s >= 0
               then betapply (incr_boundvars (s+1) v, Bound s)
               else (tr "repl" b (Bound i); raise StateChange))
         else Bound i
        | repl b s (Abs(x,T,t)) = Abs(x,T,repl (b+1) (s+1) t)
        | repl b s (u $ v) = repl b s u $ repl b s v
        | repl _ _ t = t
      
      fun inline b ((u as Abs _) $ v) =
         inline b (betapply (u,v))
        | inline b (Const(@{const_syntax Basic},_) $ f) =
          (Const(@{const_syntax Basic},dummyT) $ repl b ~1 f)
        | inline b (Const(@{const_syntax Seq},_) $ u $ v) =
          (check_occ "seq" b v;
           (Const(@{const_syntax Seq},dummyT) $ inline b u $ v))
        | inline b (Const(@{const_syntax Cond},_) $ e $ u $ v) =
          (Const(@{const_syntax Cond},dummyT) $ repl b ~1 e $ inline b u $ inline b v)
        | inline b (Const(@{const_syntax bind},_) $ e $ Abs(x,T,u)) =
          (Const(@{const_syntax bind},dummyT)
              $ repl b ~1 e 
              $ Abs(x,T, inline (b+1) u))
        | inline b (Const(@{const_syntax Guard},_) $ f $ g $ u) =
          (Const(@{const_syntax Guard},dummyT) $ f $ repl b ~1 g $ inline b u)
        | inline b (Const(@{const_syntax call},_) $ init $ p $ ret $ k) =
          (check_occ "call/p" b p;
           check_occ "call/ret" b ret;
           check_occ "call/k" b k;
           (Const(@{const_syntax call},dummyT) $ inline b init $ p $ ret $ k))
        | inline b t = (check_occ "other" b t; t)
  in
      inline 0 (incr_bv (inc,1,k))
      handle StateChange => sBind v (abs_k (incr_bv (inc+1,1,k)))
  end
  
(*
  Conservative check whether the short-circuit semantics 
  is redundant,  i.e. the rhs argument can be evaluated 
  regardless of the outcome of the lhs.
*)
val elim_short_circuit_redundant = Unsynchronized.ref true

fun short_circuit_redundant e =
    let fun loop (EVar _) = true
        | loop (ECbool _) = true
        | loop (ECchar _) = true
        | loop (ECint _) = true
        | loop (ECuint _) = true
        | loop (ENull _) = true
        | loop (EDeref(_,_)) = false
        | loop (EAddr e) = loop e
        | loop (EAcc(e,_,_)) = false
        | loop (EAssign(_,a,b)) = false
        | loop (EPre(_,EVar _)) = true
        | loop (EPre(_,e)) = false
        | loop (EPost(_,EVar _)) = true
        | loop (EPost(_,e)) = false
        | loop (EPrim1(_,e)) = loop e
        | loop (EPrim2(_,a,b)) = loop a andalso loop b
        | loop (EFree(_,e)) = false
        | loop (EFreeArray(_,a,b)) = false
        | loop (ENew _) = true
        | loop (ENewArray(_,e)) = false
        | loop (EAnd(a,b)) = loop a andalso loop b
        | loop (EOr(a,b)) = loop a andalso loop b
        | loop (ECall(f,args)) = forall loop args
  in
      !elim_short_circuit_redundant andalso loop e
  end

fun embed_exp ctx vctx e k =
  let val bind = bind ctx
      val bind0 = bind 0
      val bind1 = bind 1
      fun loop (EVar v) k = bind0 (read_var ctx vctx v) k
        | loop (ECint i) k = 
          (* can use beta-reduction here *)
          betapply (abs_k k, HOLogic.mk_number @{typ Word} i)
        | loop (ECbool b) k = 
          (* can use beta-reduction here *)                            
          betapply (abs_k k,
                    if b
                    then HOLogic.true_const 
                    else HOLogic.false_const)
        | loop (ENull _) k =
          betapply (abs_k k, @{term NULL})
        | loop (EDeref (t,p)) k = 
            loop p (sGuard "heap_acc" 
                         (heap_acc_cond ctx vctx t (Bound 0))
                         (bind1 (heap_get ctx vctx t (Bound 0)) k))
        | loop (EAcc (EDeref(ty,p), ty', f)) k =
          let val t = get_struct_field_type ctx (dest_TStruct ty') f
              val addr = field_addr ctx (Bound 0) ty' f
          in
              loop p
                (sGuard "heap_acc"
                    (heap_acc_cond ctx vctx t addr)
                    (bind1 (heap_get ctx vctx t addr) k))
          end
        | loop (EAssign(t,(EVar v), e)) k =
          loop e (sSeq (sBasic (set_var ctx vctx v (Bound 0))) 
                     (incr_bv (1,1,k)))
        | loop (EAssign (t, EAcc (EDeref (t',p), s, f), e)) k =
          let val addr = field_addr ctx (Bound 1) s f
          in
              loop p (loop e
                   (sSeq (sGuard "heap_acc"
                             (heap_acc_cond ctx vctx t addr)
                             (sBasic (heap_put ctx vctx t addr (Bound 0)))) 
                       (incr_bv (2,0,k))))
          end
        | loop (EAssign(t,(EDeref (t',p)), e)) k =
          (if t <> t'
           then error "*p = e with incompatible types"
           else ();
           loop p (loop e 
              (sSeq (sGuard "heap_acc"
                        (heap_acc_cond ctx vctx t (Bound 1))
                        (sBasic (heap_put ctx vctx t (Bound 1) (Bound 0)))) 
                  (incr_bv (2,0,k)))))
        | loop (EPrim1((primop,ta),a)) k =
          (* use beta-reduction instead of bind for functional update *)
          loop a (betapply (abs_k (incr_bv (1,1,k)), embed_primop ctx primop [ta] [ Bound 0 ]))
        | loop (EPrim2((primop,ta,tb),a,b)) k =
          (* use beta-reduction instead of bind for functional update *)
          loop a (loop b (betapply (abs_k (incr_bv (2,1,k)), 
                                    embed_primop ctx primop [ta,tb] [ Bound 1, Bound 0])))
        | loop (EPost((primop,ta), EVar v)) k =
          bind0 (read_var ctx vctx v) 
               (sSeq (sBasic (set_var ctx vctx v 
                                 (embed_primop ctx primop [ta] [ Bound 0 ]))) k)
        | loop (EPre ((primop,ta), EVar v)) k =
          bind0 (read_var ctx vctx v) 
               (sSeq (sBasic (set_var ctx vctx v (embed_primop ctx primop [ta] [ Bound 0 ]))) 
                   (bind0 (read_var ctx vctx v) k))
        | loop (EAnd (a,b)) k =
          if short_circuit_redundant b
          then (loop a (loop b (betapply (abs_k (incr_bv (2,1,k)),
                                          (@{const HOL.conj} $ Bound 1 $ Bound 0)))))
          else (loop a (sCond (sCollect (sK (Bound 0)))
                           (loop b k)
                           (betapply (abs_k k, sFalse))))
        | loop (EOr (a,b)) k =
          if short_circuit_redundant b
          then (loop a (loop b (betapply (abs_k (incr_bv (2,1,k)),
                                          (@{const HOL.disj} $ Bound 1 $ Bound 0)))))
          else (loop a (sCond (sCollect (sK (Bound 0)))
                           (betapply (abs_k k, sTrue))
                           (loop b k)))
        | loop (ECall(f,args)) k =
          let val {params,rty,...} = CSyntax.get_fun_info ctx f
              val vctx' = vctx_mk (SOME f) (map mk_local 
                                               (params @ [(return_name,rty)]))
              val init =
                 Abs("s",dummyT,
                     fold (curry op $)
                        (map_index 
                            (fn (i,(p,_)) =>
                               set_var ctx vctx' p (Bound (i + 1)))
                            params)
                        (Bound 0))
              val ret =
                 Abs("s",dummyT,
                     Abs("t",dummyT,
                         set_var ctx vctx' return_name 
                            (read_var ctx vctx' return_name $ (Bound 0))
                            $ (replace_globals ctx (Bound 0) $ (Bound 1))))
          in
              fold loop args
              (sCall init (Free(suffix Hoare.proc_deco f,dummyT)) ret
                (Abs("_",dummyT, 
                     Abs("s",dummyT,
                         bind (length args+2)
                            (read_var ctx vctx' return_name) 
                               k))))
          end
        | loop e k = error("Embedding/exp not done: "^CSyntax.string_of_exp e)
  in
      loop e k
  end

fun contract (Const(@{const_syntax Seq},_) $ a $ Const(@{const_syntax Skip},_)) = 
    contract a
  | contract (Const(@{const_syntax Seq},_) $ Const(@{const_syntax Skip},_) $ b) =
    contract b
  | contract (Const(@{const_syntax Cond},_) 
                 $ (Const(@{const_syntax Collect},_) 
                       $ (Abs(_,_, (Const(@{const_syntax True},_)))))
                 $ u $ v) = contract u
  | contract (Const(@{const_syntax Cond},_) 
                 $ (Const(@{const_syntax Collect},_) 
                       $ (Abs(_,_, (Const(@{const_syntax False},_)))))
                 $ u $ v) = contract v
  | contract (a $ b) = contract a $ contract b
  | contract (Abs(x,T,t)) = Abs(x,T,contract t)
  | contract t = t
  
fun embed_sm ctx vctx syn =
  let fun loop vctx SSkip = sSkip
        | loop vctx (SSeq(a,b)) = 
          sSeq (loop vctx a) (loop vctx b)
        | loop vctx (SExec e) = 
          embed_exp ctx vctx e sSkip
        | loop vctx (SIf (e,a,b)) =
          embed_exp ctx vctx e
           (sCond (sCollect (sK (Bound 0))) (loop vctx a) (loop vctx b))
        | loop vctx (SWhile((vctx',I),e,b)) =
          (s_catch ctx vctx @{const ThrowBreak}
              (sWhile (embed_HOL_assn ctx vctx vctx' false I) (sCollect (sK sTrue))
                (embed_exp ctx vctx e
                  (sCond (sCollect (sK (Bound 0)))
                      (loop vctx b)
                      (s_throw ctx vctx @{const ThrowBreak}))))
              sSkip)
        | loop vctx (SBlock(decls,body)) =
          let val vctx = vctx_add_cur (map mk_local decls) vctx
              fun vars f = fold (fn d => fn r => f d $ r) decls (Bound 0)
              fun havoc_var (x,t) =
                 set_var ctx vctx x sUndefined
              val init = 
                 Abs("s",dummyT, vars havoc_var)
              fun rest_var (x,t) =
                set_var ctx vctx x (read_var ctx vctx x $ (Bound 1))
              val restore =
                 Abs("s",dummyT, Abs("t",dummyT, vars rest_var))                                      
          in
              sBlock init (loop vctx body) restore 
                 (sK (sK sSkip))
          end
        | loop vctx (SGhost(vs,GAssign(x,t))) =
            sBasic (Abs("s",dummyT, 
                        set_var ctx vctx x (embed_HOL ctx vctx vs 0 t) $ Bound 0))
        | loop (vctx as VarContext {fun_name,...}) (SReturn e) =
          (case fun_name of
               NONE => error "Cannot embed return with function context"
             | SOME name =>
               embed_exp ctx vctx e 
                  (sSeq (sBasic (set_var ctx vctx return_name (Bound 0)))
                      (s_throw ctx vctx @{const ThrowReturn})))
        | loop vctx sm =
          error ("Embedding/sm not done: "^CSyntax.string_of_sm sm)
  in
      contract (loop vctx syn)
  end

type globals_info = {
  current : string,
  spaces  : (string * (var_kind * ty)) list Symtab.table
}                      

structure Globals = Generic_Data(
  type T = globals_info 
  val empty = {current="", spaces = Symtab.empty}
  val extend = I
  fun merge ({current,spaces},{spaces=spaces',...}) = 
    {current=current, spaces = Symtab.merge (K true) (spaces,spaces')}
)

fun get_globals ctx name =
  let val {spaces,...} = Globals.get (Context.Proof ctx)
  in case Symtab.lookup spaces name of
         NONE => error ("undefined global context "^name)
       | SOME vs => vs
  end

fun get_current_globals ctx =
  let val {spaces,current} = Globals.get (Context.Proof ctx)
  in case Symtab.lookup spaces current of
         NONE => []
       | SOME vs => vs
  end


val base_globals_name = "base_state"

val setup_base_globals =
 Context.theory_map
   (Globals.put ({current = base_globals_name,
                 spaces = Symtab.make
                            [(base_globals_name, [])] }))

fun embed_statement ctx [ sm ] =
  let val vs = get_current_globals ctx
               |> map (fn (x,(_,ty)) => (x,(Local,ty)))
      val vs' = map CSyntax.decl_ty_as_term vs
      val syn = CSyntax.tr_sm ctx vs' NONE sm
      val _ = tracing ("Syntax tree:\n"^CSyntax.string_of_sm syn)
      val res = embed_sm ctx (vctx_mk NONE vs) syn
      val _ = TraceAux.trace_list "embedding result" [ Syntax.pretty_term ctx res ] 
  in
      res
  end

fun heap_assn_tr ctx [ P ] =
  P $ (read_var ctx (vctx_mk NONE []) heap_var)

fun add_state_space name parents fun_name_opt decls extra_comps thy =
  let open TraceAux
      val statetypeN = "StateType";
      val ctx = Proof_Context.init_global thy
      val vctx0 = vctx_mk fun_name_opt decls
      fun prep (x,(_,ty)) =
        (var_name vctx0 x, embed_type_HOL ctx ty)
      
      val _ =
         trace_list ("in "^name^" declaring components")
            (map (fn (x,T) => Pretty.block [ Pretty.str x,
                                             Pretty.str " :: ",
                                             Syntax.pretty_typ ctx T ])
                (map prep decls @ extra_comps))
  in
      StateSpace.define_statespace_i
          (SOME statetypeN) (* follow hoare_syntax / hoare? *)
          [] (* args *) name (* name *) 
          (map (fn p => ([],p,[])) parents)
          (map prep decls @ extra_comps) thy (* the components *)
  end
  
(* WORKAROUND FOR LOST INFO IN NEW THEORY:
  Sign.full_name gets used but the namespace is never defined *)
fun add_base_state_spaces thy =
  let val heap = Hoare.varname heap_var
      val thrown = Hoare.varname thrown_name 
  in
      thy |> add_state_space base_globals_name [] NONE [] 
               [(heap,@{typ heap}),
                (thrown, @{typ throw_case})]
      |> add_state_space (base_globals_name^"_globals") [] NONE [] 
               [(heap,@{typ heap}), 
                 (thrown, @{typ throw_case})]
  end handle ERROR _ => thy
      
fun declare_global gctx mk_Simpl_globals ((name, decls : string option), ghost) =
  let val ctx = Context.proof_of gctx
      val ghost = map (fn(x,T) => (Hoare.varname x, Syntax.read_typ ctx T)) ghost      
      val ghost_vars = map (fn (x,_) => (unsuffix Hoare.deco x, (GhostGlobal, TVoid))) ghost
      val decls = Option.map stripYXML decls
      val thy = Proof_Context.theory_of ctx
      val syn = Sign.syn_of thy
      val vars = 
         case decls of
           NONE => []
           | SOME decls =>
             let val decls = Syntax.parse_term ctx ("VARDECLS ("^decls^")")
                 val _ = TraceAux.trace_list "Declarations" [ Syntax.pretty_term ctx decls ]
                 val vars = tr_vardecls ctx Global decls
                    |> map decl_term_as_ty
                 val _ = write_big_list "vars are"
                           (map (fn (f,(_,t)) => Pretty.block [ Pretty.str f,
                                                                Pretty.str " : ",
                                                                Syntax.pretty_term ctx 
                                                                (ty_as_term t) ])
                               vars)
             in
                 vars
             end
      val suff = if mk_Simpl_globals
                 then suffix "_globals"
                 else I
  in
      gctx 
      |> Context.map_theory add_base_state_spaces
      |> Globals.map (fn {spaces,current} =>
                         {current=name,
                           spaces=Symtab.insert (op=) (name,vars @ ghost_vars) spaces})                           
      |> Context.map_theory (add_state_space 
                                (suff name)
                                (* [suff base_globals_name] *)
                                [ base_globals_name^"_globals" ]
                                NONE vars ghost)
  end

val decl_inputP =
   Parse.name -- Scan.option Parse.string --
      Scan.repeat (Parse.name --| Parse.$$$ "::" -- Parse.!!! Parse.typ)
      
val _ =
  Outer_Syntax.command
      "declare_global" "declare variables as global" Keyword.thy_decl
      (decl_inputP >> 
          (fn (name,globals) =>
		     Toplevel.generic_theory
			 (fn ctx => declare_global ctx true (name,globals))))

val _ =
  Outer_Syntax.command
      "declare_global_test" "declare variables as locals for tests" Keyword.thy_decl
      (decl_inputP >>
          (fn (name,globals) =>
		     Toplevel.generic_theory
			 (fn ctx => declare_global ctx false (name,globals))))

val setup =
    Sign.add_advanced_trfuns 
	([],
	 [ ("@statement", embed_statement),
           ("@heap", heap_assn_tr) ],
         [],
         []) #>
    setup_base_globals

end