structure Heaps =
struct

open TraceAux

val (heap_trace, heap_trace_setup) =
   Attrib.config_bool (Binding.name "trace_heaps") (fn _ => false);

val (heap_fail_trace, heap_fail_trace_setup) =
   Attrib.config_bool (Binding.name "trace_heaps_fail") (fn _ => false);

fun tr ctx f=
  if Config.get ctx heap_trace
  then (ignore (f ()); ())
  else ()

fun tr_fail ctx f =
  if Config.get ctx heap_fail_trace andalso
      not (Config.get ctx heap_trace)
  then (ignore (f()); ())
  else ()
  
fun fail_list msg lst =
   raise Fail (Pretty.string_of (Pretty.big_list msg lst))

(* General tools *)

fun map_remainder f xs =
  let fun loop xs [] = []
        | loop xs (y :: ys) =
          f y (xs @ ys) :: loop (xs @ [y]) ys
  in
      loop [] xs
  end

fun fold_map_remainder f xs z =
  let fun loop xs [] z = ([],z)
        | loop xs (y :: ys) z =
        let val (y',z') = f y (xs @ ys) z
            val (r,z'') = loop (xs @ [y]) ys z'
        in
            (y' :: r, z'')
        end
  in
      loop [] xs z
  end

fun conj_elims thms =
    let fun loop [] res = res
	  | loop (thm :: thms) res =
	    (case prop_of thm of
		 @{term "Trueprop"} $ (@{const "HOL.conj"} $ _ $ _) =>
		 let val (a,b) = HOLogic.conj_elim thm
		 in
		     loop (a :: b :: thms) res
		 end
	       | _ => loop thms (thm :: res))
    in
	loop thms []
    end

(*
 Generic tools for heap premises
*)
fun is_heap_prop params t =
  case  t of
      @{const Trueprop} $ (_ $ Bound h) =>
      (let val (nm,T) = nth params h
       in
           T = @{typ heap}
       end)
    | @{const Trueprop} $ (_ $ Free(h as (_,@{typ heap}))) =>
      true
    | _ => false

(*
  Cleanup the heaps in the premises such that those
  inserted later have precedence. This matches the
  structured reasoning introduced by Subgoal etc.
*)
fun thin_indexed_tac [] i thm = all_tac thm
  | thin_indexed_tac xs i thm =
    let val nhyps = length (Logic.prems_of_goal (prop_of thm) i)
      fun rot k = PRIMITIVE (Thm.rotate_rule k i)
      fun loop [] rem_prems = rot rem_prems
        | loop (x :: xs') rem_prems =
            rot (x - 1) THEN
            etac @{thm thin_rl} i THEN
            loop (map (fn i => i - x) xs') (rem_prems - x)
  in
      loop xs nhyps thm
  end


fun cleanup_heaps ctx i =
  COND (fn thm => nprems_of thm >= i)
  (SUBGOAL (fn (t,i) =>
    let val _ = tr ctx (fn () => trace_list "cleanup_heaps" [ pretty_term ctx t ])
        val params = Logic.strip_params t
        val hyps   = Logic.strip_assums_hyp t
        val is_heap = is_heap_prop (rev params)
        val heaps   = map_filter I
                        (map_index (fn (i,h) =>
                                      if is_heap h
                                      then SOME (i + 1) else NONE) hyps)
    in
        if length heaps <= 1
        then all_tac
        else thin_indexed_tac (take (length heaps - 1) heaps) i
    end) i)
    all_tac

fun dest_heap_prem thm =
  case prop_of thm of
      (@{const Trueprop} $ (_ $ (Free(_,@{typ heap})))) =>
      let val (P,h) = Thm.dest_comb (Thm.dest_arg (cprop_of thm))
      in
          SOME(P,h,thm)
      end
    | _ => NONE

fun heap_prem_spatial thm =
  #1 (the (dest_heap_prem thm))
  handle Option.Option =>
         error ("Not a heap prem "^string_of_thm
                  (Proof_Context.init_global (theory_of_thm thm)) thm)

fun the_heap_prem prems =
  case distinct (Thm.eq_thm o pairself #3) (map_filter dest_heap_prem prems) of
      [ p ] => SOME p
    | [] => NONE
    | _ => NONE

fun with_heap_prem prems f =
  case the_heap_prem prems of
      NONE => error "expected single heap premise"
    | SOME (_,_,hthm) => f hthm

fun heap_tac ctx tac cleanup =
  Subgoal.FOCUS_PREMS
  (fn (focus as ({context=ctx',prems,...} : Subgoal.focus)) =>
      case the_heap_prem prems of
          SOME (P,h,thm) => tac ctx' focus thm
        | NONE => no_tac) ctx THEN'
  (fn i => COND (K cleanup) (cleanup_heaps ctx i) all_tac)

fun heap_meth f cleanup ctx =
  METHOD (K (heap_tac ctx f cleanup 1))

(*
   computing disjointness of pointers
*)
structure Addr_Thms = Named_Thms
(val name = @{binding heap_addrs}; 
 val description = "forward computation of heap addrs");
structure Zero_Off_Thms = Named_Thms
(val name = @{binding zero_offset};
 val description="simplification rule deriving zero offsets in heap_addrs");

fun compute_disjoint_list ctx hthm =
  let val tr_old = tr
      fun tr f = tr_old ctx f
      val pretty_cterm = pretty_cterm ctx
      val pretty_thm = pretty_thm ctx
      val P = heap_prem_spatial hthm
      val () = tr (fn () => trace_list "fwd_disjoint" [pretty_cterm P])
      val atomic = Addr_Thms.get ctx
      val disj_star = Thm.permute_prems 0 1 @{thm disjoint_list_star}
      val thy = theory_of_thm hthm
      fun loop P =
        (tr (fn () => trace_list "loop" [ pretty_cterm P ]);
        case term_of P of
            (@{const star} $ _ $ _) =>
            let val (P',Q') = Thm.dest_binop P
                val disj_inst  = instantiate' [] [SOME P',NONE, SOME Q'] disj_star
                val left = Drule.lift_all (Thm.cprem_of disj_inst 1) (loop P')
                val right = Drule.lift_all (Thm.cprem_of disj_inst 2) (loop Q')
                val _ = tr (fn () => trace_list "combine"
                          [ pretty_thm disj_inst,
                            pretty_thm left,
                            pretty_thm right ])
                val res = right INCR_COMP (left COMP_INCR disj_inst)
                val _ = tr (fn () => trace_thm ctx "res" res)
            in
                res
            end
          | t =>
            (let fun matches thm =
                 let val thy' = Theory.merge (thy, theory_of_thm thm)
                     val t' = term_of (Thm.dest_fun (Thm.dest_arg (Thm.cprem_of thm 1)))
                     val p = Syntax.pretty_term ctx
                 in
                     Pattern.matches thy' (t', t)
                 end handle (CTERM _) => false
                   | (THM _) => false
             in
                 case find_first matches atomic of
                     SOME thm =>
                     (tr (fn () => trace_thm ctx "atomic result" thm);
                      thm)
                   | NONE => @{thm disjoint_list_empty}
             end))
      val loop_res = hthm COMP (loop P)
      val _ = tr (fn () => trace_thm ctx "loop-res" loop_res)
  in
      loop_res
  end

fun thm_proves cprop =
  curry (op aconvc) cprop o cprop_of

fun destruct_disjoint_list ctx thm =
  let val rw = Simplifier.rewrite (simpset_of ctx addsimps [ @{thm Int_Un_distrib}, @{thm Int_Un_distrib2} ])
      fun simp thm =
        let val eq = rw (cprop_of thm)
        in
            if Thm.is_reflexive eq
            then NONE
            else SOME (Thm.equal_elim eq thm)
        end
      fun prep thm =
        filter (not o thm_proves @{cprop True}) (conj_elims [thm])
      fun loop [] res = res
        | loop (thm :: thms) res =      
          (case get_first (try (fn d => thm RS d))
                          @{thms disjoint_list_destructD} of
               NONE => 
               (case simp thm of
                    NONE => loop thms (thm :: res)
                  | SOME thm' => loop (prep thm' @ thms) res)
             | SOME thm' =>
               loop (prep thm' @ thms) res)
      val res1 = loop (prep thm) []

      val zero_off_ss =
         simpset_of ctx addsimps (Zero_Off_Thms.get ctx)
      fun fwd_zero_off thm =
        let val thm' = Simplifier.simplify zero_off_ss thm
        in if prop_of thm aconv prop_of thm'
           then NONE
           else SOME thm'
        end
  in
      map_filter fwd_zero_off (res1) @ res1
  end

fun reduce_to_False info =
  case filter (thm_proves @{cprop False}) info of
      [] => info
    | thm :: _ => [thm]

fun fwd_disjointness raw ctx _ h =
  cut_facts_tac
    (if raw
     then [compute_disjoint_list ctx h]
     else
        reduce_to_False (destruct_disjoint_list ctx (compute_disjoint_list ctx h))) 1

val disjoint_simproc =
  let val cache = Unsynchronized.ref NONE
      fun mk_disj_info ctx hthm =
        let fun update () =
              let val res =
                          fold (Net.insert_term (K false))
                            (map (fn thm => (prop_of thm,thm))
                              (reduce_to_False
                                (destruct_disjoint_list ctx
                                  (compute_disjoint_list ctx hthm))))
                            Net.empty
              in
                  cache := SOME (hthm, res);
                res
              end
        in
            case !cache of
                NONE => update()
              | SOME (hthm', res) =>
                if Thm.eq_thm (hthm, hthm')
                then res
                else update()
        end
      fun proc psi ss ct =
      let val ctx = Simplifier.the_context ss
          val thy = Proof_Context.theory_of ctx
      in
          case the_heap_prem (Simplifier.prems_of ss) of
             SOME (_,_,hthm) =>
             let val info = mk_disj_info ctx hthm
                 fun direct_prove st =
                   get_first (fn thm =>
                                (SOME (thm COMP st)
                                 handle THM _ => NONE))
                             (Net.unify_term info (term_of (cprem_of st 1)))
                 fun sym st =
                    direct_prove (@{thm sym} RS st)
                    handle THM _ =>
                   (direct_prove (@{thm not_sym} RS st)
                    handle THM _ =>
                   (direct_prove (@{thm disjoint_sym} RS st)
                    handle THM _ => NONE))

                 fun prove_from_false () =
                   case filter (fn thm => prop_of thm aconv @{prop False})
                          (Net.unify_term info @{prop False}) of
                     [] => NONE
                    | thm :: _ => SOME (thm COMP_INCR FalseE COMP_INCR @{thm Eq_TrueI})
             in
                case term_of ct of
                  Const(@{const_name "Lattices.inf"},_) $ _ $ _ =>
                  let val T = rep_cterm ct |> #T
                      val psi = cterm_of thy (HOLogic.mk_Trueprop (HOLogic.mk_eq (term_of ct, Var(("res",0),T))))
                      val st = Drule.instantiate' [] [SOME psi] @{thm "Pure.asm_rl"} RS eq_reflection
                  in
                      case direct_prove st of
                         NONE => (case direct_prove (@{thm Int_commute} RS (@{thm trans} RS st)) of
                         NONE => NONE
                       | some => some)
                       | some => some
                  end
               | _ =>
                let val toTrue = Drule.instantiate' [] [SOME ct] @{thm Eq_TrueI}
                    val toFalse = Drule.instantiate' [] [SOME ct] @{thm Eq_FalseI}
                       |> Raw_Simplifier.rewrite_rule [@{thm not_not}]
                in
                    case direct_prove toTrue of
                  NONE => (case direct_prove toFalse of
                  NONE => (case sym toTrue of
                  NONE => (case sym toFalse of
                  NONE => (case prove_from_false () of
                    NONE => NONE
                  | some => some)
                  | some => some)
                  | some => some)
                  | some => some)
                  | some => some
                 end
             end
           | NONE => NONE
      end
  in
      Simplifier.make_simproc
      { name = "disjoint by heap",
        lhss = [ @{cpat "(?a :: addr) ~= ?b"},
                 @{cpat "(?a :: addr) ~: ?B"},
                 @{cpat "(?A :: addr set) Un ?B = {}"},
                 @{cpat "(?a :: addr) = ?b"},
                 @{cpat "(?a :: addr) : ?B"},
                 @{cpat "(?A :: addr set) Int ?B"},
                 @{cpat "False"} ],
        proc = proc,
        identifier = []
      }
  end

val disjoint_simproc_setup =
  Context.theory_map (Simplifier.map_ss (fn ss => ss addsimprocs [ disjoint_simproc ]))


(*============================================================
               Side conditions in heap components
  ============================================================*)

fun compute_heap_sc ctx hthm thms =
  let val tr_old = tr
      fun tr f = tr_old ctx f
      val pretty_cterm = pretty_cterm ctx
      val pretty_thm = pretty_thm ctx
      val P = heap_prem_spatial hthm
      val () = tr (fn () => trace_list "heap_sc" [pretty_cterm P])
      val thy = theory_of_thm hthm
      fun loop P =
        (tr (fn () => trace_list "loop" [ pretty_cterm P ]);
        case term_of P of
            (@{const star} $ _ $ _) =>
            let val (P',Q') = Thm.dest_binop P
                val disj_inst  = instantiate' [] [SOME P',NONE, SOME Q'] 
                                   (Thm.permute_prems 0 1 @{thm side_cond_star})
                val left = Drule.lift_all (Thm.cprem_of disj_inst 1) (loop P')
                val right = Drule.lift_all (Thm.cprem_of disj_inst 2) (loop Q')
                val _ = tr (fn () => trace_list "combine"
                          [ pretty_thm disj_inst,
                            pretty_thm left,
                            pretty_thm right ])
                val res = right INCR_COMP (left COMP_INCR disj_inst)
                val _ = tr (fn () => trace_thm ctx "res" res)
            in
                res
            end
          | t =>
            (let fun matches thm =
                 let val thy' = Theory.merge (thy, theory_of_thm thm)
                     val t' = term_of (Thm.dest_fun (Thm.dest_arg (Thm.cprem_of thm 1)))
                     val p = Syntax.pretty_term ctx
                 in
                     Pattern.matches thy' (t', t)
                 end handle (CTERM _) => false
                   | (THM _) => false
             in
                 case find_first matches thms of
                     SOME thm => thm
                   | NONE => @{thm side_cond_TrueI}
             end))
      val loop_res = hthm COMP (loop P)

      val _ = tr (fn () => trace_thm ctx "loop-res" loop_res)
  in
      loop_res
  end

fun heap_sc_tac thms ctx _ hthm =
  let val res = compute_heap_sc ctx hthm thms
                |> HOLogic.conj_elims
                |> filter (not o thm_proves @{cprop True})
  in
      case res of
          [] => no_tac
        | _ => cut_facts_tac res 1
  end

(*============================================================
               Automatic unfolding
  ============================================================*)

datatype step = L | R
type path = step list

fun string_of_path [] = "<empty>"
  | string_of_path [L] = "L"
  | string_of_path [R] = "R"
  | string_of_path (L :: xs) = "L"^string_of_path xs
  | string_of_path (R :: xs) = "R"^string_of_path xs

val pretty_path = Pretty.str o string_of_path

(*
  The bound variables in the atoms refer to ex-bound
  variables in the unfold entry.

  The 'unfoldable' flag is set to "false" if the pre-processing has
  already accounted for all possible unfoldings, such that they do
  not need to be addressed at run-time.
*)
type atom = {
  t : term,
  pos : path,
  unfoldable : bool
}

fun atom_map_term f { t, pos,unfoldable } = { t = f t, pos = pos, unfoldable = unfoldable }
fun atom_map_pos f { t, pos,unfoldable } = { t = t, pos = f pos, unfoldable = unfoldable }
fun atom_prefix_pos pos' = atom_map_pos (fn pos => pos' @ pos)
fun atom_suffix_pos pos' = atom_map_pos (fn pos => pos @ pos')

fun atom_subst_ex subst {t,pos,unfoldable} =
  { t = subst_bounds (subst,t),
    pos = pos,
    unfoldable = unfoldable}

fun atom_incr_bv i lev {t,pos,unfoldable} =
  {t=incr_bv (i,lev,t), pos=pos,unfoldable=unfoldable}

type unfold_step = {
  ex  : (string * typ) list,
  pos : path,
  thm : thm
}

fun unfold_step_map_thm f {ex,pos,thm} =
  {ex=ex,pos=pos, thm = f thm }

(*
  An unfold entry compounds several given unfold rules.

  Contrary to lightweight separation, it is not possible
  to create a single unfold theorem, because the premises
  of inner applications my reference the ex-bounds variables
  of the outer applications.

  However, at the meta-level, we can simulate the extraction
  of the exs, and express the pre-conditions in terms of these
  as a flat list: given an instance of the ex-bounds, i.e. of the
  theorems, the pre-conditions are also flat.
*)
type unfold_entry = {
  left   : atom,
  right  : atom,
  rem    : atom list,
  ex     : (string * typ) list, (* concatenated from all steps *)
  pre    : term list, (* props, the premises to be proven *)
  sc     : term, (* the side-condition on the rhs *)
  maxidx : int, (* of all the unfold_steps *)
  unfold : unfold_step list, (* how to unfold this step *)
  left_only : bool, (* whether this is strictly a left-unfolding rule *)
  trail  : string list (* tmp: break cycles by force, internal *)
}

type unfold_data = {
  steps : (bool * thm) list,
  net   : unfold_entry Net.net option
}

fun subst_ex ex t =
  subst_bounds (map Free ex, t)

(*
fun maxidx_of_entry ({left,right,rem,ex,pre,sc,unfold,trail}:unfold_entry) =
  fold (curry Int.max)
    (map maxidx_of_term
        ([ #t left, #t right,sc ] @ pre @ map #t rem)) ~1
*)

fun env_map_maxidx f (Envir.Envir {maxidx,tenv,tyenv}) =
  Envir.Envir{maxidx=f maxidx,tenv=tenv,tyenv=tyenv}

fun unfold_entry_prepare ex_subst env
  ({left,right,rem,ex,pre,sc,unfold,maxidx,left_only,trail}:unfold_entry)=
  let val idx = Envir.maxidx_of env
      val t = Logic.incr_indexes ([], idx) #>
              curry subst_bounds ex_subst             
      val a = atom_map_term t
      val entry =
         { left = a left,
           right = a right,
           rem   = map a rem,
           ex = ex, (* still required? *)
           pre = map t pre,
           sc = t sc,
           maxidx = idx + maxidx,
           unfold = map (unfold_step_map_thm (Thm.incr_indexes idx)) unfold,
           left_only = left_only,
           trail = trail }
  in
     (entry : unfold_entry,
      env_map_maxidx (K (idx + maxidx)) env)
  end

fun ex_to_vars ex idx =
  let val names = Name.variant_list [] (map fst ex)
  in
      map (fn (x,T) => Var((x,idx),T)) 
      (names ~~ map snd ex)
  end

fun ex_to_frees ex ctx =
  let val (xs',ctx') = Variable.variant_fixes (map fst ex) ctx
  in
      (map Free (xs' ~~ map snd ex), ctx')
  end

fun pretty_atom ctx ({t,pos,unfoldable} : atom) =
  Pretty.block [ Syntax.pretty_term ctx t,
                 Pretty.str ":",
                 pretty_path pos,
                 if unfoldable
                 then Pretty.str "(unfoldable)"
                 else Pretty.str "" ]

fun pretty_atoms ctx atoms =
  Pretty.separate " * " (map (pretty_atom ctx) atoms)

fun pretty_ex ex =
  Pretty.str (case ex of
                  [] => "\<E> --"
                | _ =>  "\<E> "^commas (map fst ex)^"")

fun pretty_ex_terms ctx [] = Pretty.str "\<E> --"
  | pretty_ex_terms ctx ts =
    Pretty.block (Pretty.str "\<E> ":: (Pretty.separate " | " (map (pretty_term ctx) ts)))

fun pretty_unfold_list ctx unfolds =
  Pretty.big_list "unfold"
    (map (fn ({pos,ex,thm} : unfold_step) =>
            Pretty.block [ Pretty.str "@",
                           pretty_path pos,
                           Pretty.str ": ",
                           pretty_ex ex,
                           Pretty.str ". ",
                           pretty_thm ctx thm ])
        unfolds)

fun pretty_unfold_entry ctx ({ left, right, ex, sc, rem, unfold, maxidx, ... } : unfold_entry) =
  let val prep = atom_map_term (subst_ex (rev ex))
  in
  Pretty.chunks [ Pretty.block [ pretty_atom ctx (prep left),
                                 Pretty.str " -> ",
                                 pretty_atom ctx (prep right) ],
                   pretty_ex ex,
                   pretty_term ctx (subst_ex (rev ex) sc),
                   Pretty.big_list "remainder" (map (pretty_atom ctx o prep) rem),
                   Pretty.str ("maxidx: "^string_of_int maxidx),
                   pretty_unfold_list ctx unfold
                 ]
  end

(*
  Proof search happens at the meta-level, so we can use flat variants of
  the occurring variables. The worst that can happen is that proof
  reconstruction fails
*)
fun flat_vars t =
  let fun flat (((x,i),T),n) =
        let val (args,T') = strip_type T
            val (xs,ys) = chop n args
        in
            ((x,i), Logic.rlist_abs (map (pair "_") xs, Var((x^"_",i), ys ---> T')))
        end
      fun all (t as (a $ b)) res =
          let val (f,args) = strip_comb t
          in
              if is_Var f
              then case AList.lookup (op=) res (dest_Var f) of
                       NONE => (dest_Var f, length args) :: res
                     | SOME len =>
                        (if len <> length args
                         then error ("Cannot flatten: different argument lengths for "^
                                        Syntax.string_of_term_global Pure.thy f)
                         else ();
                         res)
              else all f (fold all args res)
          end
        | all (Abs(x,T,t)) res = all t res
        | all _ res = res 
      
      val occs = all t []
  in
      (Envir.beta_norm (Term.subst_Vars (map flat occs) t))
  end

fun flat_vars_ct ct =
  cterm_of (theory_of_cterm ct) (flat_vars (term_of ct))

(* override for testing; we loose the bounds 
val flat_vars = I 
*)

fun dest_spatial ctx unfoldable t =
  let fun loop pos (@{const star} $ l $ r) res =
        (res |> loop (L :: pos) l
             |> loop (R :: pos) r)
        | loop pos (Const(@{const_name heap_ex},_) $ (Abs(_,_,b))) res =
          loop pos b res
        | loop pos (@{const heap_conj} $ l $ _) res =
          loop pos l res
        | loop pos t res =
             ({ t=t, pos=rev pos, unfoldable = unfoldable  }: atom ) :: res
      fun loop_ex (Const(@{const_name heap_ex},_) $ (Abs(x,T,b))) =
          (x,T) :: loop_ex b
        | loop_ex t = []

      fun loop_sc (Const(@{const_name heap_ex},_) $ (Abs(x,T,b))) =
        loop_sc b
        | loop_sc (@{const heap_conj} $ _ $ c) = c
        | loop_sc _ = @{term True}
      val atoms = loop [] t []
      val sc = loop_sc t
      (*
      val _ = trace_list "dest_spatial" 
                (pretty_term ctx t ::
                    pretty_term ctx sc ::
                    map (pretty_atom ctx []) atoms)
                    *)
  in
      (loop_ex t, sc, rev atoms)
  end

fun dest_unfold_rl thm =
  let val prems = Logic.strip_imp_prems (prop_of thm)
      val prop = HOLogic.dest_Trueprop (Logic.strip_imp_concl (prop_of thm))
      val (l,r) = HOLogic.dest_eq prop
  in
      (prems,l,r)
  end

fun trail_entry l =
  case head_of l of
      Var _ => "var"
    | Const(c,_) => c
    | Free(x,_) => x

fun make_refl (i,(left_only,thm)) ctx =
  let val ((_,[thm']),ctx') = Variable.import false [thm] ctx
      val (prems, l,r) = dest_unfold_rl thm'
      val (ex,sc,atoms) = dest_spatial ctx false r
      fun mk a rem =
        { left = { t=l,pos=[],unfoldable=false },
          right = a,
          ex    = ex,
          sc    = sc,
          pre   = prems,
          rem   = rem,
          unfold = [{pos=[],ex=ex,thm=thm'}],
          maxidx = ~1,
          trail   = [ trail_entry l ],
          left_only = left_only
        } : unfold_entry
  in
      (map_remainder mk atoms, ctx')
  end

fun make_key l r =
  @{const HOL.conj} $ l $ r

(*
   The bounds get replaced by variables for the lookup
   to succeed for unfold-right. It's a conservative
   filter anyway.
*)
fun make_key_entry ctx e =
  let val l = #t (#left e)
      val r = #t (#right e)
      val bnos = (fold (curry Int.max) (loose_bnos l @ loose_bnos r) 0) + 1
      val s = curry subst_bounds (replicate  bnos(Var (("x",0), @{typ unit})))
      val k = make_key (s l) (s r)
  in
      k
  end

local
fun make_insts ctx bsubst env ts =
  let val Tinsts =
         fold Term.add_tvars ts []
         |> map (fn v => (TVar v, Envir.norm_type (Envir.type_env env) (TVar v)))
      val tinsts = 
         fold Term.add_vars ts []
         |> map (fn v =>
                   (Var v,
                    Term.subst_bounds
                       (bsubst, (Envir.norm_term env (Var v)))))
      fun pretty_map (a,b) = Pretty.block [ a, Pretty.str " -> ", b ]
      val _ = tr ctx (fn () =>
                trace_list "norm_thm - bsubst & insts"
                (Pretty.block (Pretty.str "bsubst:" ::
                                  Pretty.separate " " (map (pretty_term ctx) bsubst)) ::
                    map (pretty_map o pairself (pretty_typ ctx)) Tinsts @
                    map (pretty_map o pairself (pretty_term ctx)) tinsts))
      val thy = Proof_Context.theory_of ctx
      val Tinsts' = map (pairself (ctyp_of thy)) Tinsts
      val tinsts' = map (pairself (cterm_of thy)) tinsts
  in
      (Tinsts',tinsts')
  end
in
fun norm_cterms ctx bsubst env cts =
  map (Thm.instantiate_cterm (make_insts ctx bsubst env (map term_of cts))) cts

fun norm_thms ctx bsubst env thms =
  map (Thm.instantiate (make_insts ctx bsubst env (map prop_of thms))) thms
end

fun rhs_conv cv ct =
  let open Conv
  in case term_of ct of
         @{const "==>"} $ _ $ _ =>
         arg_conv (rhs_conv cv) ct
       | Const(@{const_name "=="},_) $ _ $ _ =>
         arg_conv cv ct
       | @{const "Trueprop"} $ _ =>
         arg_conv cv ct
       | Const(@{const_name "HOL.eq"},_) $ _ $ _ =>
         arg_conv cv ct
       | _ => raise Fail("rhs_conv: "^string_of_cterm (Proof_Context.init_global (theory_of_cterm ct)) ct)
  end

fun make_data ctx (steps : (bool * thm) list)=
  let val s = map_index I (distinct (Thm.eq_thm o pairself snd) steps)
      val (init,ctx') = apfst flat (fold_map make_refl s ctx)
      val unet = fold (Net.insert_term (K false))
                   (map (fn s =>
                            let val (prems,l,r) = dest_unfold_rl (snd (snd s))
                            in (l, (l,r,s)) end)
                       s)
                   Net.empty
      
      fun unify pat obj = Pattern.unify (Proof_Context.theory_of ctx) (pat,obj) (Envir.empty 0)
      val _ = tr ctx (fn () => trace_list "init" 
                        (Pretty.separate "----"
                            (map (pretty_unfold_entry ctx) init)))

      fun enable_unfold {left,right,ex,sc,pre,rem,unfold,maxidx,left_only,trail} =
        { left=left,
          right= { t= #t right, pos = #pos right, unfoldable = true },
          ex= ex, sc=sc,pre=pre,rem=rem,
          unfold=unfold,maxidx=maxidx,left_only=left_only,trail=trail }

      fun step (ue as {left,right,ex,sc,pre,rem,unfold,maxidx,left_only,trail}) 
        (l,r,(i,(left_only',s))) =
        let val tr_entry = trail_entry l
        in
        if member (op=) trail tr_entry
        then ([], false)
          (* allow run-time unfoldings on matching atoms, but do not perform the unfoldings
            let val subst = unify (#t right) l
                val _ = tr ctx (fn () =>
                      trace_list "enable runtime unfolding on right of"
                                  [ pretty_unfold_entry ctx ue ])
            in
                ([],true)
            end handle Pattern.Unif => ([],false) *)
        else
        let val (ex0,_,_) = dest_spatial ctx false r
            val right = atom_incr_bv (length ex0) 0 right
                         (* place right side below the new existentials
                            before unification *)
            val subst = unify (#t right) l
            val r' = Envir.norm_term subst r (* TODO normalize all, or match *)
            val _ = tr ctx (fn () =>
                      trace_list "step with right"
                      [ pretty_term ctx (#t right),
                        pretty_term ctx l,
                        pretty_term ctx r,
                        pretty_term ctx r' ])
            val (ex',sc',atoms) = dest_spatial ctx false r'
            val atoms' = map (atom_prefix_pos (#pos right)) atoms
            val rem_incr = map (atom_incr_bv (length ex') 0) rem
            val (prems', _,_) = dest_unfold_rl s

            (* the only bounds in the prems' are those placed there
               by unification: simulate this behaviour *)
            val pre' = prems' |> map (Envir.norm_term subst)
                              |> map (incr_boundvars (length ex))

            fun variant_ex_names ex ex' =
              Name.variant_list (map fst ex) (map fst ex') ~~ map snd ex' 
            
            (* val unfold' = Drule.implies_intr_list hyps unfold' *)

            fun has_vars pos t =
                case Term.add_vars t [] of
                     [] => false
                  | vs => 
                    ((if Config.get ctx heap_trace
                     then tracing else warning)
                     ("Dropping step: variables in step result "^pos^
                         commas (map (Syntax.string_of_term ctx o Var) vs)^
                         "\n"^Syntax.string_of_term ctx t);
                     true)
            
            fun mk a' rem' =
              let val r = { left = left, 
                            right = { t = incr_bv (length ex', length ex', #t a'),
                                      pos = #pos a',
                                      unfoldable = false (* assume that unfolds will take place
                                                            during preparation, until proven otherwise *)
                                   },
                            ex = ex @ variant_ex_names ex ex',
                            pre = pre @ pre',
                            sc = @{const HOL.conj} 
                                   $ incr_bv (length ex',0, sc)
                                   $ sc',
                            rem   = rem_incr @ rem',
                            unfold = unfold @ [ {pos= #pos right,ex=ex',thm=s}],
                            maxidx = ~1, (* no variables here! *)
                            left_only = left_only orelse left_only',
                            trail = tr_entry :: trail
                          } : unfold_entry
              in
                  tr ctx (fn () => trace_list "mkstep"
                            [ pretty_unfold_entry ctx r ]);
                  if has_vars "left" (#t (#left r)) orelse
                     has_vars "right" (#t (#right r)) orelse
                     exists (has_vars "pre") (#pre r) orelse
                     exists (has_vars "rem" o #t) (#rem r)
                  then NONE
                  else SOME r
              end
        in
            (map_filter I (map_remainder mk atoms'),false)
        end handle Pattern.Unif => ([],false)
      end
      
      fun sat [] res = res
        | sat ((s : unfold_entry) :: queue) res =
          let val stepped = map (step s) (Net.unify_term unet (#t (#right s)))
          in
              if exists snd stepped
              then sat (flat (map fst stepped) @ queue) (enable_unfold s :: res)
              else sat (flat (map fst stepped) @ queue) (s :: res)
          end

      val sat_res = sat init []
      val _ = tr ctx (fn () => trace_list "sat result"
                        (map (pretty_unfold_entry ctx) sat_res))

      val prep_term = singleton (Variable.export_terms ctx' ctx)
      val prep_atom = atom_map_term prep_term
      val prep_thm =
          singleton (Variable.export ctx' ctx)
      
      fun prep_entry {left,right,rem,ex,sc,pre,unfold,maxidx,left_only,trail} =
        let val steps = map (unfold_step_map_thm prep_thm) unfold
        in
            { left = prep_atom left,
              right = prep_atom right,
              rem = map prep_atom rem,
              ex = ex,
              pre = map prep_term pre,
              sc = prep_term sc,
              unfold = steps,
              maxidx = fold (curry Int.max o Thm.maxidx_of o #thm) steps ~1,
              left_only = left_only,
              trail = []
            }
        end
      val entries = map prep_entry sat_res
      val _ = tr ctx (fn () => trace_list "prepared"
                (map (pretty_unfold_entry ctx) entries))

      val net = fold (Net.insert_term (K false))
                  (map (fn e => (make_key_entry ctx e, e)) entries)
                  Net.empty
  in
      {steps = map snd s, net = SOME net } : unfold_data
  end

fun data_merge ({steps=steps, ...}, {steps=steps', ...}) =
  {steps = steps @ steps', net = NONE }

structure UnfoldData = Generic_Data(
  type T = unfold_data
  val empty = {steps = [], net= SOME Net.empty }
  val extend = I
  val merge = data_merge
);

fun consolidate_data thy =
  let val data = UnfoldData.get (Context.Theory thy)
  in
      case #net data of
          NONE => SOME (Context.theory_map
                           (UnfoldData.put (make_data (Proof_Context.init_global thy)
                                           (#steps data))) thy)
        | SOME net => NONE
  end

fun add_thm left_only thm (data as ({steps,...} : unfold_data)) =
  if exists (curry Thm.eq_thm thm o snd) steps
  then data
  else make_data (Proof_Context.init_global (theory_of_thm thm)) ((left_only,thm) :: steps)

fun del_thm thm ({steps,net} : unfold_data) =
  make_data (Proof_Context.init_global (theory_of_thm thm))
            (remove (Thm.eq_thm o pairself snd) (false,thm) steps)

fun add_unfold_eq left_only = UnfoldData.map o (add_thm left_only)

fun add left_only = Thm.declaration_attribute (add_unfold_eq left_only)
val del = Thm.declaration_attribute (UnfoldData.map o del_thm);

val attr_setup =
  Attrib.setup (Binding.name "sep_unfold") (Attrib.add_del (add false) del) ("declare automatic unfolding") #>
  Attrib.setup (Binding.name "sep_unfold_left") (Attrib.add_del (add true) del) ("declare automatic unfolding")

(*==================================================================
 Normalize a term to pull those positions mentioned in ps
 to the front, pushing the others to the back.
 ==================================================================*)
val star_extract_left =
   @{lemma  "(P \<star> Q) \<star> R == P \<star> Q \<star> R"
   by (rule eq_reflection) simp}

val star_extract_right =
   @{lemma "Q \<star> (P \<star> R) == P \<star> (Q \<star> R)"
   by (rule eq_reflection) (rule star_swap_front)}

val star_comm = @{thm star_comm} RS eq_reflection

fun assert msg t =
  if not t then error msg else ()

fun path_take_hd p (p' :: ps) =
  if p <> p'
  then error ("Unexpected: take_hd "^string_of_path [p']^" /= "^string_of_path [p])
  else ps

fun path_cons_hd p ps = p :: ps
fun path_app ps ps' = ps @ ps

val star_extractLL = @{thm star_extractLL} RS eq_reflection
val star_extractLR = @{thm star_extractLR} RS eq_reflection

fun rearrange_conv ctx ps =
  let open Conv
      fun loop ps =
         cases ps
      and cases [] = all_conv
        | cases [ [] ] = all_conv
        | cases ([L] :: ps) =
          arg_conv (loop (map (path_take_hd R) ps))
        | cases (ps as ((L :: L :: _) :: _)) =
          rewr_conv star_extractLL then_conv
          loop (map (fn (L :: L :: pos) => L :: pos
                       | (L :: R :: pos) => R :: L :: pos
                       | (R :: pos) => R :: R :: pos) ps)
        | cases (ps as ((L :: R :: _) :: _)) =
          rewr_conv star_extractLR then_conv
          loop (map (fn (L :: L :: pos) => R :: L :: pos
                       | (L :: R :: pos) => L :: pos
                       | (R :: pos) => R :: R :: pos) ps)
        | cases (ps as ((R :: _) :: _)) =
          rewr_conv star_comm then_conv
          loop (map (fn (L :: pos) => R :: pos | (R :: pos) => L :: pos) ps)
        | cases ps =
          raise Fail("rearrange_loop "^commas (map string_of_path ps))
  in
      loop ps
  end

fun rearrange_heap_prop ctx ps P =
  Conv.arg_conv (Conv.fun_conv (rearrange_conv ctx ps)) P

fun rearrange_heap_prem ctx ps thm =
  Thm.equal_elim (rearrange_heap_prop ctx ps (cprop_of thm)) thm

(*============================================================
   The matching procedure
  ============================================================*)

(* applied unfolding rule
   it is given by the absolute (!) path at which
   it needs to be applied and the existential instance,
   which may be either frees or variables.
*)
type unfold_inst = {
  pos : path,
  ex  : term list,
  thm : thm (* with incremented indexes *)
}

fun pretty_unfold_inst ctx env ({ pos,ex,thm } : unfold_inst) =
  let val prep = Envir.norm_term env
  in
      Pretty.block [ pretty_path pos,
                     Pretty.str " ",
                     pretty_ex_terms ctx (map prep ex),
                     Pretty.str ": ",
                     pretty_term ctx (prep (prop_of thm)) ]
  end

(*
  One side of a matching is made up from the atoms, which
  must be interpreted under the existentials of the unfold_insts.
*)
type side = {
  atoms  : atom list,
  unfold : unfold_inst list
}

fun pretty_side ctx env ({atoms,unfold}:side) =
  let val prep = Envir.norm_term env
  in
      Pretty.chunks [
        Pretty.block (Pretty.separate " * "
                         (map (pretty_atom ctx o atom_map_term prep) atoms)),
        Pretty.big_list "unfold"
                  (Pretty.separate "-------"
                      (map (pretty_unfold_inst ctx env) unfold))
      ]
  end

fun side_map_atoms f ({atoms,unfold} : side) =
   {atoms = f atoms,unfold=unfold} : side

fun side_map_unfold f ({atoms,unfold} : side) =
   {atoms = atoms,unfold=f unfold} : side

fun side_add_unfolds u' =
  side_map_unfold (fn u => u @ u')

fun side_ex ({unfold,...}:side) =
  Library.foldr (fn ({ex,...},res) => ex @ res) (unfold,[])

datatype eq_type = Cancel | Frame
fun pretty_eq_type Cancel = Pretty.str "cancel"
  | pretty_eq_type Frame = Pretty.str "frame"

(*
 Heuristics: solve pre-conditions immediately
*)
datatype cond_cat =
  CondPre | CondSc

fun pretty_cond_cat CondPre = Pretty.str "pre"
  | pretty_cond_cat CondSc  = Pretty.str "side"

fun tag_cond_pre p = (CondPre,p)
fun tag_cond_side p = (CondSc,p)

(*
 Variable atoms on the left are never unified, while those on the
 right receive the residue, or emp, for a frame computation.

 The unfold theorems are computed lazily in order to be fast.
*)
datatype match_state = State of {
  left    : side,
  right   : side,
  eqs     : (atom * atom * eq_type) list, (* matched atoms (eq after env.norm)*)
  prems   : thm list, (* the available knowledge; extended by unfolding on the left *)
  pending : (cond_cat * term) list, (* pending side-conditions of unfolding rules *)
  solved  : thm list, (* previously solved side-conditions *)
  env     : Envir.env,
  names   : Proof.context (* keep track of frees for inventing ex-bounds *)
}

fun state_map_pending f (State {left,right,eqs,prems,pending,solved,env,names}) =
  State {left=left,right=right,eqs=eqs,prems=prems,
          pending=f pending,solved=solved,
          env=env,names=names}

fun state_map_solved f (State {left,right,eqs,prems,pending,solved,env,names}) =
  State {left=left,right=right,eqs=eqs,prems=prems,
          pending=pending,solved=f solved,
          env=env,names=names}

fun state_map_env f (State {left,right,eqs,prems,pending,solved,env,names}) =
  State {left=left,right=right,eqs=eqs,prems=prems,
          pending=pending,solved=solved,
          env=f env,names=names}

fun pretty_pending ctx (cat,t) =
  Pretty.block [ pretty_cond_cat cat,
                 Pretty.str ": ",
                 pretty_term ctx t ]

fun pretty_match_state ctx (State {left,right,pending,solved,eqs,prems,env,...}) =
  let val prep = Envir.norm_term env
      val prepa = atom_map_term prep
  in
  Pretty.chunks [ Pretty.block [ Pretty.str "left: ",
                                 pretty_side ctx env left ],
                  Pretty.block [ Pretty.str "right: ",
                                 pretty_side ctx env right ],
                  Pretty.big_list "eqs:"
                     (map (fn (a,b,eqt) =>
                              Pretty.block [ pretty_atom ctx (prepa a),
                                             Pretty.str " == ",
                                             pretty_atom ctx (prepa b),
                                             Pretty.str " ",
                                             pretty_eq_type eqt ])
                         eqs),
                  Pretty.big_list "pending:"
                    (map (pretty_pending ctx o apsnd prep) pending),
                  Pretty.big_list "solved:"
                    (map (pretty_thm ctx) solved),
                  Pretty.big_list "prems:"
                     (map (pretty_thm ctx) prems) ]
  end

fun init_state ctx left right prems =
  let val maxidx = fold (Term.maxidx_term ) [left,right] ~1;
      val idx    = maxidx + 1
      val (exL,scL,atomsL) = dest_spatial ctx true left
      val (exR,scR,atomsR) = dest_spatial ctx true right
      (*
      val atomsL = map (atom_map_term flat_vars) atomsL
      val atomsR = map (atom_map_term flat_vars) atomsR
      *)
      val cert = cterm_of (Proof_Context.theory_of ctx)
      val (exL,ctx) = ex_to_frees exL ctx
      val exR = ex_to_vars exR idx
      val idx = idx + 1
      fun mk_unfold_inst_refl ex =
        { pos = [], ex = ex, thm = Thm.incr_indexes idx (@{thm refl} RS @{thm eq_reflection}) }
      val init =
      State {
        left = { atoms  = map (atom_subst_ex (rev exL)) atomsL,
                 unfold = [ mk_unfold_inst_refl exL ] },
        right = { atoms = map (atom_subst_ex (rev exR)) atomsR,
                  unfold = [ mk_unfold_inst_refl exR ] },
        eqs   = [],
        prems = prems,
        pending = [],
        solved = [],
        env   = Envir.empty (idx + 1),
        names = fold Variable.declare_names [left,right] ctx
      }
      val _ = tr ctx (fn () => trace_list "initial heap match state"
                        [ pretty_match_state ctx init ])
  in
          init
  end

fun seq_map_prod f xs ys =
  Seq.maps (fn x => Seq.maps (fn y => f x y) ys) xs

fun fold_atom_pairs f (State s) unit =
  fold_product f
      (map_remainder pair (#atoms (#left s)))
      (map_remainder pair (#atoms (#right s)))
      unit

(*
  Sanity check on environment to track misbehaviour where it originates:
  lookup all vars bound in the environment. The "var has distinct types"
  originates there.
*)
(*
  local
fun lookup_check check tenv (xi, T) =
  (case Vartab.lookup tenv xi of
    NONE => NONE
  | SOME (U, t) => if check (T, U) then SOME t else var_clash xi T U);
*)
fun check_env (env as Envir.Envir {tenv,...})=
  ()
  (*
  List.app (fn v => ignore (Envir.lookup (env,v)))
     (Vartab.keys tenv)
  *)

local
fun body_type env =
  let
    val tyenv = Envir.type_env env;
    fun bT (Type ("fun", [_, T])) = bT T
      | bT (T as TVar v) =
          (case Type.lookup tyenv v of
            NONE => T
          | SOME T' => bT T')
      | bT T = T;
  in bT end;

fun binder_types env =
  let
    val tyenv = Envir.type_env env;
    fun bTs (Type ("fun", [T, U])) = T :: bTs U
      | bTs (TVar v) =
          (case Type.lookup tyenv v of
            NONE => []
          | SOME T' => bTs T')
      | bTs _ = [];
  in bTs end;

fun strip_type env T = (binder_types env T, body_type env T);


(*Abstraction over a list of types, like list_abs*)
fun types_abs ([], u) = u
  | types_abs (T :: Ts, u) = Abs ("", T, types_abs (Ts, u));

(*Abstraction over the binder of a type*)
fun type_abs (env, T, t) = types_abs (binder_types env T, t);
  
  (*For smash_flexflex1*)
fun var_head_of (env,t) : indexname * typ =
  (case head_of (strip_abs_body (Envir.norm_term env t)) of
    Var (v, T) => (v, T)
  | _ => error "smash failed");  (*not flexible, cannot use trivial substitution*)

fun body_type env =
  let
    val tyenv = Envir.type_env env;
    fun bT (Type ("fun", [_, T])) = bT T
      | bT (T as TVar v) =
          (case Type.lookup tyenv v of
            NONE => T
          | SOME T' => bT T')
      | bT T = T;
  in bT end;


(*Eliminate a flex-flex pair by the trivial substitution, see Huet (1975)
  Unifies ?f(t1...rm) with ?g(u1...un) by ?f -> %x1...xm.?a, ?g -> %x1...xn.?a
  Unfortunately, unifies ?f(t,u) with ?g(t,u) by ?f, ?g -> %(x,y)?a,
  though just ?g->?f is a more general unifier.
  Unlike Huet (1975), does not smash together all variables of same type --
    requires more work yet gives a less general unifier (fewer variables).
  Handles ?f(t1...rm) with ?f(u1...um) to avoid multiple updates. *)
fun smash_flexflex1 ((t, u), env) : Envir.env =
  let
    val vT as (v, T) = var_head_of (env, t)
    and wU as (w, U) = var_head_of (env, u);
    val (env', var) = Envir.genvar (#1 v) (env, body_type env T);
    val env'' = Envir.vupdate ((wU, type_abs (env', U, var)), env');
  in
    if vT = wU then env''  (*the other update would be identical*)
    else Envir.vupdate ((vT, type_abs (env', T, var)), env'')
  end;


(*Smash all flex-flexpairs.  Should allow selection of pairs by a predicate?*)
fun smash_flexflex (env, tpairs) : Envir.env =
  List.foldr smash_flexflex1 env tpairs;
 
 in
     
fun no_flex_flex (env,flex_flex) =
  case flex_flex of
      [] => env
    | ff => (warning (Pretty.string_of
                        (Pretty.big_list "Flex-flex pairs in result env"
                            (map (fn (x,y) =>
                                    Pretty.block [ 
                                      Syntax.pretty_term_global Pure.thy x,
                                      Pretty.str " =?= ",
                                      Syntax.pretty_term_global Pure.thy y
                                                 ]) ff)));
             smash_flexflex (env,flex_flex))
end      

fun unify_atoms ctx aps env =
    case Seq.pull (Unify.unifiers (Proof_Context.theory_of ctx, env,
                                     map (pairself #t) aps)) of
        NONE => Seq.empty
      | SOME(x,_) => Seq.single (no_flex_flex x)
  
fun cancel_direct ctx (State s) =
  let val thy = Proof_Context.theory_of ctx
      val prep = Envir.norm_term (#env s)
      val prepa = atom_map_term prep
      fun match_pair (l, ls) (r, rs) =
        let fun mk_state env' =
            let val () = tr ctx (fn () => trace_list "match_pair"
                               [ pretty_atom ctx (prepa l),
                                 pretty_atom ctx (prepa r) ])
                val _ = check_env env'
            in
              State {
                left = side_map_atoms (K ls) (#left s),
                right = side_map_atoms (K rs) (#right s),
                eqs   = (l,r,Cancel) :: #eqs s,
                prems = #prems s,
                pending = #pending s,
                solved = #solved s,
                env   = env',
                names = #names s }
            end
        in
            Seq.map mk_state (unify_atoms ctx [(l,r)] (#env s))
        end
      val is_var = is_Var o head_of
      fun add_possible_match (x as (l,_)) (y as (r,_)) z =
        if not (is_var (#t l)) andalso
            not (is_var (#t r)) andalso
            (Term.could_unify (#t l, #t r))
        then (x,y) :: z
        else z
  in
      Seq.FIRST (map (fn p => fn () => match_pair (fst p) (snd p))
                   (fold_atom_pairs add_possible_match (State s) [])) ()
  end

fun fold_atoms pos [] = (@{const emp},[])
  | fold_atoms pos [(x : atom)] = (#t x, [atom_map_pos (K pos) x])
  | fold_atoms pos ((x : atom) :: xs) =
    let val (t',xs') = fold_atoms (pos @ [ R ]) xs
    in (@{const star} $ #t x $ t',
        atom_map_pos (K (pos @ [ L ])) x :: xs')
    end

fun fill_frame ctx (State s) =
  let val thy = Proof_Context.theory_of ctx
  in
      case (#atoms (#right s)) of
          [ {t,pos,...} ] =>
          if is_Var (head_of t)
          then let val (t', left' : atom list) = fold_atoms pos (#atoms (#left s))
                   val _ = tr ctx (fn () => trace_list "fill frame"
                                      [ pretty_path pos,
                                        Pretty.str " : ",
                                        pretty_term ctx t,
                                        Pretty.str " -> ",
                                        pretty_term ctx t' ])
                   val env' =  no_flex_flex
                                 (Seq.hd (Unify.unifiers (thy,(#env s),[(t, t')])))
                                   handle Option =>
                                      fail_list "Expected unifier with frame"
                                          [ pretty_term ctx t,
                                            pretty_term ctx t' ]
                   val _ = check_env env'
               in Seq.single
                   (State {
                     left = side_map_atoms (K []) (#left s),
                     right = side_map_atoms (K []) (#right s),
                     eqs   = (map2 (fn a => fn b =>
                                     (a,b,Frame))
                                 (#atoms (#left s)) left') @ #eqs s,
                     prems = #prems s,
                     pending = #pending s,
                     solved = #solved s,
                     env   = env',
                     names = #names s })
               end
          else Seq.empty
        | _ => Seq.empty
  end

fun is_ground t =
   Term.fold_aterms (fn t => fn res => res andalso not (is_Var t)) t true

fun dest_conj (Const ("HOL.conj", _) $ t $ t') =
    dest_conj t @ dest_conj t'
  | dest_conj t = [t];

val solve_tac =
  Unsynchronized.ref (fn ctx => fn prems => fn ss =>
         asm_full_simp_tac ss 1)

val safe_solve_tac =
  Unsynchronized.ref (fn ctx => fn prems => fn ss =>
         safe_asm_full_simp_tac ss 1 THEN
         TRY (rtac @{thm refl} 1))

fun is_eq_prop (@{const Trueprop} $ (Const(@{const_name "HOL.eq"},_) $ _ $ _)) = true
  | is_eq_prop _ = false

fun sort_pending ts =
  let fun cat 1 (CondPre,t) = true
        | cat 2 (_,t) = is_eq_prop t andalso is_ground t
        | cat 3 (_,t) = is_eq_prop t
        | cat 4 (_,t) = is_ground t
        | cat 5 (_,t) = true
        | cat _ _ = false
  in
      flat (partition_list cat 1 5 ts)
  end

fun is_safe_solvable ctx t =
  case t of
      (@{const Trueprop} $ (Const(@{const_name "HOL.eq"},_) $ a $ b)) =>
      is_ground a orelse is_ground b
    | _ => false

fun solve_pending ctx only_pre (State s) =
  exception_trace (fn () =>
  let val thy = Proof_Context.theory_of ctx
      fun prove tac p =
        let val _ = tr ctx (fn () => trace_list "prove pre" [ pretty_term ctx p ])
        in (* Note: prove would zero the var indexes, leading to clashes *)
            SOME (Goal.prove_internal [] (cterm_of thy p) (K tac))
            handle exn => if Exn.is_interrupt exn
                          then reraise exn
                          else NONE
        end

      val ss = Simplifier.add_prems (#prems s)
                  (simpset_of ctx addsimps (#prems s))

      fun new_env env p' thm =
        Seq.hd (Unify.smash_unifiers thy [ (p', prop_of thm) ] env)
        handle exn => (if not (Exn.is_interrupt exn)
                       then fail_list "exn in new_env in solve_pending"
                              [ pretty_term ctx p',
                                pretty_term ctx (Envir.norm_term env p'),
                                pretty_term ctx (prop_of thm) ]
                       else ();
                       reraise exn)

      fun propagate (P as (c,p)) (pend,solved,env,false) =
        if is_eq_prop p
        then let val thm =
                    Goal.conclude (@{thm refl} INCR_COMP (Goal.init (cterm_of thy p)))
             in
                 (pend, thm :: solved, new_env env p thm, false)
             end
            handle THM _ => (P :: pend, solved,env, false)
        else (P :: pend, solved,env, false)

      fun solve (P as (_,p)) (pend,solved,env,true) = (P :: pend,solved,env,true)
        | solve (P as (c,p)) (pend,solved,env,false) =
          let val p' = Envir.norm_term env p
              val new_env = new_env env p'
          in
              if only_pre andalso c <> CondPre
              then propagate (c,p') (pend,solved,env,false)
              else if is_ground p'
              then case prove (!solve_tac ctx (#prems s) ss) p' of
                       NONE => (P :: pend, solved, env, true)
                     | SOME thm => (pend, thm :: solved, new_env thm, false)
              else case prove (!safe_solve_tac ctx (#prems s) ss) p' of
                       NONE => (P :: pend, solved, env, is_safe_solvable ctx p')
                     | SOME thm => (pend, thm :: solved, new_env thm, false)
          end

      val (pend',solved,env',fail) = fold solve (#pending s) ([],[],#env s,false)
      val pretty_pend = (pretty_pending ctx o apsnd (Envir.norm_term env'))
      val res_state = State s
                          |> state_map_pending (K (rev pend'))
                          |> state_map_solved (fn s => s @ (rev solved))
                          |> state_map_env (K env')
      val _ = tr ctx (fn () =>
                    trace_list "after solve pending"
                [ (if fail then Pretty.str "FAILED" else Pretty.str "success"),
                  Pretty.big_list "solved" (map (pretty_thm ctx) (rev solved)),
                  Pretty.big_list "remaining" (map pretty_pend (rev pend'))
                ])
      val _ = check_env env'
      val _ = if fail
              then tr_fail ctx (fn () =>
                trace_list "failed to prove side-conditions"
                    (map pretty_pend pend' @
                       [ Pretty.str "---- state ----",
                         pretty_match_state ctx (State s) ]))
              else ()
  in
      if fail
      then Seq.empty
      else Seq.single res_state
  end)
      
fun make_unfold_inst ctx ex pos e =
  let val (ex_rem,unfold') =
         fold (fn {pos=pos',ex=ex'',thm=thm} =>
                 fn (ex',res) =>
                   let val (ex_here,ex_rem) = chop (length ex'') ex'
                   in
                       (ex_rem, { pos = pos @ pos',
                                  ex = ex_here,
                                  thm=thm } :: res)
                   end)
               (#unfold e) (ex,[])
      val _ = if ex_rem <> []
              then error ("expected empty ex-remainder for step\n"^
                             Pretty.string_of (pretty_unfold_entry ctx e))
              else ()
  in
      rev unfold'
  end

fun unfold_left ctx (State s) =
  let val cert = cterm_of (Proof_Context.theory_of ctx)
      val lookup = Net.unify_term (the (#net (UnfoldData.get (Context.Proof ctx))))
      fun unfold_pair ((left as {t,pos,unfoldable},remL),(right as {t=t',pos=pos',...},remR), e0 : unfold_entry) =
        let val prep = Envir.norm_term (#env s)
            val prepa = atom_map_term prep
            val (ex',names') = ex_to_frees (#ex e0) (#names s)
            val (e,env) = unfold_entry_prepare (rev ex') (#env s) e0
            val _ = tr ctx (fn () => trace_list "unfold pair left"
                      [ Pretty.block [ pretty_atom ctx (prepa left),
                                       Pretty.str " ~ ",
                                       pretty_atom ctx (#left e) ],
                        Pretty.block [ pretty_atom ctx (prepa right),
                                       Pretty.str " ~ ",
                                       pretty_atom ctx (#right e) ],
                        Pretty.big_list "step rem"
                           (map (pretty_atom ctx) (#rem e)),
                        pretty_unfold_list ctx (#unfold e),
                        Pretty.big_list "pending pre"
                          (map (pretty_pending ctx o apsnd prep) (#pending s)) ])

            val unfold' = make_unfold_inst ctx ex' pos e

            fun mk_res env' =
              exception_trace (fn () =>
              let (*val sc' = #sc e |> Envir.norm_term env'
                                  |> curry op $ @{const Trueprop}
                                  |> cert
                                  |> Thm.assume *)
                  val _ = check_env env'
                  val res =
                     State {
                       left = { atoms = remL @ map (atom_prefix_pos pos) (#rem e),
                                unfold = #unfold (#left s) @ unfold'
                              },
                       right = #right s |> side_map_atoms (K remR),
                       eqs   = ({t = #t (#right e),
                                 pos = #pos left @ #pos (#right e),
                                 unfoldable = #unfoldable (#right e) },
                                right, Cancel) ::  #eqs s,
                       prems = #prems s, (* @ [ sc' ], *)
                       pending = sort_pending (#pending s @ (map tag_cond_pre (#pre e))),
                       solved = #solved s,
                       env   = env',
                       names = names'
                    }
                  val _ = tr ctx (fn () =>
                                    trace_list "unfold-left result" [
                                 pretty_match_state ctx res ])
              in
                  res
              end)
        in
            Seq.maps (solve_pending ctx true)
              (Seq.map mk_res
                (unify_atoms ctx [(left, #left e), (right, #right e)] env))
        end
      val is_var = is_Var o head_of
      fun add_unfold_entries (l as (a,_)) (r as (b,_)) res =
        if is_var (#t a) orelse is_var (#t b)
        then res
        else map (fn e => (l,r,e)) (lookup (make_key (#t a) (#t b))) @ res
  in
      Seq.maps unfold_pair
          (Seq.of_list (fold_atom_pairs add_unfold_entries (State s) []))
  end

fun unfold_right ctx (State s) =
  let val cert = cterm_of (Proof_Context.theory_of ctx)
      val lookup = Net.unify_term (the (#net (UnfoldData.get (Context.Proof ctx))))
      fun unfold_pair ((left as {t,pos,...},remL),(right as {t=t',pos=pos',unfoldable},remR), e0 : unfold_entry) =
        if #left_only e0
        then (tr ctx (fn () => trace_list "skipping left-only entry"
                         [ pretty_unfold_entry ctx e0 ]);
              Seq.empty)
        else
        let val prep = Envir.norm_term (#env s)
            val prepa = atom_map_term prep
            val idx = Envir.maxidx_of (#env s) + 1
            val ex' = ex_to_vars (#ex e0) idx
            val env = env_map_maxidx (K idx) (#env s)
            val (e,env) = unfold_entry_prepare (rev ex') env e0
            val _ = tr ctx (fn () => trace_list "unfold pair right"
                      [ Pretty.block [ pretty_atom ctx (prepa left),
                                       Pretty.str " ~ ",
                                       pretty_atom ctx (#right e) ],
                        Pretty.block [ pretty_atom ctx (prepa right),
                                       Pretty.str " ~ ",
                                       pretty_atom ctx (#left e) ],
                        pretty_ex_terms ctx ex',
                        Pretty.big_list "step rem"
                           (map (pretty_atom ctx) (#rem e)),
                        pretty_unfold_list ctx (#unfold e),
                        Pretty.big_list "pending pre"
                          (map (pretty_pending ctx o apsnd prep) (#pending s)),
                        Pretty.big_list "sc of unfolding"
                          [ pretty_term ctx (prep (#sc e)) ] ])

            val unfold' = make_unfold_inst ctx ex' pos' e

            fun mk_res env' =
              exception_trace (fn () =>
              let val res =
                     State {
                       left = #left s |> side_map_atoms (K remL),
                       right = { atoms = remR @ map (atom_prefix_pos pos') (#rem e),
                                 unfold = #unfold (#right s) @ unfold'
                               },
                       eqs   = (left,
                                {t = #t (#right e),
                                 pos = #pos right @ #pos (#right e),
                                 unfoldable = #unfoldable (#right e) },
                                Cancel) ::  #eqs s,
                       prems = #prems s,
                       pending = sort_pending (#pending s @
                                               map tag_cond_pre (#pre e) @
                                               map (tag_cond_side o HOLogic.mk_Trueprop) (dest_conj (#sc e))),
                       solved = #solved s,
                       env   = env',
                       names = #names s
                    }
                  val _ = tr ctx (fn () =>
                                    trace_list "unfold-right result" [
                                 pretty_match_state ctx res ])
                  val _ = check_env env'
              in
                  res
              end)
        in
            Seq.maps (solve_pending ctx true)
              (Seq.map mk_res
                (unify_atoms ctx [(left, #right e), (right, #left e)] env))
        end
      val is_var = is_Var o head_of
      fun add_unfold_entries (l as (a,_)) (r as (b,_)) res =
        if is_var (#t a) orelse is_var (#t b)
        then res
        else map (fn e => (l,r,e)) (lookup (make_key (#t b) (#t a))) @ res
  in
      Seq.maps unfold_pair
          (Seq.of_list (fold_atom_pairs add_unfold_entries (State s) []))
  end


(*
  Main function: compute a sequence of theorems
  "[| left'' ==> right'' |] ==> left' ==> right'"
  (under the given premises),
  where left' and right' are instances of left and right,
  and left'' and right'' contain the non-matching pairs.

  Implementation: state -> state seq, i.e. search for
  possible outcomes.
*)
fun strategy_full ctx left right prems =
  let open Seq
      val solve =
        REPEAT
           (cancel_direct ctx ORELSE
            fill_frame ctx ORELSE
            unfold_left ctx ORELSE
            unfold_right ctx)
  in
      solve (init_state ctx left right prems)
  end

fun strategy_cancel ctx left right prems =
  let open Seq
      val solve =
        REPEAT (cancel_direct ctx ORELSE
                fill_frame ctx)
  in
      solve (init_state ctx left right prems)
  end

fun match_components ctx heap_prem i thm =
 (tr ctx (fn () => trace_list "match components"
            [ pretty_thm ctx heap_prem,
              pretty_cterm ctx (cprem_of thm 1) ]);
  (case prop_of heap_prem of
      @{const Trueprop} $ ((@{const Heap.star} $ _ $ _) $ _) =>
      (rtac (heap_prem COMP_INCR @{thm star_match_components}) THEN'
       (Subgoal.FOCUS_PREMS (fn {context=ctx', prems=[heap_prem'], ...} =>
           ((match_components ctx' heap_prem' ORELSE'
             cut_facts_tac [ heap_prem' ])) 1) ctx))
    | _ => (TRY o rtac @{thm heap_unify_with_emp} THEN'
            (rtac heap_prem ORELSE'
            (cut_facts_tac [ heap_prem ])))) i thm)

val norm_ss = HOL_basic_ss addsimps @{thms norm_heap_simps}
val norm_heap_conv = Simplifier.rewrite norm_ss

fun mk_meta_eq thm =
  the_default thm (try (curry op RS thm) @{thm eq_reflection})

(**
 TODO determine the precise solve conditions for the specific rule.
*)
fun apply_unfold_conv ctx pos thm solved =
  let open Conv
      (* adapted from Conv.rewr_conv to handle pre-conditions after the match *)
      fun rewr_conv ctx ct =
        let val thm' = thm
                         |> Thm.incr_indexes (#maxidx (Thm.rep_cterm ct) + 1)
                         |> mk_meta_eq
            val lhs = thm' |> cprop_of |> Drule.strip_imp_concl |> Thm.dest_binop |> fst
            val thm'' = Thm.rename_boundvars (Thm.term_of lhs) (Thm.term_of ct) thm'
            val thm''' =
               Drule.instantiate_normalize (Thm.match (lhs, ct)) thm''
               handle Pattern.MATCH => raise CTERM ("apply unfold conv", [lhs, ct])
            val _ = tr ctx (fn () => trace_list "solve pre-conds on" [ pretty_thm ctx thm''' ])
            fun solve_pre thm =
              if nprems_of thm = 0
              then thm
              else case (get_first (try (fn s => s COMP thm)) solved) of
                       SOME thm' => solve_pre thm'
                     | NONE => error (string_of_list
                                    "Failed to solve pre-condition in unfold app"
                                    (pretty_thm ctx thm :: map (pretty_thm ctx) solved))

        in
            solve_pre thm'''
        end

      fun loop_spatial ctx bsubst [] =
          rewr_conv ctx 
        | loop_spatial ctx bsubst (L :: p') =
          arg1_conv (loop_spatial ctx bsubst p')
        | loop_spatial ctx bsubst (R :: p') =
          arg_conv (loop_spatial ctx bsubst p')

      fun loop_ex ctx bsubst ct =
        case term_of ct of
            (Const(@{const_name "heap_ex"},_) $ _) =>
            binder_conv
               (fn (ct',ctx') => loop_ex ctx' (term_of ct' :: bsubst))
               ctx ct
          | (@{const heap_conj} $ _ $ _) =>
            fun_conv (loop_ex ctx bsubst) ct
          | _ => loop_spatial ctx bsubst pos ct
  in
      loop_ex ctx [] then_conv
      norm_heap_conv
  end

fun norm_by_state_subst ctx ex_elims (State s) =
  let val ex_free_subst = (side_ex (#left s)~~ ex_elims)
      val _ = tr ctx (fn () =>
                let val ctx = Proof_Context.init_global Pure.thy
                in tracing ("free subst "^
                             commas (map (fn (a,b) =>
                                            string_of_term ctx a^
                                            " -> "^
                                            string_of_term ctx b) ex_free_subst))
              end)
  in
      Envir.norm_term (#env s)#>
      subst_free ex_free_subst
  end

fun instantiate_cterm_by_state_subst ctx ex_elims (State s) ct =
  let val norm = norm_by_state_subst ctx ex_elims (State s)
      val { t, thy_ref, ... } = rep_cterm ct
      val thy = Theory.deref thy_ref
      val Ts = Term.add_tvars t []
      val _ = if Ts <> []
              then error "Sorry, not done: instantiation of type unknowns in direct match"
              else ()
      val xs = map (fn x => pairself (cterm_of thy) (Var x, norm (Var x)))
                       (Term.add_vars t [])
  in
      Thm.instantiate_cterm ([],xs) ct
  end


(*
  All the proof execution tactics work on current
  ctx, ex-names, and heap premise. Then, they take
  the necessary information for their specific work
  from the given state.

  They take a continuation argument k which determines
  the next step to be invoked after the processing is done.
*)
fun direct_proof tac ctx xs hthm target_thm (State s) =
  exception_trace (fn () =>
  let val target = cprem_of target_thm 1
      val Q = target
         |> Thm.dest_arg |> Thm.dest_fun
         |> instantiate_cterm_by_state_subst ctx xs (State s)

      fun match_pos which =
        map (fn (a,b,_) => (#pos (which(a,b)))) (#eqs s)
      val hthm' = rearrange_heap_prem ctx (match_pos fst) hthm
      val _ = tr ctx (fn () => trace_list "rearrange P" [ pretty_thm ctx hthm' ])
      val _ = tr ctx (fn () => trace_list "Q" [ pretty_cterm ctx Q ])
      val h = cprop_of hthm |> Thm.dest_arg |> Thm.dest_arg
      val normQ = Drule.fun_cong_rule
                    (rearrange_conv ctx (match_pos snd) Q)
                    h
      val _ = tr ctx (fn () => trace_list "rearrange Q" [ pretty_cterm ctx (Thm.rhs_of normQ) ])
      val back = normQ COMP_INCR @{thm meta_eq_to_obj_eq} RS iffD2
      val hthm'' =
                (target_thm
             |> curry (op RS) back
             |> match_components ctx hthm' 1
             |> Seq.hd
             |> Drule.flexflex_unique) (* TODO: better to flatten hthm' like left/right *)
             handle Option =>
              fail_list "match_components" [ pretty_thm ctx hthm',
                                             pretty_thm ctx target_thm,
                                             pretty_thm ctx back ]
      val _ = tr ctx (fn () => trace_list "result of direct_proof" 
                        [ pretty_thm ctx hthm'' ])
  in
      tac ctx hthm''
      (* TODO: rtac back *)
  end)

(*
  A weaker variant, which merely rearranges the heap premise
  and inserts the re-arranged version. It orders the eqs of the
  state such that the "cancel" entries come first.
*)
fun rearrange_left ctx xs hthm (State s) thm =
  exception_trace (fn () =>
  let val Q = cprem_of thm 1
         |> Thm.dest_arg |> Thm.dest_fun

      fun order_elem (L,R) = LESS
        | order_elem (R,L) = GREATER
        | order_elem _ = EQUAL
      
      val order_path = Library.list_ord order_elem

      fun pref ((a,b,Frame), (a',b',Cancel)) = GREATER
        | pref ((a,b,Cancel), (a',b',Frame)) = LESS
        | pref ((a,b,Frame), (a',b',Frame)) = EQUAL
        | pref ((a,b,Cancel), (a',b',Cancel)) =
          order_path (#pos b, #pos b')
      
      val match_pos =
        map (fn (a,b,_) => (#pos a)) (sort pref (#eqs s))

      val hthm' = rearrange_heap_prem ctx match_pos hthm
      val _ = tr ctx (fn () => trace_list "rearrange P" [ pretty_thm ctx hthm' ])
  in
      (cut_facts_tac [ hthm' ] 1) thm
  end)

fun unfold_right_proof target_thm k ctx ex_elims hthm (State s) =
  exception_trace (fn () =>
  let open Conv
      val cert = cterm_of (Proof_Context.theory_of ctx)
      val cert_typ = ctyp_of (Proof_Context.theory_of ctx)
     val prep = norm_by_state_subst ctx ex_elims (State s)
     fun loop [] = all_tac         
       | loop ({pos,ex,thm} :: unfolds') =
         CSUBGOAL (fn (ct,i) =>
         let val app_eq =
                arg_conv (fun_conv (apply_unfold_conv ctx pos thm (#solved s)))
                            ct
             val _ = tr ctx (fn () => trace_list "unfold right" [
                          pretty_ex_terms ctx (map prep ex),
                          Pretty.block (Pretty.str "obtain yielded ex " ::
                                         (Pretty.separate " "
                                             (map (pretty_term ctx) ex_elims))),
                          pretty_thm ctx app_eq,
                          Pretty.block [ Pretty.str "to: ", pretty_cterm ctx ct ]
                     ])
             fun heap_exI t =
               rtac (instantiate' [ (SOME (cert_typ (Term.fastype_of t))) ]
                                  [ NONE, SOME (cert t) ] @{thm heap_exI}) 1
             fun eqTrue thm =
               thm COMP_INCR @{thm Eq_TrueI}

             val solved_ss =
                map eqTrue (filter (not o thm_proves @{cprop True})
                               (#solved s))
         in
             compose_tac (false,app_eq COMP_INCR Drule.equal_elim_rule2,1) 1 THEN
             EVERY (map (heap_exI o prep) ex) THEN
             TRY (rtac @{thm heap_conjI} 1 THEN
                  SELECT_GOAL
                    (InstDefs.instantiate_defs_tac ctx 1 THEN
                     IF_UNSOLVED
                      (asm_simp_tac (simpset_of ctx addsimps solved_ss) 1)) 2) THEN
             loop unfolds'
         end) 1
      val target_thm' =
         target_thm |> loop (#unfold (#right s))
                    |> Seq.hd
  in      
      k ctx ex_elims hthm target_thm' (State s)
  end)

fun unfold_left_proof k ctx xs hthm (State s)=
  exception_trace (fn () =>
  let open Conv
      val tr = tr ctx
      fun loop ctx [] xs hthm =
          k ctx xs hthm (State s)
        | loop ctx ({pos,ex,thm} :: unfolds') xs hthm =
          let val app_eq =
                 arg_conv (fun_conv (apply_unfold_conv ctx pos thm (#solved s)))
                           (cprop_of hthm)
              val _ = tr (fn () => trace_list "unfold left eq" [ pretty_thm ctx app_eq ])
              val hthm' = Thm.equal_elim app_eq hthm
              val _ = tr (fn () => trace_list "exec unfold step left"
                        [ pretty_thm ctx hthm,
                          pretty_thm ctx thm,
                          pretty_thm ctx hthm' ])
          in
              OBTAIN (K (REPEAT_DETERM (eresolve_tac [ @{thm heap_exE}, @{thm heap_conjE} ] 1))) [ hthm' ]
              (loop_step unfolds' xs) 1 ctx
          end
        and loop_step unfolds' xs ctx xs' prems =
          let val (hthm'' :: sc'') = prems
              val _ = tr (fn () => trace_list "OBTAIN yields" [
                        pretty_thm ctx hthm'',
                        Pretty.big_list "side-conds"
                            (map (pretty_thm ctx) sc'')
                     ])
          in
              loop ctx unfolds' (xs @ (map (term_of o snd) xs')) hthm''
          end
  in
      loop ctx (#unfold (#left s)) [] hthm
  end)

fun is_finished (State s) =
  let fun is_empty [] = true
        | is_empty _ = false
  in
      is_empty (#atoms (#left s)) andalso
      is_empty (#atoms (#right s)) andalso
      is_empty (#pending s)
  end

fun is_finished_partly (State s) =
  let fun is_empty [] = true
        | is_empty _ = false
  in
      not (is_empty (#eqs s)) andalso
      is_empty (#pending s)
  end

(* Drivers for top-level invocations *)

(* Match heaps in a forward fashion
   against a given target; the result is
   not applied immediately, but is passed on
   to a given tactic.
*)
fun match_heaps_fwd strategy ctx hthm prems target tac =
  let val (P,h,_) = the (dest_heap_prem hthm)
  in
  case term_of target of
      @{const Trueprop} $ (_ $ h') =>
      if term_of h = h'
      then let val target = flat_vars_ct target
               fun exec_proof (State s) =
                 let val _ = tr ctx (fn () => trace_list "Match result"
                               [ pretty_match_state ctx (State s) ])
                     val target_thm =
                           Thm.instantiate ([],[(@{cpat "PROP ?psi"}, target)]) @{thm "Pure.asm_rl"}
                 in
                     case Seq.pull (solve_pending ctx false (State s)) of
                         SOME(State s,_) =>
                         if is_finished_partly (State s)
                         then unfold_left_proof (unfold_right_proof target_thm (direct_proof tac))
                             ctx [] hthm (State s)
                         else (tr_fail ctx (fn () =>
                                              trace_list "discard unfinished state"
                                              [ pretty_match_state ctx (State s) ]);
                               no_tac)
                       | NONE =>
                         (tr_fail ctx (fn () =>
                              trace_list "discard with unsolvable side-conditions"
                                          [ pretty_match_state ctx (State s) ]);
                          no_tac)
                 end
               val Q = target |> Thm.dest_arg |> Thm.dest_fun
           in
               fn thm =>
                 Seq.maps (fn x => exception_trace (fn () => exec_proof x thm))
                   (strategy ctx (term_of P) (term_of Q) prems)
           end
      else no_tac
    | _ => no_tac
  end

fun protect_freeze thm =
  let val thy = theory_of_thm thm
      fun mk_inst (v as ((x,i),T)) =
        (cterm_of thy (Var v), 
         cterm_of thy (Free("_protect_"^x^":"^string_of_int i, T)))
      val inst = map mk_inst (Term.add_vars (prop_of thm) [])
  in
      Thm.instantiate ([],inst) thm
  end
fun protect_thaw thm =
  let val thy = theory_of_thm thm
      val frees = filter (String.isPrefix "_protect_" o fst)
                    (Term.add_frees (prop_of thm) [])
      fun var_of (nm,T) =
        unprefix "_protect_" nm
        |> String.tokens (curry op= #":")
        |> (fn [x,i] => (Var((x,the(Int.fromString i)),T))
              | xs => error ("Sorry, protect_thaw: "^commas xs))
  in  thm 
      |> Drule.forall_intr_list (map (cterm_of thy o Free) frees)
      |> Drule.forall_elim_list (map (cterm_of thy o var_of) frees)
  end
  
 fun match_heaps_tac_base strategy ctx hthm prems =
  CSUBGOAL
  (fn (ct,i) =>
     (match_heaps_fwd strategy ctx hthm prems ct 
       (fn ctx' => fn hthm' => 
           (* workaround: some internals of comp_no_flatten fail
              because of remaining flex-flex pairs *)
           rtac (protect_freeze hthm') 1) THEN
      PRIMITIVE protect_thaw)) 1
  
fun match_heaps_tac strategy ctx (g as {prems,...} : Subgoal.focus) hthm =
  match_heaps_tac_base strategy ctx hthm prems

(*
fun heap_expose_tac ctx (g as {prems,...} : Subgoal.focus) hthm ts thm =
  let val _ = tr ctx (fn () => trace_list "heap_expose" (map (pretty_term ctx) ts))
      fun exec_proof (State s) =
        let val _ = tr ctx (fn () => trace_list "Match result"
                      [ pretty_match_state ctx (State s) ])
        in
            if is_finished (State s)
            then unfold_left_proof rearrange_left ctx [] hthm (State s) thm
            else (tr_fail ctx (fn () =>
                                 trace_list "discard unfinished state"
                                 [ pretty_match_state ctx (State s) ]);
                  Seq.empty)
        end
      val P = heap_prem_spatial hthm
      val idx = #maxidx (rep_cterm P) + 1
      val (_,Q) = fold (fn t => fn (i,F) =>
                      (i+1, @{const star} $ (Const(@{const_name points_to},_) $ t $ Var(("v"^string_of_int i,idx), @{typ val})) $ F))
                    (rev ts) (1,Var(("R",idx), @{typ hassn}))
      val _ = tr ctx (fn () => trace_list "expose target" [ pretty_term ctx Q ])
  in
      Seq.maps (fn x => exception_trace (fn () => exec_proof x))
      (match_heaps ctx (term_of P) Q prems)
  end

fun heap_expose_tac_raw ts_raw ctx g hthm =
  heap_expose_tac ctx g hthm (map (Syntax.read_term ctx) ts_raw)
*)

(*============================================================
     top-level tactics
  ============================================================*)
local
open Method
in

val tactic_setup =
  Method.setup (Binding.name "fwd_disjoint")
    (Scan.lift (Scan.optional (Args.$$$ "raw" |-- Scan.succeed true) false) >>
        (fn raw => heap_meth (fwd_disjointness raw) false))
    "add disjointness list to premises" #>
  Method.setup (Binding.name "heap_sc")
    (Attrib.thms >> (fn thms => heap_meth (heap_sc_tac thms) false))
    "extract side-conditions from heap components" #>
  Method.setup (Binding.name "heap")
    (Scan.succeed (heap_meth (match_heaps_tac strategy_full) true))
    "match heaps in premise and conclusion" #>
  Method.setup (Binding.name "heap_cancel")
    (Scan.succeed (heap_meth (match_heaps_tac strategy_cancel) true))
    "match heaps in premise and conclusion; only trivial matches" #>
  Method.setup (Binding.name "heap_cleanup_stale")
    (Scan.succeed (fn ctx => Method.SIMPLE_METHOD (cleanup_heaps ctx 1)))
    "remove old (first) heap from premises after calls"

(* General setup *)
val setup =
  attr_setup #>
  tactic_setup #>
  Addr_Thms.setup #>
  Zero_Off_Thms.setup #>
  Theory.at_begin consolidate_data #>
  heap_trace_setup #>
  heap_fail_trace_setup #>
  disjoint_simproc_setup

end
end