structure PrepVC =
struct

open TraceAux

val (prep_vc_trace, prep_vc_trace_setup) =
   Attrib.config_bool @{binding "trace_prep_vc"} (fn _ => false);

fun tr ctx f=
  if Config.get ctx prep_vc_trace
  then (ignore (f ()); ())
  else ()

fun string_of_bool b = if b then "true" else "false"

fun sHeapOp e a =
  Const(@{const_name HeapOp},dummyT) $ e $ a

fun goal_concl_conv cv ctx =
  let open Conv
      fun loop vs ctx ct =
        case term_of ct of
            (Const("all",_) $ _) =>
            Conv.forall_conv (fn (v,ctx) => loop (v :: vs) ctx) ctx ct
          | _ => Conv.concl_conv ~1 (cv vs ctx) ct
  in 
      loop [] ctx 
  end

datatype naming = 
    Naming of { names : (term * term) list, next : int }

val heap_op_prefix = "_H_"

fun invent t (nm as (Naming { names, next })) =
  case AList.lookup (op aconv) names t of
      SOME x => (x, nm)
    | NONE =>
      let val next' = next + 1
          val x = Free(heap_op_prefix^string_of_int next', dummyT)
      in      
          (x, Naming { names = ((t,x) :: names), next = next' })
      end
      
fun empty_naming nxt = Naming { names = [], next = nxt }

fun is_op_ref x = 
  String.isPrefix heap_op_prefix x
  
fun heap_refs t = 
  let fun add_heap_ref (Free(x,@{typ heap})) r =
        if not (member(op=) r x)
        then x :: r
        else r
        | add_heap_ref _ r = r
  in
      Term.fold_aterms add_heap_ref t []
  end

(*
  Extracting accesses from a given VC yields a raw list
  of accesses accs and a term t' where the bound variables 
  refer to the accesses.
  
  Merging access means updating the bounds in the terms
  for an op$ node in the term.
*)
fun pretty_acc ctx (t,x) =
  Pretty.block [ pretty_term ctx x,
                 Pretty.str ": ",
                 pretty_term ctx t ]

fun pretty_names ctx (Naming { names, ...}) =
   Pretty.chunks
     (map (pretty_acc ctx) (rev names))

fun pretty_accs ctx msg (t,naming) =
  Pretty.big_list msg
    [ pretty_names ctx naming,
      Pretty.str "-----",
      pretty_term ctx t ]
        
type accs = ((term * term) list * term)

fun eq_acc ((x,t),(x',t')) =
  x aconv x' andalso t aconv t'

val string_of_refs = commas 
val pretty_refs = Pretty.str o string_of_refs

fun trace_refs ctx msg xs =
  (tr ctx (fn () => tracing (msg^":"^string_of_refs xs)))

fun extract_accs ctx (t : term) naming : (term * naming) =
  let fun is_heap_acc (Const(c,_)) =
        c = @{const_name "heap_put"} orelse
        c = @{const_name "heap_get"}
        | is_heap_acc _ = false
      val no_bounds = forall (null o loose_bnos)
  in
      case t of
          (c as Const(@{const_name HeapOp},_)) $ e $ a =>
          (* avoid extracting the whole of the heap op *)
          let val (f,args) = strip_comb e
              val (subs,naming) = fold_map (extract_accs ctx) (f :: args) naming
              val e' = Library.foldl1 (op $) subs
              val (a',naming) = extract_accs ctx a naming
          in
              (c $ e' $ a', naming)
          end
        | (_ $ _) =>
          let val (f,args) = strip_comb t
              val (subs,naming) = fold_map (extract_accs ctx) (f :: args) naming
              val t' = Library.foldl1 (op $) subs
          in
              if is_heap_acc f andalso no_bounds subs
              then invent t' naming
              else (t',naming)
          end
        | Abs(x,T,t) =>
          let val (t',naming) = extract_accs ctx t naming
          in
              (Abs(x,T, t'),naming)
          end
        | _ => (t,naming)
  end


(*
  The dependency graph between the different heap operations
  determines which operations must be scheduled before which.
  It is governed by two kinds of edges:
  (a) Argument required for evaluation (subterm relationship),
      so that all arguments must be scheduled first
  (b) A heap_get ... H must be scheduled before a heap_put .. H,
      because otherwise the symbolic execution would destroy H
      before the get can be executed.
      
  The graph has an edge H -> H' if H' must be scheduled before H.
  
  The dependency graph includes all frees that occur as arguments
  to heap ops as well, in order to compute the depencency on guarded
  values in a simple mprettyanner.
*)
structure SG = Graph(type key=string; val ord = string_ord)

fun pretty_op ctx sg x =
  Pretty.block [ Pretty.str x,
                 Pretty.str ": ",
                 pretty_term ctx (SG.get_node sg x) ]

fun pretty_sched_graph ctx g =
  let val lin = filter is_op_ref (SG.topological_order g)
      fun pretty_edges ((x,_),xs) =
        if is_op_ref x 
        then SOME (Pretty.str (x^" -> "^commas xs))
        else NONE
  in
      Pretty.chunks 
      (map (pretty_op ctx g) (rev lin) @
       [ Pretty.str "----" ] @
       map_filter pretty_edges (SG.dest g))
  end

fun is_heap_put (Const(@{const_name heap_put},_) $ _ $ _ $ _ $ _) = true
  | is_heap_put t = false

fun scheduling_graph ctx (Naming { names, ...}) =
  let fun node (Free(H,_)) = H
        | node t = error ("node: "^Syntax.string_of_term ctx t)
  
      val names = map (apsnd node) names
      val frees = 
         distinct (op=)
           (filter (not o is_op_ref o fst) 
               (fold Term.add_frees (map fst names) []))
      
      fun subterm_edges (t,h) =
         map (fn r => (h,r)) (Term.add_free_names t [])
      
      fun heap_put_edges ((Const(@{const_name heap_put},_) $ _ $ _ $ _ $ h),x) =
          map_filter
          (fn ((Const(@{const_name heap_get},_) $ _ $ _ $ h'),x') =>
              if h = h'
              then SOME (x,x')
              else NONE
             | _ => NONE)
          names
        | heap_put_edges _ = []
      
      val g1 = 
         SG.empty 
         |> fold (SG.new_node o swap) names
         |> fold SG.new_node (map (fn (x,T) => (x, Free(x,T))) frees)
         |> fold SG.add_edge (flat (map subterm_edges names))
      val g2 = g1
         |> fold SG.add_edge (flat (map heap_put_edges names))
      
      val _ = tr ctx (fn () => trace_list "scheduling graph"
                          [ pretty_sched_graph ctx g2 ])
  in
     (g1,g2)
  end

fun is_empty []  = true
  | is_empty _ = false

(*
  Computing the info about heap and other references in a term,
  including guarding. A heap op, represented by its result name,
  cannot be executed in a term immediately.
*)
type ref_info =
  { op_refs    : string list,
    guarded    : string list, (* heap_gets *)
    forbidden  : string list, (* destructive heap_puts *)
    before_ins : string list
  }
  
fun pretty_ref_info msg ri =
  Pretty.big_list msg
     [ Pretty.str ("op-refs    : "^commas (#op_refs ri)),
       Pretty.str ("guarded    : "^commas (#guarded ri)),
       Pretty.str ("forbidden  : "^commas (#forbidden ri)),
       Pretty.str ("before_ins : "^commas (#before_ins ri)) ]

fun  merge_ref_infos ctx sg_sub sg_put t [a] = a
  | merge_ref_infos ctx sg_sub sg_put t [a',b'] =
  let fun merge f a b = distinct (op=) (f a @ f b)
      val sub = subtract (op=)
      val nodes = SG.keys sg_sub
      val node = SG.get_node sg_sub
      
  in case t of
         (@{const HOL.implies} $ a $ _)=>      
         let val a_frees= Term.add_free_names a []
             fun is_guarded x =
              (* Some reference x is guarded if one of its
                 dependants is guarded, because then it should not
                 be executed, because the op itself (transitively)
                 dependends on some value that may still be restricted.
                 Note: no push dependencies here, only value computation. 
                 Note: the frees of heap ops are in the graph and recorded
                 as dependents there. *)
              let val res =
              not (null (inter (op=)
                           (SG.all_succs sg_sub
                               (SG.immediate_succs sg_sub x))
                           a_frees))
              in
                 (*if res 
                 then 
                   Output.writeln ("Guarded: "^x^" by "^commas a_frees)                                      
                 else (); *)
                 res
              end
         in
             { op_refs = merge #op_refs a' b',
               guarded = distinct (op=)
                           (filter is_guarded nodes @
                               #guarded a' @
                               #guarded b'),
               forbidden = merge #forbidden a' b',
               before_ins = []
             }
         end
       | (@{const HOL.conj} $ _ $ _) =>
         let fun unguarded {op_refs,guarded,...} =
               sub guarded op_refs
         in
             { op_refs = merge #op_refs a' b',
               guarded = 
               (* a reference is not guarded any longer if it occurs
                  unguarded in one of the branches, because then it will
                  be executed anyway. *)
                distinct (op=)
                  (sub (unguarded b') (#guarded a') @
                   sub (unguarded a') (#guarded b')),
               forbidden = merge #forbidden a' b',
               before_ins = [] }
         end
       | t => error("merge-ref-infos:"^string_of_term ctx t)
  end
  
fun compute_ref_info ctx sg_sub sg_put t =
  let val nodes = SG.keys sg_sub
      val node = SG.get_node sg_sub
      
      fun merge f a b = distinct (op=) (f a @ f b)
      val sub = subtract (op=)
      
      val empty =
         { op_refs = [],
           guarded = [],
           forbidden = [],
           before_ins = []
         }
      
      (* For a directly required heap x, guard those heap_puts that
         would (transitively) destroy that heap during symbolic execution.
         Note, however, that gets will always be scheduled correctly,
         so they do not lead to any guarding of puts. Only direct heap references
         in the formula do. *)
      fun guarded_puts x =
        filter (is_heap_put o node) 
               (SG.all_preds sg_put (SG.immediate_preds sg_put x))
      
      fun loop (t as (@{const HOL.implies} $ a $ b)) =
          merge_ref_infos ctx sg_sub sg_put t [loop a,loop b]
        | loop (t as (@{const HOL.conj} $ a $ b)) =
          merge_ref_infos ctx sg_sub sg_put t [loop a,loop b]
        | loop (Free(x,@{typ heap})) =
            { op_refs = if is_op_ref x then [x] else [],
              guarded = [],
              forbidden = guarded_puts x,              
              before_ins = if is_op_ref x then [x] else []
            }
        | loop (Free(x,T)) =
          if is_op_ref x 
          then { op_refs = [x],
                 guarded = [],
                 forbidden = [],
                 before_ins = [x]
               }
          else empty
        | loop (a $ b) =
        let val a' = loop a
            val b' = loop b
        in
            if a' = empty andalso b' = empty
            then empty
            else
            { op_refs = merge #op_refs a' b',
              guarded = merge #guarded a' b',
              forbidden = merge #forbidden a' b',
              before_ins = merge #before_ins a' b'
            }
        end
        | loop (Abs(x,T,t)) = loop t
        | loop _ = empty
  in      
      loop t
  end
  
fun insert_accs ctx (t,naming) =
  let val (sg_sub,sg_put) = scheduling_graph ctx naming
      val sg_nodes = SG.keys sg_put
      val sg_node = SG.get_node sg_sub
      
      fun add_heap_op x a =
        sHeapOp (sg_node x) (absfree (x,dummyT) (incr_boundvars 1 a))
      
      val complete_refs_put  = distinct (op=) o SG.all_succs sg_put
      val complete_refs_subs = distinct (op=) o SG.all_succs sg_sub
      
      fun add_ops msg [] xs t k = k xs
        | add_ops msg sel xs t k =
        let val sel = distinct (op=) sel
            fun is_sel x = member (op=) sel x            
            val ys = filter is_sel xs
            val zs = filter (not o is_sel) xs
            
            val _ = tr ctx 
                      (fn () => 
                         if ys <> []
                         then trace_list ("chopping/"^msg)
                             (map (pretty_op ctx sg_sub) ys @
                              [ Pretty.big_list "at" 
                                  [ pretty_term ctx t ] ])
                         else ())
        in
            fold add_heap_op (rev ys) (k zs)
        end

      (* A heap-op is guared if some consituent of the op is
         subject to an additional assumption further down the tree.
         A heap_put is also guarded if its original heap is used 
         somewhere down the tree, because then symbolic execution would
         destroy the original heap. *)
      fun drop_guarded ri xs =
        filter (not o member (op=) (#guarded ri @ #forbidden ri)) xs
      
      (*
        References that are not actually required. Note that here
        only the the subterm order is important: a put must not to
        all gets on the heap to be executed.
      *)
      fun keep_required ri =
        let val req = SG.all_succs sg_sub (#op_refs ri)            
        in 
            filter (member (op=) req)
        end
            
      (* loop through the tree to locate possible insertion points *)
      fun loop [] t = t
        | loop xs (t as (@{const HOL.conj} $ _ $ _)) = ins_pt xs t
        | loop xs (t as (@{const HOL.implies} $ _ $ _)) = ins_pt xs t
        | loop xs (t as (Const(@{const_name Ex},_) $ _)) = ins_pt xs t
        | loop xs (a $ b) = loop xs a $ loop xs b
        | loop xs (Abs(x,T,t)) = Abs(x,T,loop xs t)
        | loop xs t = t
      
      (* Determine those that need to be inserted here, rather than
         further down, in order to avoid breaking the sharing or
         loosing referenced bound values. 
         
         Thus, we insert those values that are shared between
         different arguments.
      *)
      and ins_pt [] t = t
        | ins_pt xs t =
          let val _ = tr ctx (fn () =>
                                trace_list "ins_pt"
                                [ Pretty.str (commas xs),
                                  pretty_term ctx t ])
              val (f,args) = strip_comb t
              val ref_info = compute_ref_info ctx sg_sub sg_put 
              val ris = map ref_info args
              val ri  = merge_ref_infos ctx sg_sub sg_put t ris
              val _ = tr ctx (fn () =>
                               trace_list "ref-info"
                                 (pretty_ref_info "ref-info" ri ::
                                   (map_index (fn (i,x) =>
                                        pretty_ref_info (string_of_int (i+1)) x)
                                       ris)))
              
              fun add_imp_lhs xs =
                if f aconv @{const HOL.implies}
                then let val lhs_refs = #op_refs (ref_info (hd args))
                     in tr ctx (fn () =>
                                  trace_refs ctx "adding implies LHS to shared" lhs_refs);
                       lhs_refs @ xs
                     end
                else xs
              
              fun tx m rs = (trace_refs ctx m rs; rs)
              
              val scheduled = map #op_refs ris
                      (* Note: nothing is decided as yet,
                         so ignore put-dependencies for the moment *)
                          |> map complete_refs_subs
                          |> foldl1 (uncurry (inter (op=)))
                          |> tx "inter"
                          |> add_imp_lhs
                          |> complete_refs_put (* add the put-dependencies now *)
                          |> inter (op=) xs              
                          |> tx "completed with push"
                          |> drop_guarded ri
                          |> tx "without guarded"
                          |> keep_required ri
                             
              val _ = tr ctx (fn () => trace_refs ctx "scheduled" scheduled)
          in
              ins_remaining_free xs
              (add_ops "scheduled" scheduled xs t
                  (fn zs => list_comb (f, 
                                   map_index (fn (i,t) =>
                                                ins_force (nth ris i) zs t)
                                      args)))
          end
      
      (* Force those that are required in the subterm
         before the next ins_pt occurs. *)
      and ins_force ri [] t = t
        | ins_force ri xs t =      
          let val force =
                      #before_ins ri
                    |> complete_refs_put
                    |> inter (op=) xs
              val _ = tr ctx (fn () =>
                                trace_list "ins_force/1"
                                [ pretty_term ctx t,
                                  pretty_refs force,
                                  pretty_ref_info "ref info" ri ])
              val force = force
                 |> drop_guarded ri
                 |> keep_required ri
              val _ = trace_refs ctx  "ins_force/after drop" force
          in
              ins_remaining_free xs
              (add_ops "force" force xs t 
               (fn zs => loop zs t))
          end
      
      (* Catch-all clause: a term that after top-down scheduling
         still contains free non-scheduled refs needs to have these
         insertet. This is better than ins_force. *)
      and ins_remaining_free xs t =
         let val scheduled = Term.add_frees t [] |> map fst
                           |> filter is_op_ref
                           |> inter(op=) xs
             val _ = if scheduled <> []
                     then trace_list ("ins_remaining "^commas xs)
                         [ pretty_term ctx t ]
                     else ()
         in
            add_ops "remaining" scheduled xs t
                 (fn _ => t)
         end
  in
      loop (filter is_op_ref (rev (SG.topological_order sg_put))) t
  end

fun below_binders_conv cv ctx =
   (Conv.binder_conv (fn (_,ctx) => below_binders_conv cv ctx) ctx else_conv
    cv ctx)

fun lin_conv ps ctx ct =
  let open Conv
      val contract = Drule.beta_eta_conversion ct
      val ct = Thm.rhs_of contract
      val _ = trace_list "lin_conv" [ pretty_cterm ctx ct ]
      fun conv_step ctx ct =
         let val next_free_name =
               (* TODO: count only the _H_ entries, but Abs_conv
                  does not preserve the bound name, but replaces it by u *)
               length (filter (curry (op=) "u" o fst)
                        (Variable.dest_fixes ctx))
             
             val accs = extract_accs ctx (term_of ct) (empty_naming next_free_name)
             val _ = tr ctx (fn () => trace_list "acc_list" [ pretty_accs ctx "extracted" accs ])
             val t' = insert_accs ctx accs
             val _ = tr ctx (fn () => trace_list "with accessors" [ pretty_term ctx t' ])
             val cert = cterm_of (Proof_Context.theory_of ctx)
                          o singleton (Type_Infer_Context.infer_types ctx)
             val g = Const("==",dummyT) $ term_of ct $ t'
             val g' = cert g
             val _ = (fn () => trace_list "goal" [ pretty_cterm ctx g' ])
             val eq = Goal.prove_internal [] g' 
                        (K (simp_tac (HOL_basic_ss addsimps @{thms heap_op_unfold}) 1))
             handle THM _ => error ("Failed to prove lin_conv equality\n"^
                                   Syntax.string_of_term ctx g)
         in
             eq
         end
      fun conv_sel ctx ct =
        case term_of ct of
            Const(@{const_name All},_) $ P =>
            below_binders_conv conv_step ctx ct
          | _ => all_conv ct
      
      val eq = (conv_step ctx then_conv
                sub_conv (top_conv conv_sel) ctx) ct;

      (* val _ = (fn () => trace_list "thm" [ pretty_thm ctx eq ]) *)
      
  in
      Thm.transitive contract eq
  end

(*
 - Extract the heap operations 
 - Prune obviously contradictory if/while conditions
 
*)
fun prep_vc_tac ctx =
  ALLGOALS (CONVERSION (goal_concl_conv lin_conv ctx) THEN'
            simp_tac (Simplifier.context ctx HOL_basic_ss))

val setup =
    Method.setup (Binding.name "prep_vc")
    (Scan.succeed (fn ctx => SIMPLE_METHOD (prep_vc_tac ctx)))
    "prepare VC for separation logic treatment" #>
   prep_vc_trace_setup

end