structure SourceImport =
struct

open Ast

val mk_str = HOLogic.mk_string
fun mk_bool true = @{term True}
  | mk_bool false = @{term False}

val mk_int = HOLogic.mk_number @{typ int}
val mk_list = HOLogic.mk_list

fun mk_prim OP_PLUS = @{term OP_PLUS}
   | mk_prim OP_MINUS = @{term OP_MINUS}
   | mk_prim OP_MUL = @{term OP_MUL}
   | mk_prim OP_DIV = @{term OP_DIV}
   | mk_prim OP_LEQ = @{term OP_LEQ}   
   | mk_prim OP_GEQ = @{term OP_GEQ}   
   | mk_prim OP_EQUAL = @{term OP_EQUAL}   
   | mk_prim OP_NEQ = @{term OP_NEQ}   
   | mk_prim OP_GREATER = @{term OP_GREATER}   
   | mk_prim OP_LESS = @{term OP_LESS}   

fun mk_pre_post OP_INCR = @{term OP_INCR}
  | mk_pre_post OP_DECR = @{term OP_DECR}

fun tr_ty TVoid = @{term TVoid}
  | tr_ty TInt = @{term TInt}
  | tr_ty TUInt = @{term TUint}
  | tr_ty TChar = @{term TChar}
  | tr_ty (TPtr ty) = @{term TPtr} $ tr_ty ty
  | tr_ty (TStruct name) = @{term TStruct} $ mk_str name
  | tr_ty (TArray (ty,len)) =
    (case len of
         NONE => @{term TArray} $ tr_ty ty
       | SOME len => @{term TArrayN} $ tr_ty ty $ mk_int len)
  | tr_ty _ = error "unknown ty constructor"

fun tr_exp (EVar v) = @{term EVar} $ mk_str v
  | tr_exp (ECchar c) = @{term ECchar} $ HOLogic.mk_char (Char.ord c)
  | tr_exp (ECint i) = @{term ECint} $ mk_int i
  | tr_exp (ECuint int) = error "sorry: ECuint not done"
  | tr_exp (ENull) = @{term ENull} 
  | tr_exp (EDeref (e)) = @{term EDeref} $ tr_exp e
  | tr_exp (EAddr e) = @{term EAddr} $ tr_exp e
  | tr_exp (EDot(e,f)) = @{term EDot} $ tr_exp e $ mk_str f
  | tr_exp (EArrow(e,f)) = @{term EArrow} $ tr_exp e $ mk_str f
  | tr_exp (EIndex(e,i)) = @{term EIndex} $ tr_exp e $ tr_exp i
  | tr_exp (EAssign(a,b)) = @{term EAssign} $ tr_exp a $ tr_exp b
  | tr_exp (EPre(p,e)) = @{term EPre} $ mk_pre_post p $ tr_exp e
  | tr_exp (EPost(p,e)) = @{term EPost} $ mk_pre_post p $ tr_exp e
  | tr_exp (EPrim1(p,e)) = @{term EPrim1} $ mk_prim p $ tr_exp e
  | tr_exp (EPrim2(p,a,b)) = @{term EPrim2} $ mk_prim p $ tr_exp a $ tr_exp b
  | tr_exp (EAnd(a,b)) = @{term EAnd} $ tr_exp a $ tr_exp b
  | tr_exp (EOr(a,b)) = @{term EOr} $ tr_exp a $ tr_exp b
  | tr_exp (EFree(e)) = @{term EFree} $ tr_exp e
  | tr_exp (EFreeArray(a,b)) = error "not done: EFreeArray"
  | tr_exp (ENew ty) = @{term ENew} $ tr_ty ty
  | tr_exp (ENewArray(ty,len)) = @{term ENewArray} $ tr_ty ty $ tr_exp len
  | tr_exp (ECall(f,args)) = @{term ECall} $ mk_str f $ mk_list @{typ exp} (map tr_exp args)
  | tr_exp _ = error "unknown exp constructor"

fun tr_decl (name,ty) = HOLogic.mk_prod (mk_str name, tr_ty ty)

fun tr_sm SSkip = @{term SSkip}
  | tr_sm (SSeq(a,b)) = @{term SSeq} $ tr_sm a $ tr_sm b
  | tr_sm (SExec e) = @{term SExec} $ tr_exp e
  | tr_sm (SBlock(decls,sm)) = 
       @{term SBlock} $ mk_list @{typ "string * ty"} (map tr_decl decls) $ tr_sm sm
  | tr_sm (SIf(e,a,b)) = @{term SIf} $ tr_exp e $ tr_sm a $ tr_sm b
  | tr_sm (SWhile(_,e,a)) = @{term SWhile} $ tr_exp e $ tr_sm a
  | tr_sm SBreak = @{term SBreak}
  | tr_sm SContinue = @{term SContinue}
  | tr_sm (SReturn e) = @{term SReturn} $ tr_exp e
  | tr_sm _ = error "unknown sm constructor"
               
fun import_gen start tr ty (id,text) lthy =
  let val ast = 
         tr (SimplC.parseString start text)
      val term = singleton (Type_Infer_Context.infer_types lthy) ast
      val eq = Logic.mk_equals (Free(id,ty), term)
      val ((_,(_,def)),lthy) = Specification.definition'
                    (SOME(Binding.name id, SOME ty, NoSyn), 
                     ((Binding.name (id^"_def"), []), eq)) false lthy
      val _ = writeln (Display.string_of_thm lthy def)
  in
      lthy
  end handle (Pos.ParseError(pos, msg)) =>
     error (Pos.fmt pos^":"^msg)
  
val import_sm =
  import_gen SimplC.startSymbolSM (fn(SM sm) => tr_sm sm) @{typ sm}

val import_exp =
  import_gen SimplC.startSymbolEXP (fn(EXP sm) => tr_exp sm) @{typ exp}
  
val _ =
  Outer_Syntax.local_theory ("source_sm", Keyword.thy_decl) 
    "import a source statement"
    (Parse.short_ident -- (Parse.$$$ "=" |-- Parse.text)
        >> import_sm);

val _ =
  Outer_Syntax.local_theory ("source_exp", Keyword.thy_decl) 
    "import a source expression"
    (Parse.short_ident -- (Parse.$$$ "=" |-- Parse.text)
        >> import_exp);

end