structure SymExec =
struct

open TraceAux

fun pretty_term_bounded ctx n t =
  let fun cut 0 _ = Free("?",dummyT)
        | cut n ((b as (Const _)) $ Abs(x,T,t)) =
          b $ Abs(x,T,cut n t)
        | cut n (t as (_ $ _)) =
          let val (f,args) = Term.strip_comb t
          in
              Term.list_comb (cut (n-1) f, map (cut (n-1)) args)
          end
        | cut n (Abs(x,T,t)) = Abs(x,T,cut n t)
        | cut n t = t
  in
      pretty_term ctx (cut n t)
  end
  
open Heaps

val (sym_exec_trace, sym_exec_trace_setup) =
   Attrib.config_bool @{binding "trace_sym_exec"} (fn _ => false);

fun tr ctx f=
  if Config.get ctx sym_exec_trace
  then (ignore (f ()); ())
  else ()

fun tr_conv ctx msg ct =
  (tr ctx (fn () => trace_list msg [ pretty_cterm ctx ct ]);
   Conv.all_conv ct)

fun tr_focus msg (f as {context, params, prems,...} : Subgoal.focus) =
  (tr context
      (fn () => 
         trace_list msg
         [ Pretty.block (Pretty.str "params: " :: 
                            map (pretty_cterm context o snd) params),
           Pretty.big_list "prems:"
           (map (pretty_thm context) prems) ]);
   f)

fun tr_goal ctx msg i thm =
  (tr ctx
      (fn () => 
         if nprems_of thm < i
         then tracing (msg^": no goal "^string_of_int i)
         else trace_list msg [Syntax.pretty_term ctx (nth (Thm.prems_of thm) (i - 1))]);
   Seq.single thm)

(*
  The recursively computed state
*)
type step_state = {
  ctx : Proof.context,
  params : (string * cterm) list,
  prems  : thm list,
  heap   : thm, (* the special current heap premise among the prems *)
  ss : simpset
}

fun state_init_focus ({context,params=params',prems=prems',...} : Subgoal.focus) =
   {ctx = context,params = params',prems=prems',
    heap = case the_heap_prem prems' of
               NONE => error ("Start run with exactly one heap premise")
             | SOME (_,_,thm) => thm,
    ss = Simplifier.add_prems prems' (simpset_of context addsimps prems') } : step_state

fun state_init_empty context =
   {ctx = context,params = [],prems=[],
    heap = @{thm TrueI},
    ss = simpset_of context } : step_state


fun state_map_ctx f ({ctx,params,prems,heap,ss} : step_state) =
  {ctx=f ctx,params=params,prems=prems,heap=heap,ss=ss} : step_state

fun state_map_params f ({ctx,params,prems,heap,ss} : step_state) =
  {ctx=ctx,params=f params,prems=prems,heap=heap,ss=ss} : step_state

fun state_map_prems f ({ctx,params,prems,heap,ss} : step_state) =
  {ctx=ctx,params=params,prems=f prems,heap=heap,ss=ss} : step_state

fun state_map_heap f ({ctx,params,prems,heap,ss} : step_state) =
  {ctx=ctx,params=params,prems=prems,heap=f heap,ss=ss} : step_state

fun bind_Ex x t = Const(@{const_name heap_ex},dummyT) $ Abs(x,dummyT,t)
fun bind_Exs xs t = fold bind_Ex (rev xs) t

fun points_to_pat ctx ct p h =
  ((@{const star} 
      $ (Const(@{const_name points_to},dummyT) $ p $ ct $ Var(("v",0),dummyT))
      $ Var(("R",0),dummyT))
      $ h)
  |> singleton (Type_Infer_Context.infer_types ctx)
  
(*
fun match_state_heap pat s =
  let val ctx = #ctx s
      val (_,h,_) = the (dest_heap_prem (#heap s))
      val cpat = Thm.capply @{cterm Trueprop}
                   (Thm.capply (cterm_of (Proof_Context.theory_of (#ctx s)) pat) h)
      val _ = trace_list "match_state_heap"
                [ pretty_cterm ctx cpat ]
      val thm = SOME (Goal.prove_internal [] cpat
                      (fn _ => match_heaps_tac_base ctx (#heap s) (#prems s) ORELSE
                               (* produce a nicer error message, saying that
                                  the goal could not be proven *)
                               all_tac))
                handle THM _ => error ("Failed to match heap\n"^
                                        Display.string_of_thm ctx (#heap s)^"\n==>\n"^
                                        Syntax.string_of_term ctx (term_of cpat))
                                          
      (*
      val _ = trace_list "match_state result" [ 
                 case thm of
                     NONE => Pretty.str "--"
                   | SOME thm => pretty_thm ctx thm ]
      *)
  in
      case thm of
          NONE => NONE
        | SOME heap' => SOME (state_map_heap (K heap') s)
  end

fun match_state_points_to ct p h s =
  match_state_heap (points_to_pat (#ctx s) ct p h) s
*)

(*
  Use simplifier re-orienting to make the addition of the prems'
  over the prems safe.
*)
fun reorient_prems ctx prems prems' =
  let val (_, { mk_rews = {reorient, ...}, ... }) = Raw_Simplifier.internal_ss (simpset_of ctx)
      val thy = Proof_Context.theory_of ctx
      fun check thm =
        case prop_of thm of
            @{const Trueprop} $ (Const(@{const_name HOL.eq},_) $ lhs $ rhs) =>
            (* TODO: reorient logic *)
            thm
          | _ => thm
  in
      map check prems'
  end

fun safe_ss ctx prems prems' = 
  Simplifier.add_prems prems'
    (simpset_of ctx addsimps 
        (reorient_prems ctx prems prems'))

(*============================================================================*)

(*
  Case-splitting on a disjunction leaves the original disjunction in
  the premises when FOCUS returns.
*)
fun dest_disj t =
  let fun loop (Const ("HOL.disj", _) $ t $ t') r =
        loop t (loop t' r)
        | loop t r = t :: r
  in
      rev (loop t [])
  end

fun cleanup_split_disj ctx i =
  COND (fn thm => nprems_of thm >= i)
  (SUBGOAL (fn (t,i) =>
    let val params = Logic.strip_params t
        val hyps   = map_filter (try HOLogic.dest_Trueprop) (Logic.strip_assums_hyp t)
        fun to_be_cleaned t =
          case t of
              disj as (@{const HOL.disj} $ _ $ _) =>
              exists (fn d => member (op aconv)hyps d)
                     (dest_disj disj)
            | _ => false
        val thinable = map_filter I
                        (map_index (fn (i,h) =>
                                      if to_be_cleaned h
                                      then SOME (i + 1) else NONE) hyps)
    in
        thin_indexed_tac (take (length thinable) thinable) i
    end) i)
    all_tac



(*
  Re-focussing is a fundamental operation. We have to make sure that
  any arising equalities are applied to all existing premises.
*)
fun state_add_focus ({ctx,params,prems,heap,ss} : step_state)
  ({context,params=params',prems=prems',...} : Subgoal.focus) =
  let val ss' = safe_ss ctx prems prems'
      val simp = Simplifier.simplify ss'
      fun prep thm =
        filter (not o thm_proves @{cprop True}) (conj_elims [thm])
  in 
      {ctx = context,
        params = params @ params',
        prems = map simp prems @ flat (map prep prems'),
        heap = case the_heap_prem prems' of
                   NONE => 
                   if Thm.eq_thm_prop (heap,@{thm TrueI})
                   then error "Failed to retrieve the head premise"
                   else heap
                 | SOME (_,_,thm) => thm,
               ss = ss'  } : step_state
  end

fun with_heap_pattern pat s tac else_tac =
  let val target = cterm_of (Proof_Context.theory_of (#ctx s)) (@{const Trueprop} $ pat)
      fun tac' ctx' hthm' =
        tac (s |> state_map_heap (K hthm') |> state_map_ctx (K ctx'))
  in
      match_heaps_fwd strategy_full (#ctx s) (#heap s) (#prems s) target tac' ORELSE
      else_tac s
  end


fun with_points_to ct p h s =
  with_heap_pattern (points_to_pat (#ctx s) ct p h) s

  (*
  case match_state_points_to ct p h s of
      NONE => no_tac
    | SOME s' => 
      (trace_list "with_points_to" [ pretty_thm (#ctx s) (#heap s') ];
      tac s')
*)
val heap_op_unfold_thm = @{thm heap_op_unfold} RS eq_reflection
val heap_get_resolve_thm = @{thm heap_get_resolve_simp} RS eq_reflection

fun beta_eta_contraction_rule th =
  Thm.equal_elim (Drule.beta_eta_conversion (cprop_of th)) th

fun strip_Ex (Const(@{const_name Ex},_) $ Abs(x,T,t)) =
  let val (exs,t') = strip_Ex t
  in
      ((x,T)::exs, t')
  end
  | strip_Ex t = ([],t)

val imp_TrueI = @{lemma "P ==> True --> P" by simp}
val imp_FalseI = @{lemma "False --> P" by simp}

fun steps_tac ctx n conj_follow_both =
  let open Subgoal
      open Conv
      fun loop 0 _ = (tracing ("breaking loop after "^string_of_int n^" steps"); 
                      no_tac)
        | loop n s =
          (case get_first (fn thm => 
                            if thm_proves @{cprop False} thm
                            then SOME thm
                            else NONE) (#prems s) of
              SOME thm => rtac (thm RS @{thm FalseE}) 1
         | NONE =>
           SUBGOAL (fn(t,i) =>
           (tr ctx (fn () => trace_list "loop on" [ pretty_term_bounded ctx 5 t ]);
            case #2 (dest_comb t) of
              @{const HOL.conj} $ (Const(@{const_name heap_acc_cond},_) $ ct $ p $ h) $ _ =>
              with_points_to ct p h s
              (fn s' =>
                 rtac @{thm conjI} 1 THEN
                 rtac @{thm heap_acc_condI} 1 THEN
                 rtac (#heap s') 1 THEN
                 (* don't count side-condition as a step *)
                 TRY (SELECT_GOAL (loop n s') 1))
              (split_cases n)
            | Const(@{const_name heap_acc_cond},_) $ ct $ p $ h =>
              with_points_to ct p h s
              (fn s' =>
                 rtac @{thm heap_acc_condI} 1 THEN
                 rtac (#heap s') 1)
              (split_cases n)
            | Const(@{const_name HeapOp},_) 
                 $ (Const(@{const_name heap_get},_) $ ct $ p $ h) $ _ =>
              with_points_to ct p h s 
              (fn s' =>
                 CONVERSION
                 (arg_conv 
                     ((arg1_conv (rewr_conv (#heap s' RS heap_get_resolve_thm)))
                         then_conv
                      fun2_conv (rewr_conv heap_op_unfold_thm))) 1 THEN
                 PRIMITIVE beta_eta_contraction_rule THEN
                 TRY (SELECT_GOAL (loop (n -1) s') 1))
              (split_cases n)
            | Const(@{const_name HeapOp},_) 
                 $ (Const(@{const_name heap_put},_) $ ct $ p $ v $ h) $ _ =>
              with_points_to ct p h s
              (fn s' =>
                 rtac @{thm heap_put_forward_heap_op} 1 THEN
                 rtac (#heap s') 1 THEN
                 simp_tac (simpset_of (#ctx s)) 1 THEN (* solve cty_...ok *)
                 re_focus n s' 1)
              (split_cases n)
            | @{const HOL.conj} $ _ $ _ =>
              if conj_follow_both
              then (rtac @{thm conjI} THEN_ALL_NEW
                    TRY o SELECT_GOAL (loop (n -1) s)) i
              else (rtac @{thm conjI} THEN'
                    TRY o SELECT_GOAL (loop (n -1) s)) i
            | @{const HOL.implies} $ _ $ _ =>
              CONVERSION (arg_conv (arg1_conv (Simplifier.rewrite (#ss s)))) 1 THEN             
              (rtac imp_FalseI 1 ORELSE
               ((rtac imp_TrueI 1 ORELSE
                 rtac @{thm impI} 1) THEN
                re_focus n s 1))
            | (P as (Const(@{const_name Ex},_) $ _)) =>
              (* a function call takes this form *)
              let val (exs,H) = apsnd (hd o HOLogic.dest_conj)
                                  (strip_Ex P)
              in
                  if Heaps.is_heap_prop [] (@{const Trueprop} $ H)
                  then let val idx = maxidx_of_term H + 1
                           val H' = subst_bounds (map (fn (x,T) => Var((x,idx),T))
                                                     (rev exs), H)
                       in
                           with_heap_pattern H' s
                           (fn s' =>
                              REPEAT_DETERM (rtac @{thm exI} 1) THEN
                              TRY (rtac @{thm conjI} 1) THEN
                              rtac (#heap s') 1)
                           (K no_tac)
                       end
                  else no_tac
              end
            | (Const(@{const_name All},_) $ _) =>
              (* after the call *)
              print_tac "X" THEN
              REPEAT_DETERM (rtac @{thm allI} 1) THEN
              print_tac "Y" THEN
                rtac @{thm impI} 1 THEN
               print_tac "Z" THEN
              clarify_tac (#ctx s) 1
            | t => 
                (tr ctx (fn () => 
                          trace_list "break sym exec at"
                           [ pretty_term_bounded ctx 6 t ]);
                 no_tac))) 1)
      
        and re_focus n s =
           safe_asm_full_simp_tac (simpset_of (#ctx s)) THEN'
           IF_UNSOLVED o FOCUS (TRY o re_focus_base (n - 1) s) (#ctx s)
        and re_focus1 n s =
           safe_asm_full_simp_tac (simpset_of (#ctx s)) THEN'
           IF_UNSOLVED o FOCUS (re_focus_base (n - 1) s) (#ctx s)
        and re_focus_base n s = loop n o state_add_focus s
      
        and split_cases n s =
          no_tac
      (*
          (* idea: introduce extra parameter to "guess" heap splits *)
          split_or n s
        and split_or n s =
           let fun remove_prem p s =
                 state_map_prems (fn ps => remove Thm.eq_thm_prop p ps) s
               
               
               fun elim_or p =
                 case prop_of p of
                     @{term Trueprop} $ (@{const HOL.disj} $ _ $ _) =>
                     SOME (cut_facts_tac [ p ] 1 THEN
                           REPEAT_ALL_NEW (etac @{thm HOL.disjE}) 1 THEN
                           (* must make progress after the split *)
                           re_focus1 n (remove_prem p s) 1)
                   | _ => NONE
           in
               FIRST (map_filter elim_or (#prems s))
           end
      *)
  in
      re_focus1 (n + 1) (state_init_empty ctx)
  end

fun steps_wrapper ctx conj_both n =
  let val safe_elims =
       claset_of ctx |> Classical.rep_cs |> #safeEs
                     |> Item_Net.content
                     |> filter (has_fewer_prems 3)
      (* clarify does too much, involving negations *)      
  in
   SELECT_GOAL (* avoid case splits *)
      (REPEAT_DETERM (eresolve_tac safe_elims 1) THEN
       TRY (cleanup_heaps ctx 1) THEN
       IF_UNSOLVED
         ((steps_tac ctx n conj_both THEN_ALL_NEW
          (cleanup_heaps ctx THEN'
           cleanup_split_disj ctx)) 1))
  end


fun run_tac ctx conj_both =  
  steps_wrapper ctx conj_both ~1

fun step_tac ctx n = 
  steps_wrapper ctx false 1

(* Case splitting *)
fun split_or_tac ctx i =
  let fun APPEND tacs = fold_rev (curry op Seq.APPEND) tacs no_tac;
      fun elim_or p =
        case prop_of p of
            @{term Trueprop} $ (@{const HOL.disj} $ _ $ _) =>
            SOME (DETERM (cut_facts_tac [ p ] 1 THEN
                             REPEAT_ALL_NEW (etac @{thm HOL.disjE}) 1))
          | _ => NONE
  in
      (Subgoal.FOCUS (fn f => APPEND (map_filter elim_or (#prems f))) ctx THEN_ALL_NEW
          DETERM o cleanup_split_disj ctx) i
  end


val setup =
   Method.setup (Binding.name "run")
   (Scan.succeed (fn ctx => SIMPLE_METHOD (run_tac ctx false 1)))
   "automatic symbolic execution" #>
   Method.setup (Binding.name "run_all")
   (Scan.succeed (fn ctx => SIMPLE_METHOD (run_tac ctx true 1)))
   "automatic symbolic execution, execute both branches of a conjunction" #>
   Method.setup (Binding.name "step")
   (Scan.lift (Scan.optional Parse.nat 1) >> (fn n => (fn ctx => SIMPLE_METHOD (step_tac ctx n 1))))
   "single step symbolic execution" #>
   Method.setup (Binding.name "split_or")
   (Scan.succeed (fn ctx => SIMPLE_METHOD (split_or_tac ctx 1)))
   "split case distinction by or in premises" #>   
   sym_exec_trace_setup

end